/*! \file DynDatabaseOverride.cpp

    \brief DynRPG plugin to override default database values with tab-separated values (.txt) files.

    \author Bernard J. Badger, AKA Aubrey the Bard

    DynDatabaseOverride is a DynRPG plugin which overrides the default values in an RPG Maker 2003
    game's database with the values provided in tab-separated values (.txt) files. It can do so at
    the start of a new game, in mid-game via comment command, and when a game is loaded with memory
    of previously applied overrides.

    NOTE: While the data files used by this plugin are in fact tab-separated values files with a
    standard ending of ".txt", there are frequent references to comma-separated values (CSV, .csv)
    in this code. This is because the plugin was originally written to actually use the ".csv" file
    ending, even though the cells were separated by tabs. Then I learned that Microsoft Excel
    expects tab-separated values files to have a ".txt" ending (OpenOffice Calc is smart enough to
    work with whatever it's given). So I changed the file endings to ".txt", but did not bother
    refactoring all of the code to match. Simply understand that references to CSV and .csv are
    conceptual to data files with cells delimited by some special character, not specifically
    comma-separated values.
*/

//#define DYNDATABASEOVERRIDE_DEBUG              // Comment out to remove debugging

#include <DynRPG/DynRPG.h>
#include <cstdlib>
#include <list>
#include <set>
#include <string>
#include <fstream>
#include "CSVIterator.h"
#ifdef DYNDATABASEOVERRIDE_DEBUG
#include <sstream>          // Only needed for debug purposes
#include <iostream>         // Only needed for debug purposes
#endif // DYNDATABASEOVERRIDE_DEBUG

// Use this as a template for debug popups
// DEBUG
#ifdef DYNDATABASEOVERRIDE_DEBUG
//std::stringstream message;
//message << "message\n";
//MessageBox(
//            NULL, // We don't need a window handle
//            message.str().c_str(), // Text
//            "Debug", // Title
//            MB_OK // Flags
//);
#endif // DYNDATABASEOVERRIDE_DEBUG
// /DEBUG

// Use this as a template for writing to the debug file
// DEBUG
#ifdef DYNDATABASEOVERRIDE_DEBUG
//*ofDebug << "message\n";
#endif // DYNDATABASEOVERRIDE_DEBUG
// /DEBUG

//! Types of .csv file
enum CSVFileType
{
    CSV_FILE_TYPE_ACTOR,                                //!< Actor (not including array values like HP, MP, etc.)
    CSV_FILE_TYPE_ACTOR_MAX_HP,                         //!< Actor max HP
    CSV_FILE_TYPE_ACTOR_MAX_MP,                         //!< Actor max MP
    CSV_FILE_TYPE_ACTOR_ATTACK,                         //!< Actor attack
    CSV_FILE_TYPE_ACTOR_DEFENSE,                        //!< Actor defense
    CSV_FILE_TYPE_ACTOR_INTELLIGENCE,                   //!< Actor intelligence
    CSV_FILE_TYPE_ACTOR_AGILITY,                        //!< Actor agility
    CSV_FILE_TYPE_ACTOR_SKILL_LEARNING,                 //!< Actor skill learning
    CSV_FILE_TYPE_ACTOR_CONDITIONS,                     //!< Actor conditions
    CSV_FILE_TYPE_ACTOR_ATTRIBUTES,                     //!< Actor attributes
    CSV_FILE_TYPE_CLASS,                                //!< Class (not including array values like HP, MP, etc.)
    CSV_FILE_TYPE_CLASS_MAX_HP,                         //!< Class max HP
    CSV_FILE_TYPE_CLASS_MAX_MP,                         //!< Class max MP
    CSV_FILE_TYPE_CLASS_ATTACK,                         //!< Class attack
    CSV_FILE_TYPE_CLASS_DEFENSE,                        //!< Class defense
    CSV_FILE_TYPE_CLASS_INTELLIGENCE,                   //!< Class intelligence
    CSV_FILE_TYPE_CLASS_AGILITY,                        //!< Class agility
    CSV_FILE_TYPE_CLASS_SKILL_LEARNING,                 //!< Class skill learning
    CSV_FILE_TYPE_CLASS_CONDITIONS,                     //!< Class conditions
    CSV_FILE_TYPE_CLASS_ATTRIBUTES,                     //!< Class attributes
    CSV_FILE_TYPE_ENEMY,                                //!< Enemy (not including array values like conditions, attributes, etc.)
    CSV_FILE_TYPE_ENEMY_CONDITIONS,                     //!< Enemy conditions
    CSV_FILE_TYPE_ENEMY_ATTRIBUTES,                     //!< Enemy attributes
    CSV_FILE_TYPE_ENEMY_BEHAVIOR,                       //!< Enemy behavior
    CSV_FILE_TYPE_ENEMY_GROUP,                          //!< Enemy group (not including array values like monster list, habitat terrain, etc.)
    CSV_FILE_TYPE_ENEMY_GROUP_ENEMY_LIST,               //!< Enemy group monster list
    CSV_FILE_TYPE_ENEMY_GROUP_HABITAT_TERRAIN,          //!< Enemy group habitat terrain
    CSV_FILE_TYPE_ITEM,                                 //!< Item (not including array values like conditions, attributes, etc.)
    CSV_FILE_TYPE_ITEM_USABLE_BY_HERO,                  //!< Item usable by hero
    CSV_FILE_TYPE_ITEM_CONDITIONS,                      //!< Item conditions
    CSV_FILE_TYPE_ITEM_ATTRIBUTES,                      //!< Item attributes
    CSV_FILE_TYPE_ITEM_WEAPON_ANIMS,                    //!< Item weapon anims
    CSV_FILE_TYPE_ITEM_USABLE_BY_CLASS,                 //!< Item usable by class
    CSV_FILE_TYPE_ITEM_CONDITION_FLAG,                  //!< Item condition flag
    CSV_FILE_TYPE_SKILL,                                //!< Skill (not including array values like conditions, attributes, etc.)
    CSV_FILE_TYPE_SKILL_CONDITIONS,                     //!< Skill conditions
    CSV_FILE_TYPE_SKILL_ATTRIBUTES,                     //!< Skill attributes
    CSV_FILE_TYPE_SKILL_SKILL_ANIMS,                    //!< Skill skill anims
    CSV_FILE_TYPE_TERRAIN,                              //!< Terrain
    CSV_FILE_TYPE_TILESET,                              //!< Tileset (not including array values like lower tile terrain ID, etc.)
    CSV_FILE_TYPE_TILESET_LOWER_TILE_TERRAIN_ID,        //!< Tileset lower tile terrain ID
    CSV_FILE_TYPE_TILESET_LOWER_TILE_PASSABILITY,       //!< Tileset lower tile passability
    CSV_FILE_TYPE_TILESET_UPPER_TILE_PASSABILITY        //!< Tileset upper tile passability
};

//! Indices of data in an actor .csv file
enum CSVIndexActor
{
    CSV_INDEX_ACTOR_NOTES_OPTIONAL,                     //!< Zero-based index of an actor's "Notes (optional)" data in the .csv file
    CSV_INDEX_ACTOR_ID,                                 //!< Zero-based index of an actor's "ID" data in the .csv file
    CSV_INDEX_ACTOR_NAME,                               //!< Zero-based index of an actor's "Name" data in the .csv file
    CSV_INDEX_ACTOR_DEGREE,                             //!< Zero-based index of an actor's "Degree" data in the .csv file
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    CSV_INDEX_ACTOR_CHARSET_FILE_NAME,                  //!< Zero-based index of an actor's "Charset file name" data in the .csv file
//    CSV_INDEX_ACTOR_CHARSET_ID,                         //!< Zero-based index of an actor's "Charset ID" data in the .csv file
//    CSV_INDEX_ACTOR_CHARSET_TRANSPARENT,                //!< Zero-based index of an actor's "Charset Transparent" data in the .csv file
    CSV_INDEX_ACTOR_MIN_LEVEL,                          //!< Zero-based index of an actor's "Min Level" data in the .csv file
    CSV_INDEX_ACTOR_MAX_LEVEL,                          //!< Zero-based index of an actor's "Max Level" data in the .csv file
    CSV_INDEX_ACTOR_ALLOW_CRITICAL_HIT,                 //!< Zero-based index of an actor's "Allow critical hit" data in the .csv file
    CSV_INDEX_ACTOR_CRITICAL_HIT_PROBABILITY,           //!< Zero-based index of an actor's "Critical hit probability" data in the .csv file
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    CSV_INDEX_ACTOR_FACESET_FILE_NAME,                  //!< Zero-based index of an actor's "Faceset file name" data in the .csv file
//    CSV_INDEX_ACTOR_FACESET_ID,                         //!< Zero-based index of an actor's "Faceset ID" data in the .csv file
    CSV_INDEX_ACTOR_TWO_WEAPONS,                        //!< Zero-based index of an actor's "Two weapons" data in the .csv file
    CSV_INDEX_ACTOR_AI_CONTROL,                         //!< Zero-based index of an actor's "AI control" data in the .csv file
    CSV_INDEX_ACTOR_LOCK_EQUIPMENT,                     //!< Zero-based index of an actor's "Lock equipment" data in the .csv file
    CSV_INDEX_ACTOR_MIGHTY_GUARD,                       //!< Zero-based index of an actor's "Mighty guard" data in the .csv file
    CSV_INDEX_ACTOR_EXP_PRIMARY,                        //!< Zero-based index of an actor's "Exp primary" data in the .csv file
    CSV_INDEX_ACTOR_EXP_SECONDARY,                      //!< Zero-based index of an actor's "Exp secondary" data in the .csv file
    CSV_INDEX_ACTOR_EXP_TERTIARY,                       //!< Zero-based index of an actor's "Exp tertiary" data in the .csv file
    CSV_INDEX_ACTOR_WEAPON_ID,                          //!< Zero-based index of an actor's "Weapon ID" data in the .csv file
    CSV_INDEX_ACTOR_SHIELD_ID,                          //!< Zero-based index of an actor's "Shield ID" data in the .csv file
    CSV_INDEX_ACTOR_ARMOR_ID,                           //!< Zero-based index of an actor's "Armor ID" data in the .csv file
    CSV_INDEX_ACTOR_HELMET_ID,                          //!< Zero-based index of an actor's "Helmet ID" data in the .csv file
    CSV_INDEX_ACTOR_ACCESSORY_ID,                       //!< Zero-based index of an actor's "Accessory ID" data in the .csv file
    CSV_INDEX_ACTOR_UNARMED_BATTLE_ANIMATION_ID,        //!< Zero-based index of an actor's "Unarmed battle animation ID" data in the .csv file
    CSV_INDEX_ACTOR_CLASS_ID,                           //!< Zero-based index of an actor's "Class ID" data in the .csv file
    CSV_INDEX_ACTOR_BATTLE_GRAPHIC_ID,                  //!< Zero-based index of an actor's "Battle graphic ID" data in the .csv file
    CSV_INDEX_ACTOR_X_POS_IN_BATTLE,                    //!< Zero-based index of an actor's "X pos in battle" data in the .csv file
    CSV_INDEX_ACTOR_Y_POS_IN_BATTLE                     //!< Zero-based index of an actor's "Y pos in battle" data in the .csv file
};

//! Indices of data in an actor stat (HP/MP/Attack/Defense/Intelligence/Agility) .csv file
enum CSVIndexActorStat
{
    CSV_INDEX_ACTOR_STAT_NOTES_OPTIONAL,                //!< Zero-based index of an actor stat's "Notes (optional)" data in the .csv file
    CSV_INDEX_ACTOR_STAT_ID,                            //!< Zero-based index of an actor stat's "ID" data in the .csv file
    CSV_INDEX_ACTOR_STAT_LEVEL_1                        //!< Zero-based index of an actor stat's "Level 1" data in the .csv file (levels past this are simply calculated)
};

//! Indices of data in an actor skill learning .csv file
enum CSVIndexActorSkillLearning
{
    CSV_INDEX_ACTOR_SKILL_LEARNING_NOTES_OPTIONAL,      //!< Zero-based index of an actor skill learning's "Notes (optional)" data in the .csv file
    CSV_INDEX_ACTOR_SKILL_LEARNING_ID,                  //!< Zero-based index of an actor skill learning's "ID" data in the .csv file
    CSV_INDEX_ACTOR_SKILL_LEARNING_ACTOR_ID,            //!< Zero-based index of an actor skill learning's "Actor ID" data in the .csv file
    CSV_INDEX_ACTOR_SKILL_LEARNING_LEVEL_REQUIREMENT,   //!< Zero-based index of an actor skill learning's "Level requirement" data in the .csv file (levels past this are simply calculated)
    CSV_INDEX_ACTOR_SKILL_LEARNING_SKILL_ID             //!< Zero-based index of an actor skill learning's "Skill ID" data in the .csv file (levels past this are simply calculated)
};

//! Indices of data in an actor condition .csv file
enum CSVIndexActorCondition
{
    CSV_INDEX_ACTOR_CONDITION_NOTES_OPTIONAL,           //!< Zero-based index of an actor condition's "Notes (optional)" data in the .csv file
    CSV_INDEX_ACTOR_CONDITION_ACTOR_ID,                 //!< Zero-based index of an actor condition's "Actor ID" data in the .csv file
    CSV_INDEX_ACTOR_CONDITION_CONDITION_ID,             //!< Zero-based index of an actor condition's "Condition ID" data in the .csv file
    CSV_INDEX_ACTOR_CONDITION_SUSCEPTIBILITY_RANK       //!< Zero-based index of an actor condition's "Susceptibility rank" data in the .csv file
};

//! Indices of data in an actor attributes .csv file
enum CSVIndexActorAttribute
{
    CSV_INDEX_ACTOR_ATTRIBUTE_NOTES_OPTIONAL,           //!< Zero-based index of an actor attribute's "Notes (optional)" data in the .csv file
    CSV_INDEX_ACTOR_ATTRIBUTE_ACTOR_ID,                 //!< Zero-based index of an actor attribute's "Actor ID" data in the .csv file
    CSV_INDEX_ACTOR_ATTRIBUTE_ATTRIBUTE_ID,             //!< Zero-based index of an actor attribute's "Attribute ID" data in the .csv file
    CSV_INDEX_ACTOR_ATTRIBUTE_SUSCEPTIBILITY_RANK       //!< Zero-based index of an actor attribute's "Susceptibility rank" data in the .csv file
};

//! Indices of data in an class .csv file
enum CSVIndexClass
{
    CSV_INDEX_CLASS_NOTES_OPTIONAL,                     //!< Zero-based index of an class's "Notes (optional)" data in the .csv file
    CSV_INDEX_CLASS_ID,                                 //!< Zero-based index of an class's "ID" data in the .csv file
    CSV_INDEX_CLASS_NAME,                               //!< Zero-based index of an class's "Name" data in the .csv file
    CSV_INDEX_CLASS_TWO_WEAPONS,                        //!< Zero-based index of an class's "Two weapons" data in the .csv file
    CSV_INDEX_CLASS_LOCK_EQUIPMENT,                     //!< Zero-based index of an class's "Lock equipment" data in the .csv file
    CSV_INDEX_CLASS_AI_CONTROL,                         //!< Zero-based index of an class's "AI control" data in the .csv file
    CSV_INDEX_CLASS_MIGHTY_GUARD,                       //!< Zero-based index of an class's "Mighty guard" data in the .csv file
    CSV_INDEX_CLASS_EXP_PRIMARY,                        //!< Zero-based index of an class's "Exp primary" data in the .csv file
    CSV_INDEX_CLASS_EXP_SECONDARY,                      //!< Zero-based index of an class's "Exp secondary" data in the .csv file
    CSV_INDEX_CLASS_EXP_TERTIARY,                       //!< Zero-based index of an class's "Exp tertiary" data in the .csv file
    CSV_INDEX_CLASS_BATTLE_GRAPHIC_ID,                  //!< Zero-based index of an class's "Battle graphic ID" data in the .csv file
    CSV_INDEX_CLASS_BATTLE_COMMAND_0                    //!< Zero-based index of an class's "Battle command 0" data in the .csv file (battle commands past this are simply calculated)
};

//! Indices of data in an class stat (HP/MP/Attack/Defense/Intelligence/Agility) .csv file
enum CSVIndexClassStat
{
    CSV_INDEX_CLASS_STAT_NOTES_OPTIONAL,                //!< Zero-based index of an class stat's "Notes (optional)" data in the .csv file
    CSV_INDEX_CLASS_STAT_ID,                            //!< Zero-based index of an class stat's "ID" data in the .csv file
    CSV_INDEX_CLASS_STAT_LEVEL_1                        //!< Zero-based index of an class stat's "Level 1" data in the .csv file (levels past this are simply calculated)
};

//! Indices of data in an class skill learning .csv file
enum CSVIndexClassSkillLearning
{
    CSV_INDEX_CLASS_SKILL_LEARNING_NOTES_OPTIONAL,      //!< Zero-based index of an class skill learning's "Notes (optional)" data in the .csv file
    CSV_INDEX_CLASS_SKILL_LEARNING_ID,                  //!< Zero-based index of an class skill learning's "ID" data in the .csv file
    CSV_INDEX_CLASS_SKILL_LEARNING_CLASS_ID,            //!< Zero-based index of an class skill learning's "Class ID" data in the .csv file
    CSV_INDEX_CLASS_SKILL_LEARNING_LEVEL_REQUIREMENT,   //!< Zero-based index of an class skill learning's "Level requirement" data in the .csv file (levels past this are simply calculated)
    CSV_INDEX_CLASS_SKILL_LEARNING_SKILL_ID             //!< Zero-based index of an class skill learning's "Skill ID" data in the .csv file (levels past this are simply calculated)
};

//! Indices of data in an class condition .csv file
enum CSVIndexClassCondition
{
    CSV_INDEX_CLASS_CONDITION_NOTES_OPTIONAL,           //!< Zero-based index of an class condition's "Notes (optional)" data in the .csv file
    CSV_INDEX_CLASS_CONDITION_CLASS_ID,                 //!< Zero-based index of an class condition's "Class ID" data in the .csv file
    CSV_INDEX_CLASS_CONDITION_CONDITION_ID,             //!< Zero-based index of an class condition's "Condition ID" data in the .csv file
    CSV_INDEX_CLASS_CONDITION_SUSCEPTIBILITY_RANK       //!< Zero-based index of an class condition's "Susceptibility rank" data in the .csv file
};

//! Indices of data in an class attributes .csv file
enum CSVIndexClassAttribute
{
    CSV_INDEX_CLASS_ATTRIBUTE_NOTES_OPTIONAL,           //!< Zero-based index of an class attribute's "Notes (optional)" data in the .csv file
    CSV_INDEX_CLASS_ATTRIBUTE_CLASS_ID,                 //!< Zero-based index of an class attribute's "Class ID" data in the .csv file
    CSV_INDEX_CLASS_ATTRIBUTE_ATTRIBUTE_ID,             //!< Zero-based index of an class attribute's "Attribute ID" data in the .csv file
    CSV_INDEX_CLASS_ATTRIBUTE_SUSCEPTIBILITY_RANK       //!< Zero-based index of an class attribute's "Susceptibility rank" data in the .csv file
};

//! Indices of data in a skill .csv file
enum CSVIndexSkill
{
    CSV_INDEX_SKILL_NOTES_OPTIONAL,                     //!< Zero-based index of a skill's "Notes (optional)" data in the .csv file
    CSV_INDEX_SKILL_ID,                                 //!< Zero-based index of a skill's "ID" data in the .csv file
    CSV_INDEX_SKILL_NAME,                               //!< Zero-based index of a skill's "Name" data in the .csv file
    CSV_INDEX_SKILL_DESCRIPTION,                        //!< Zero-based index of a skill's "Description" data in the .csv file
    CSV_INDEX_SKILL_USE_MESSAGE_2K,                     //!< Zero-based index of a skill's "Use message 2K" data in the .csv file
    CSV_INDEX_SKILL_USE_MESSAGE_2K_ALT,                 //!< Zero-based index of a skill's "Use message 2K alt" data in the .csv file
    CSV_INDEX_SKILL_EVASION_TYPE,                       //!< Zero-based index of a skill's "Evasion type" data in the .csv file
    CSV_INDEX_SKILL_TYPE,                               //!< Zero-based index of a skill's "Type" data in the .csv file
    CSV_INDEX_SKILL_MP_COST,                            //!< Zero-based index of a skill's "MP cost" data in the .csv file
    CSV_INDEX_SKILL_TARGET,                             //!< Zero-based index of a skill's "Target" data in the .csv file
    CSV_INDEX_SKILL_SWITCH_ID,                          //!< Zero-based index of a skill's "Switch ID" data in the .csv file
    CSV_INDEX_SKILL_BATTLE_ANIM_ID,                     //!< Zero-based index of a skill's "Battle animation ID" data in the .csv file
    CSV_INDEX_SKILL_USE_IN_FIELD,                       //!< Zero-based index of a skill's "Use in field" data in the .csv file
    CSV_INDEX_SKILL_USE_IN_BATTLE,                      //!< Zero-based index of a skill's "Use in battle" data in the .csv file
    CSV_INDEX_SKILL_ATTACK_INFLUENCE,                   //!< Zero-based index of a skill's "Attack influence" data in the .csv file
    CSV_INDEX_SKILL_INTELLIGENCE_INFLUENCE,             //!< Zero-based index of a skill's "Intelligence influence" data in the .csv file
    CSV_INDEX_SKILL_VARIANCE,                           //!< Zero-based index of a skill's "Variance" data in the .csv file
    CSV_INDEX_SKILL_EFFECT_RATING,                      //!< Zero-based index of a skill's "Effect rating" data in the .csv file
    CSV_INDEX_SKILL_SUCCESS_RATE,                       //!< Zero-based index of a skill's "Success rate" data in the .csv file
    CSV_INDEX_SKILL_AFFECT_HP,                          //!< Zero-based index of a skill's "Affect HP" data in the .csv file
    CSV_INDEX_SKILL_AFFECT_MP,                          //!< Zero-based index of a skill's "Affect MP" data in the .csv file
    CSV_INDEX_SKILL_AFFECT_STR,                         //!< Zero-based index of a skill's "Affect str" data in the .csv file
    CSV_INDEX_SKILL_AFFECT_DEF,                         //!< Zero-based index of a skill's "Affect def" data in the .csv file
    CSV_INDEX_SKILL_AFFECT_INT,                         //!< Zero-based index of a skill's "Affect int" data in the .csv file
    CSV_INDEX_SKILL_AFFECT_AGI,                         //!< Zero-based index of a skill's "Affect agi" data in the .csv file
    CSV_INDEX_SKILL_ABSORB,                             //!< Zero-based index of a skill's "Absorb" data in the .csv file
    CSV_INDEX_SKILL_IGNORE_DEF,                         //!< Zero-based index of a skill's "Ignore def" data in the .csv file
    CSV_INDEX_SKILL_ATTRIBUTE_RESIST,                   //!< Zero-based index of a skill's "Attribute resist" data in the .csv file
    CSV_INDEX_SKILL_IS_MP_PERCENT,                      //!< Zero-based index of a skill's "Is MP percent" data in the .csv file
    CSV_INDEX_SKILL_MP_COST_PERCENT,                    //!< Zero-based index of a skill's "MP cost percent" data in the .csv file
    CSV_INDEX_SKILL_CONDITION_FLAG                      //!< Zero-based index of a skill's "Condition flag" data in the .csv file
};

//! Indices of data in a skill condition .csv file
enum CSVIndexSkillCondition
{
    CSV_INDEX_SKILL_CONDITION_NOTES_OPTIONAL,           //!< Zero-based index of a skill condition's "Notes (optional)" data in the .csv file
    CSV_INDEX_SKILL_CONDITION_SKILL_ID,                 //!< Zero-based index of a skill condition's "Skill ID" data in the .csv file
    CSV_INDEX_SKILL_CONDITION_CONDITION_ID,             //!< Zero-based index of a skill condition's "Condition ID" data in the .csv file
    CSV_INDEX_SKILL_CONDITION_AFFECTS_CONDITION         //!< Zero-based index of a skill condition's "Affects condition" data in the .csv file
};

//! Indices of data in a skill attribute .csv file
enum CSVIndexSkillAttribute
{
    CSV_INDEX_SKILL_ATTRIBUTE_NOTES_OPTIONAL,           //!< Zero-based index of a skill attribute's "Notes (optional)" data in the .csv file
    CSV_INDEX_SKILL_ATTRIBUTE_SKILL_ID,                 //!< Zero-based index of a skill attribute's "Skill ID" data in the .csv file
    CSV_INDEX_SKILL_ATTRIBUTE_CONDITION_ID,             //!< Zero-based index of a skill attribute's "Condition ID" data in the .csv file
    CSV_INDEX_SKILL_ATTRIBUTE_HAS_ATTRIBUTE             //!< Zero-based index of a skill attribute's "Has attribute" data in the .csv file
};

//! Indices of data in a skill animation .csv file
enum CSVIndexSkillAnimation
{
    CSV_INDEX_SKILL_ANIMATION_NOTES_OPTIONAL,           //!< Zero-based index of a skill animation's "Notes (optional)" data in the .csv file
    CSV_INDEX_SKILL_ANIMATION_SKILL_ID,                 //!< Zero-based index of a skill animation's "Skill ID" data in the .csv file
    CSV_INDEX_SKILL_ANIMATION_ACTOR_ID,                 //!< Zero-based index of a skill animation's "Actor ID" data in the .csv file
    CSV_INDEX_SKILL_ANIMATION_BATTLER_POSE,             //!< Zero-based index of a skill animation's "Battler pose" data in the .csv file
    CSV_INDEX_SKILL_ANIMATION_MOVE_BEFORE_ATTACK,       //!< Zero-based index of a skill animation's "Move before attack" data in the .csv file
    CSV_INDEX_SKILL_ANIMATION_ADD_AFTER_IMAGE           //!< Zero-based index of a skill animation's "Add after image" data in the .csv file
};

//! Indices of data in an item .csv file
enum CSVIndexItem
{
    CSV_INDEX_ITEM_NOTES_OPTIONAL,                      //!< Zero-based index of an item's "Notes (optional)" data in the .csv file
    CSV_INDEX_ITEM_ID,                                  //!< Zero-based index of an item's "ID" data in the .csv file
    CSV_INDEX_ITEM_NAME,                                //!< Zero-based index of an item's "Name" data in the .csv file
    CSV_INDEX_ITEM_DESCRIPTION,                         //!< Zero-based index of an item's "Description" data in the .csv file
    CSV_INDEX_ITEM_TYPE,                                //!< Zero-based index of an item's "Type" data in the .csv file
    CSV_INDEX_ITEM_PRICE,                               //!< Zero-based index of an item's "Price" data in the .csv file
    CSV_INDEX_ITEM_NUM_USES,                            //!< Zero-based index of an item's "Number of uses" data in the .csv file
    CSV_INDEX_ITEM_ATTACK,                              //!< Zero-based index of an item's "Attack" data in the .csv file
    CSV_INDEX_ITEM_DEFENSE,                             //!< Zero-based index of an item's "Defense" data in the .csv file
    CSV_INDEX_ITEM_INTELLIGENCE,                        //!< Zero-based index of an item's "Intelligence" data in the .csv file
    CSV_INDEX_ITEM_AGILITY,                             //!< Zero-based index of an item's "Agility" data in the .csv file
    CSV_INDEX_ITEM_TWO_HANDED,                          //!< Zero-based index of an item's "Two-handed" data in the .csv file
    CSV_INDEX_ITEM_MP_CONSUMED,                         //!< Zero-based index of an item's "MP consumed" data in the .csv file
    CSV_INDEX_ITEM_HIT_PROBABILITY,                     //!< Zero-based index of an item's "Hit probability" data in the .csv file
    CSV_INDEX_ITEM_CRITICAL_BONUS,                      //!< Zero-based index of an item's "Critical bonus" data in the .csv file
    CSV_INDEX_ITEM_BATTLE_ANIMATION_ID,                 //!< Zero-based index of an item's "Battle animation ID" data in the .csv file
    CSV_INDEX_ITEM_ALWAYS_ATTACK_FIRST,                 //!< Zero-based index of an item's "Always attack first" data in the .csv file
    CSV_INDEX_ITEM_ATTACK_TWICE,                        //!< Zero-based index of an item's "Attack twice" data in the .csv file
    CSV_INDEX_ITEM_ATTACK_ALL,                          //!< Zero-based index of an item's "Attack all" data in the .csv file
    CSV_INDEX_ITEM_IGNORE_ENEMY_EVASION,                //!< Zero-based index of an item's "Ignore enemy evasion" data in the .csv file
    CSV_INDEX_ITEM_PREVENT_CRITICAL_HITS,               //!< Zero-based index of an item's "Prevent critical hits" data in the .csv file
    CSV_INDEX_ITEM_INCREASED_EVASION,                   //!< Zero-based index of an item's "Increased evasion" data in the .csv file
    CSV_INDEX_ITEM_HALF_MP_CONSUMPTION,                 //!< Zero-based index of an item's "Half MP consumption" data in the .csv file
    CSV_INDEX_ITEM_NO_TERRAIN_DAMAGE,                   //!< Zero-based index of an item's "No terrain damage" data in the .csv file
    CSV_INDEX_ITEM_CURSED,                              //!< Zero-based index of an item's "Cursed" data in the .csv file
    CSV_INDEX_ITEM_TARGET,                              //!< Zero-based index of an item's "Target" data in the .csv file
    CSV_INDEX_ITEM_HP_RESTORE_PERCENT,                  //!< Zero-based index of an item's "HP restore percent" data in the .csv file
    CSV_INDEX_ITEM_HP_RESTORE_UNITS,                    //!< Zero-based index of an item's "HP restore units" data in the .csv file
    CSV_INDEX_ITEM_MP_RESTORE_PERCENT,                  //!< Zero-based index of an item's "MP restore percent" data in the .csv file
    CSV_INDEX_ITEM_MP_RESTORE_UNITS,                    //!< Zero-based index of an item's "MP restore units" data in the .csv file
    CSV_INDEX_ITEM_ONLY_IN_FIELD,                       //!< Zero-based index of an item's "Only usable on the field map" data in the .csv file
    CSV_INDEX_ITEM_ONLY_ON_DEAD,                        //!< Zero-based index of an item's "Only usable on dead characters" data in the .csv file
    CSV_INDEX_ITEM_ADD_MAX_HP,                          //!< Zero-based index of an item's "Add max HP" data in the .csv file
    CSV_INDEX_ITEM_ADD_MAX_MP,                          //!< Zero-based index of an item's "Add max MP" data in the .csv file
    CSV_INDEX_ITEM_ADD_ATTACK,                          //!< Zero-based index of an item's "Add attack" data in the .csv file
    CSV_INDEX_ITEM_ADD_DEFENSE,                         //!< Zero-based index of an item's "Add defense" data in the .csv file
    CSV_INDEX_ITEM_ADD_INTELLIGENCE,                    //!< Zero-based index of an item's "Add intelligence" data in the .csv file
    CSV_INDEX_ITEM_ADD_AGILITY,                         //!< Zero-based index of an item's "Add agility" data in the .csv file
    CSV_INDEX_ITEM_MESSAGE_SHOWN,                       //!< Zero-based index of an item's "Message shown when used" data in the .csv file
    CSV_INDEX_ITEM_INVOKE_SKILL,                        //!< Zero-based index of an item's "Invoke skill when used" data in the .csv file
    CSV_INDEX_ITEM_SKILL_ID,                            //!< Zero-based index of an item's "Skill ID" data in the .csv file
    CSV_INDEX_ITEM_SWITCH_ID,                           //!< Zero-based index of an item's "Switch ID" data in the .csv file
    CSV_INDEX_ITEM_USE_IN_FIELD,                        //!< Zero-based index of an item's "Activation allowed in field" data in the .csv file
    CSV_INDEX_ITEM_USE_IN_BATTLE,                       //!< Zero-based index of an item's "Activation allowed in battle" data in the .csv file
    CSV_INDEX_ITEM_CONDITION_EFFECTIVENESS,             //!< Zero-based index of an item's "Condition effectiveness" data in the .csv file
    CSV_INDEX_ITEM_CONDITION_RESIST_OR_INFLICT,         //!< Zero-based index of an item's "Condition resist or inflict" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_TRAJECTORY,                   //!< Zero-based index of an item's "Weapon trajectory" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_TARGETTING                    //!< Zero-based index of an item's "Weapon targetting" data in the .csv file
};

//! Indices of data in an item usable by actor .csv file
enum CSVIndexItemUsableByActor
{
    CSV_INDEX_ITEM_USABLE_BY_ACTOR_NOTES_OPTIONAL,      //!< Zero-based index of an item usable by actor's "Notes (optional)" data in the .csv file
    CSV_INDEX_ITEM_USABLE_BY_ACTOR_ITEM_ID,             //!< Zero-based index of an item usable by actor's "Item ID" data in the .csv file
    CSV_INDEX_ITEM_USABLE_BY_ACTOR_ACTOR_ID,            //!< Zero-based index of an item usable by actor's "Actor ID" data in the .csv file
    CSV_INDEX_ITEM_USABLE_BY_ACTOR_CAN_USE              //!< Zero-based index of an item usable by actor's "Can use" data in the .csv file
};

//! Indices of data in an item usable by class .csv file
enum CSVIndexItemUsableByClass
{
    CSV_INDEX_ITEM_USABLE_BY_CLASS_NOTES_OPTIONAL,      //!< Zero-based index of an item usable by class's "Notes (optional)" data in the .csv file
    CSV_INDEX_ITEM_USABLE_BY_CLASS_ITEM_ID,             //!< Zero-based index of an item usable by class's "Item ID" data in the .csv file
    CSV_INDEX_ITEM_USABLE_BY_CLASS_ACTOR_ID,            //!< Zero-based index of an item usable by class's "Actor ID" data in the .csv file
    CSV_INDEX_ITEM_USABLE_BY_CLASS_CAN_USE              //!< Zero-based index of an item usable by class's "Can use" data in the .csv file
};

//! Indices of data in an item condition .csv file
enum CSVIndexItemCondition
{
    CSV_INDEX_ITEM_CONDITION_NOTES_OPTIONAL,            //!< Zero-based index of an item condition's "Notes (optional)" data in the .csv file
    CSV_INDEX_ITEM_CONDITION_ITEM_ID,                   //!< Zero-based index of an item condition's "Item ID" data in the .csv file
    CSV_INDEX_ITEM_CONDITION_CONDITION_ID,              //!< Zero-based index of an item condition's "Condition ID" data in the .csv file
    CSV_INDEX_ITEM_CONDITION_AFFECTS_CONDITION          //!< Zero-based index of an item condition's "Affects condition" data in the .csv file
};

//! Indices of data in an item attribute .csv file
enum CSVIndexItemAttribute
{
    CSV_INDEX_ITEM_ATTRIBUTE_NOTES_OPTIONAL,            //!< Zero-based index of an item attribute's "Notes (optional)" data in the .csv file
    CSV_INDEX_ITEM_ATTRIBUTE_ITEM_ID,                   //!< Zero-based index of an item attribute's "Item ID" data in the .csv file
    CSV_INDEX_ITEM_ATTRIBUTE_ATTRIBUTE_ID,              //!< Zero-based index of an item attribute's "Attribute ID" data in the .csv file
    CSV_INDEX_ITEM_ATTRIBUTE_HAS_ATTRIBUTE              //!< Zero-based index of an item attribute's "Has attribute" data in the .csv file
};

//! Indices of data in an item attribute .csv file
enum CSVIndexItemWeaponAnimation
{
    CSV_INDEX_ITEM_WEAPON_ANIMATION_NOTES_OPTIONAL,         //!< Zero-based index of an item weapon animation's "Notes (optional)" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_ITEM_ID,                //!< Zero-based index of an item weapon animation's "Item ID" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_ACTOR_ID,               //!< Zero-based index of an item weapon animation's "Actor ID" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_BATTLER_POSE,           //!< Zero-based index of an item weapon animation's "Battler pose" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_ANIMATION_TYPE,         //!< Zero-based index of an item weapon animation's "Animation type" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_WEAPON_ANIMATION_ID,    //!< Zero-based index of an item weapon animation's "Weapon animation ID" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_MOVEMENT,               //!< Zero-based index of an item weapon animation's "Movement before attack" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_ADD_AFTERIMAGE,         //!< Zero-based index of an item weapon animation's "Add afterimage" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_NUM_ATTACKS,            //!< Zero-based index of an item weapon animation's "Number of attacks" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_RANGED_WEAPON,          //!< Zero-based index of an item weapon animation's "Ranged weapon" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_RANGED_ANIMATION_ID,    //!< Zero-based index of an item weapon animation's "Ranged weapon animation ID" data in the .csv file
    CSV_INDEX_ITEM_WEAPON_ANIMATION_RANGED_ANIMATION_SPEED  //!< Zero-based index of an item weapon animation's "Ranged weapon animation speed" data in the .csv file
};

//! Indices of data in an enemy .csv file
enum CSVIndexEnemy
{
    CSV_INDEX_ENEMY_NOTES_OPTIONAL,                     //!< Zero-based index of an enemy's "Notes (optional)" data in the .csv file
    CSV_INDEX_ENEMY_ID,                                 //!< Zero-based index of an enemy's "ID" data in the .csv file
    CSV_INDEX_ENEMY_NAME,                               //!< Zero-based index of an enemy's "Name" data in the .csv file
    CSV_INDEX_ENEMY_FILE_NAME,                          //!< Zero-based index of an enemy's "File name" data in the .csv file
    CSV_INDEX_ENEMY_COLOR_ALTERATION,                   //!< Zero-based index of an enemy's "Color alteration" data in the .csv file
    CSV_INDEX_ENEMY_MAX_HP,                             //!< Zero-based index of an enemy's "Max HP" data in the .csv file
    CSV_INDEX_ENEMY_MAX_MP,                             //!< Zero-based index of an enemy's "Max MP" data in the .csv file
    CSV_INDEX_ENEMY_ATTACK,                             //!< Zero-based index of an enemy's "Attack" data in the .csv file
    CSV_INDEX_ENEMY_DEFENSE,                            //!< Zero-based index of an enemy's "Defense" data in the .csv file
    CSV_INDEX_ENEMY_INTELLIGENCE,                       //!< Zero-based index of an enemy's "Agility" data in the .csv file
    CSV_INDEX_ENEMY_AGILITY,                            //!< Zero-based index of an enemy's "Intelligence" data in the .csv file
    CSV_INDEX_ENEMY_IS_TRANSPARENT,                     //!< Zero-based index of an enemy's "Is transparent" data in the .csv file
    CSV_INDEX_ENEMY_EXP,                                //!< Zero-based index of an enemy's "Exp" data in the .csv file
    CSV_INDEX_ENEMY_GOLD,                               //!< Zero-based index of an enemy's "Gold" data in the .csv file
    CSV_INDEX_ENEMY_ITEM_ID,                            //!< Zero-based index of an enemy's "Item ID" data in the .csv file
    CSV_INDEX_ENEMY_ITEM_PROBABILITY,                   //!< Zero-based index of an enemy's "Item probability" data in the .csv file
    CSV_INDEX_ENEMY_ALLOW_CRITICAL_HIT,                 //!< Zero-based index of an enemy's "Allow critical hit" data in the .csv file
    CSV_INDEX_ENEMY_CRITICAL_HIT_PROBABILITY,           //!< Zero-based index of an enemy's "Critical hit probability" data in the .csv file
    CSV_INDEX_ENEMY_OFTEN_MISS,                         //!< Zero-based index of an enemy's "Often miss" data in the .csv file
    CSV_INDEX_ENEMY_IS_FLYING                           //!< Zero-based index of an enemy's "Is flying" data in the .csv file
};

//! Indices of data in an enemy condition .csv file
enum CSVIndexEnemyCondition
{
    CSV_INDEX_ENEMY_CONDITION_NOTES_OPTIONAL,           //!< Zero-based index of an enemy condition's "Notes (optional)" data in the .csv file
    CSV_INDEX_ENEMY_CONDITION_ENEMY_ID,                 //!< Zero-based index of an enemy condition's "Enemy ID" data in the .csv file
    CSV_INDEX_ENEMY_CONDITION_CONDITION_ID,             //!< Zero-based index of an enemy condition's "Condition ID" data in the .csv file
    CSV_INDEX_ENEMY_CONDITION_SUSCEPTIBILITY_RANK       //!< Zero-based index of an enemy condition's "Susceptibility rank" data in the .csv file
};

//! Indices of data in an enemy attributes .csv file
enum CSVIndexEnemyAttribute
{
    CSV_INDEX_ENEMY_ATTRIBUTE_NOTES_OPTIONAL,           //!< Zero-based index of an enemy attribute's "Notes (optional)" data in the .csv file
    CSV_INDEX_ENEMY_ATTRIBUTE_ENEMY_ID,                 //!< Zero-based index of an enemy attribute's "Enemy ID" data in the .csv file
    CSV_INDEX_ENEMY_ATTRIBUTE_ATTRIBUTE_ID,             //!< Zero-based index of an enemy attribute's "Attribute ID" data in the .csv file
    CSV_INDEX_ENEMY_ATTRIBUTE_SUSCEPTIBILITY_RANK       //!< Zero-based index of an enemy attribute's "Susceptibility rank" data in the .csv file
};

//! Indices of data in an enemy behavior .csv file
enum CSVIndexEnemyBehavior
{
    CSV_INDEX_ENEMY_BEHAVIOR_NOTES_OPTIONAL,            //!< Zero-based index of an enemy behavior's "Notes (optional)" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_ENEMY_ID,                  //!< Zero-based index of an enemy behavior's "Enemy ID" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_BEHAVIOR_ID,               //!< Zero-based index of an enemy behavior's "Behavior ID" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_PRECONDITION,              //!< Zero-based index of an enemy behavior's "Precondition" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_PRECONDITION_PARAM_1,      //!< Zero-based index of an enemy behavior's "Precondition param 1" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_PRECONDITION_PARAM_2,      //!< Zero-based index of an enemy behavior's "Precondition param 2" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_PRECONDITION_SWITCH_ID,    //!< Zero-based index of an enemy behavior's "Precondition switch ID" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_PRIORITY,                  //!< Zero-based index of an enemy behavior's "Priority" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_SWITCH_ON,                 //!< Zero-based index of an enemy behavior's "Switch on" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_SWITCH_ON_ID,              //!< Zero-based index of an enemy behavior's "Switch on ID" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_SWITCH_OFF,                //!< Zero-based index of an enemy behavior's "Switch off" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_SWITCH_OFF_ID,             //!< Zero-based index of an enemy behavior's "Switch off ID" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_ACTION_KIND,               //!< Zero-based index of an enemy behavior's "Action kind" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_BASIC_ACTION_TYPE,         //!< Zero-based index of an enemy behavior's "Basic action type" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_SKILL_ID,                  //!< Zero-based index of an enemy behavior's "Skill ID" data in the .csv file
    CSV_INDEX_ENEMY_BEHAVIOR_TRANSFORM_ENEMY_ID         //!< Zero-based index of an enemy behavior's "Transform enemy ID" data in the .csv file
};

//! Indices of data in an attribute .csv file
enum CSVIndexAttribute
{
    CSV_INDEX_ATTRIBUTE_NOTES_OPTIONAL,                 //!< Zero-based index of an attribute's "Notes (optional)" data in the .csv file
    CSV_INDEX_ATTRIBUTE_ID,                             //!< Zero-based index of an attribute's "ID" data in the .csv file
    CSV_INDEX_ATTRIBUTE_NAME,                           //!< Zero-based index of an attribute's "Name" data in the .csv file
    CSV_INDEX_ATTRIBUTE_TYPE,                           //!< Zero-based index of an attribute's "Type" data in the .csv file
    CSV_INDEX_ATTRIBUTE_DAMAGE_A,                       //!< Zero-based index of an attribute's "Damage A" data in the .csv file
    CSV_INDEX_ATTRIBUTE_DAMAGE_B,                       //!< Zero-based index of an attribute's "Damage B" data in the .csv file
    CSV_INDEX_ATTRIBUTE_DAMAGE_C,                       //!< Zero-based index of an attribute's "Damage C" data in the .csv file
    CSV_INDEX_ATTRIBUTE_DAMAGE_D,                       //!< Zero-based index of an attribute's "Damage D" data in the .csv file
    CSV_INDEX_ATTRIBUTE_DAMAGE_E                        //!< Zero-based index of an attribute's "Damage E" data in the .csv file
};

//! Indices of data in a condition .csv file
enum CSVIndexCondition
{
    CSV_INDEX_CONDITION_NOTES_OPTIONAL,                 //!< Zero-based index of a condition's "Notes (optional)" data in the .csv file
    CSV_INDEX_CONDITION_ID,                             //!< Zero-based index of a condition's "ID" data in the .csv file
    CSV_INDEX_CONDITION_NAME,                           //!< Zero-based index of a condition's "Name" data in the .csv file
    CSV_INDEX_CONDITION_PERSISTS_AFTER_BATTLE,          //!< Zero-based index of a condition's "Persists after battle" data in the .csv file
    CSV_INDEX_CONDITION_COLOR,                          //!< Zero-based index of a condition's "Color" data in the .csv file
    CSV_INDEX_CONDITION_PRIORITY,                       //!< Zero-based index of a condition's "Priority" data in the .csv file
    CSV_INDEX_CONDITION_ACTION_RESTRICTION,             //!< Zero-based index of a condition's "Action restriction" data in the .csv file
    CSV_INDEX_CONDITION_HIT_PERCENTAGE,                 //!< Zero-based index of a condition's "Hit percentage" data in the .csv file
    CSV_INDEX_CONDITION_EVADE_ALL_ATTACKS,              //!< Zero-based index of a condition's "Evade all attacks" data in the .csv file
    CSV_INDEX_CONDITION_REFLECT_SKILLS,                 //!< Zero-based index of a condition's "Reflect skills" data in the .csv file
    CSV_INDEX_CONDITION_LOCK_EQUIPMENT,                 //!< Zero-based index of a condition's "Equipment cannot be changed" data in the .csv file
    CSV_INDEX_CONDITION_SUSCEPTIBILITY_A,               //!< Zero-based index of a condition's "Susceptibility A" data in the .csv file
    CSV_INDEX_CONDITION_SUSCEPTIBILITY_B,               //!< Zero-based index of a condition's "Susceptibility B" data in the .csv file
    CSV_INDEX_CONDITION_SUSCEPTIBILITY_C,               //!< Zero-based index of a condition's "Susceptibility C" data in the .csv file
    CSV_INDEX_CONDITION_SUSCEPTIBILITY_D,               //!< Zero-based index of a condition's "Susceptibility D" data in the .csv file
    CSV_INDEX_CONDITION_SUSCEPTIBILITY_E,               //!< Zero-based index of a condition's "Susceptibility E" data in the .csv file
    CSV_INDEX_CONDITION_RECOVERY_TURNS,                 //!< Zero-based index of a condition's "Recovery turns" data in the .csv file
    CSV_INDEX_CONDITION_RECOVERY_TURNS_PERCENT,         //!< Zero-based index of a condition's "Recovery turns percent" data in the .csv file
    CSV_INDEX_CONDITION_RECOVERY_PHYSICAL_PERCENT,      //!< Zero-based index of a condition's "Recovery physical percent" data in the .csv file
    CSV_INDEX_CONDITION_BASE_STAT_ALTERATION,           //!< Zero-based index of a condition's "Base stat alteration" data in the .csv file
    CSV_INDEX_CONDITION_ALTER_ATTACK,                   //!< Zero-based index of a condition's "Alters attack" data in the .csv file
    CSV_INDEX_CONDITION_ALTER_DEFENSE,                  //!< Zero-based index of a condition's "Alters defense" data in the .csv file
    CSV_INDEX_CONDITION_ALTER_INTELLIGENCE,             //!< Zero-based index of a condition's "Alters intelligence" data in the .csv file
    CSV_INDEX_CONDITION_ALTER_AGILITY,                  //!< Zero-based index of a condition's "Alters agility" data in the .csv file
    CSV_INDEX_CONDITION_PREVENT_SKILL_BY_ATTACK,        //!< Zero-based index of a condition's "Prevent skill by attack" data in the .csv file
    CSV_INDEX_CONDITION_PREVENT_SKILL_ATTACK_INFLUENCE, //!< Zero-based index of a condition's "Prevent skill attack influence" data in the .csv file
    CSV_INDEX_CONDITION_PREVENT_SKILL_BY_INTELLIGENCE,  //!< Zero-based index of a condition's "Prevent skill by intelligence" data in the .csv file
    CSV_INDEX_CONDITION_PREVENT_SKILL_INTELLIGENCE_INFLUENCE,   //!< Zero-based index of a condition's "Prevent skill intelligence influence" data in the .csv file
    CSV_INDEX_CONDITION_HP_ALTERATION_TYPE,             //!< Zero-based index of a condition's "HP alteration type" data in the .csv file
    CSV_INDEX_CONDITION_HP_DAMAGE_PERCENT,              //!< Zero-based index of a condition's "HP damage percent" data in the .csv file
    CSV_INDEX_CONDITION_HP_DAMAGE_UNITS,                //!< Zero-based index of a condition's "HP damage units" data in the .csv file
    CSV_INDEX_CONDITION_HP_DAMAGE_ON_STEP,              //!< Zero-based index of a condition's "HP damage on step" data in the .csv file
    CSV_INDEX_CONDITION_HP_DAMAGE_STEP_NUM,             //!< Zero-based index of a condition's "HP damage step number" data in the .csv file
    CSV_INDEX_CONDITION_MP_ALTERATION_TYPE,             //!< Zero-based index of a condition's "MP alteration type" data in the .csv file
    CSV_INDEX_CONDITION_MP_DAMAGE_PERCENT,              //!< Zero-based index of a condition's "MP damage percent" data in the .csv file
    CSV_INDEX_CONDITION_MP_DAMAGE_UNITS,                //!< Zero-based index of a condition's "MP damage units" data in the .csv file
    CSV_INDEX_CONDITION_MP_DAMAGE_ON_STEP,              //!< Zero-based index of a condition's "MP damage on step" data in the .csv file
    CSV_INDEX_CONDITION_MP_DAMAGE_STEP_NUM,             //!< Zero-based index of a condition's "MP damage step number" data in the .csv file
    CSV_INDEX_CONDITION_ANIMATION_ID,                   //!< Zero-based index of a condition's "Animation ID" data in the .csv file
    CSV_INDEX_CONDITION_MESSAGE_2K_ALLY_STATE,          //!< Zero-based index of a condition's "Message 2K ally state" data in the .csv file
    CSV_INDEX_CONDITION_MESSAGE_2K_ENEMY_STATE,         //!< Zero-based index of a condition's "Message 2K enemy state" data in the .csv file
    CSV_INDEX_CONDITION_MESSAGE_2K_ALREADY_IN,          //!< Zero-based index of a condition's "Message 2K already in" data in the .csv file
    CSV_INDEX_CONDITION_MESSAGE_2K_REGULAR,             //!< Zero-based index of a condition's "Message 2K regular" data in the .csv file
    CSV_INDEX_CONDITION_MESSAGE_2K_RECOVERY             //!< Zero-based index of a condition's "Message 2K recovery" data in the .csv file
};

const int ACTOR_MAX_LEVEL = 99;                 //!< The maximum level of an actor
const int CLASS_MAX_LEVEL = ACTOR_MAX_LEVEL;    //!< The maximum level of a class
const int CLASS_BATTLE_COMMANDS = 6;            //!< The number of battle commands of a class (DynRPG has an array of size 7, but changing the last value seems to have no effect, and the RM2K3 editor shows only 6)

const int NUM_SKILL_TYPE = 256;                 //!< The number of Skill Types, including custom ones
const int NUM_ANIM_BATTLER_POSE = 32;           //!< The number of Animation Battler Poses, including custom ones

const char CSV_CELL_SEPARATOR = '\t';                   //!< Character which signifies separation between cells in a CSV file
const char CSV_LINE_SEPARATOR = '\n';                   //!< Character which signifies separation between lines in a CSV file

std::string CSV_FILE_PATH = "DynDatabaseOverrideFiles\\";                               //!< File path for input and output CSV files
std::string STARTING_DATABASE_FILE_NAME_BASE = "Starting";                              //!< Default file name base for starting database files (can be overridden by DynRPG.ini configuration)
std::string DEFAULT_DATABASE_FILE_NAME_BASE = "Default";                                //!< File name base for default database files (ones which record the database immediately after the starting database files are applied)
std::string SAVE_STATE_FILE_NAME_BASE = "SaveState";                                    //!< File name base for save state database files (ones which record the differences from the default database for a save state)
std::string ACTOR_DATABASE_FILE_NAME_END = "_Actors.txt";                               //!< File name end for actor database files
std::string ACTOR_HP_DATABASE_FILE_NAME_END = "_Actors_HP.txt";                         //!< File name end for actor HP database files
std::string ACTOR_MP_DATABASE_FILE_NAME_END = "_Actors_MP.txt";                         //!< File name end for actor MP database files
std::string ACTOR_ATTACK_DATABASE_FILE_NAME_END = "_Actors_Attack.txt";                 //!< File name end for actor attack database files
std::string ACTOR_DEFENSE_DATABASE_FILE_NAME_END = "_Actors_Defense.txt";               //!< File name end for actor defense database files
std::string ACTOR_INTELLIGENCE_DATABASE_FILE_NAME_END = "_Actors_Intelligence.txt";     //!< File name end for actor intelligence database files
std::string ACTOR_AGILITY_DATABASE_FILE_NAME_END = "_Actors_Agility.txt";               //!< File name end for actor agility database files
std::string ACTOR_SKILL_LEARNING_DATABASE_FILE_NAME_END = "_Actors_Skill_Learning.txt"; //!< File name end for actor skill learning database files
std::string ACTOR_CONDITIONS_DATABASE_FILE_NAME_END = "_Actors_Conditions.txt";         //!< File name end for actor conditions database files
std::string ACTOR_ATTRIBUTES_DATABASE_FILE_NAME_END = "_Actors_Attributes.txt";         //!< File name end for actor attributes database files
std::string CLASS_DATABASE_FILE_NAME_END = "_Classes.txt";                              //!< File name end for class database files
std::string CLASS_HP_DATABASE_FILE_NAME_END = "_Classes_HP.txt";                        //!< File name end for class HP database files
std::string CLASS_MP_DATABASE_FILE_NAME_END = "_Classes_MP.txt";                        //!< File name end for class MP database files
std::string CLASS_ATTACK_DATABASE_FILE_NAME_END = "_Classes_Attack.txt";                //!< File name end for class attack database files
std::string CLASS_DEFENSE_DATABASE_FILE_NAME_END = "_Classes_Defense.txt";              //!< File name end for class defense database files
std::string CLASS_INTELLIGENCE_DATABASE_FILE_NAME_END = "_Classes_Intelligence.txt";    //!< File name end for class intelligence database files
std::string CLASS_AGILITY_DATABASE_FILE_NAME_END = "_Classes_Agility.txt";              //!< File name end for class agility database files
std::string CLASS_SKILL_LEARNING_DATABASE_FILE_NAME_END = "_Classes_Skill_Learning.txt";    //!< File name end for class skill learning database files
std::string CLASS_CONDITIONS_DATABASE_FILE_NAME_END = "_Classes_Conditions.txt";        //!< File name end for class conditions database files
std::string CLASS_ATTRIBUTES_DATABASE_FILE_NAME_END = "_Classes_Attributes.txt";        //!< File name end for class attributes database files
std::string SKILL_DATABASE_FILE_NAME_END = "_Skills.txt";                               //!< File name end for skill database files
std::string SKILL_CONDITIONS_DATABASE_FILE_NAME_END = "_Skills_Conditions.txt";         //!< File name end for skill condition database files
std::string SKILL_ATTRIBUTES_DATABASE_FILE_NAME_END = "_Skills_Attributes.txt";         //!< File name end for skill attribute database files
std::string SKILL_ANIMATIONS_DATABASE_FILE_NAME_END = "_Skills_Animations.txt";         //!< File name end for skill animation database files
std::string ITEM_DATABASE_FILE_NAME_END = "_Items.txt";                                 //!< File name end for item database files
std::string ITEM_USABLE_BY_ACTORS_DATABASE_FILE_NAME_END = "_Items_Usable_By_Actors.txt";   //!< File name end for item usable by actors database files
std::string ITEM_USABLE_BY_CLASSES_DATABASE_FILE_NAME_END = "_Items_Usable_By_Classes.txt"; //!< File name end for item usable by classes database files
std::string ITEM_CONDITIONS_DATABASE_FILE_NAME_END = "_Items_Conditions.txt";           //!< File name end for item conditions database files
std::string ITEM_ATTRIBUTES_DATABASE_FILE_NAME_END = "_Items_Attributes.txt";           //!< File name end for item attributes database files
std::string ITEM_WEAPON_ANIMATIONS_DATABASE_FILE_NAME_END = "_Items_Weapon_Animations.txt"; //!< File name end for item weapon animations database files
std::string ENEMY_DATABASE_FILE_NAME_END = "_Enemies.txt";                              //!< File name end for enemy database files
std::string ENEMY_CONDITIONS_DATABASE_FILE_NAME_END = "_Enemies_Conditions.txt";        //!< File name end for enemy conditions database files
std::string ENEMY_ATTRIBUTES_DATABASE_FILE_NAME_END = "_Enemies_Attributes.txt";        //!< File name end for enemy attributes database files
std::string ENEMY_BEHAVIORS_DATABASE_FILE_NAME_END = "_Enemies_Behaviors.txt";          //!< File name end for enemy behaviors database files
std::string ATTRIBUTE_DATABASE_FILE_NAME_END = "_Attributes.txt";                              //!< File name end for attribute database files
std::string CONDITION_DATABASE_FILE_NAME_END = "_Conditions.txt";                              //!< File name end for condition database files

const int CSV_ROWS_ACTOR = 30;                          //!< Number of rows in an Actor CSV file
std::string CSV_HEADER_ACTOR[] =
//! Array of header strings for an Actor CSV file
{
    "Notes (optional)",
    "ID",
    "Name",
    "Degree",
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    "Charset file name",
//    "Charset ID",
//    "Charset Transparent",
    "Min level",
    "Max level",
    "Allow critical hit",
    "Critical hit probability",
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    "Faceset file name",
//    "Faceset ID",
    "Two weapons",
    "AI control",
    "Lock equipment",
    "Mighty guard",
    "Exp primary",
    "Exp secondary",
    "Exp tertiary",
    "Weapon ID",
    "Shield ID",
    "Armor ID",
    "Helmet ID",
    "Accessory ID",
    "Unarmed battle animation ID",
    "Class ID",
    "Battle graphic ID",
    "X pos in battle",
    "Y pos in battle"
};

const int CSV_ROWS_ACTOR_STAT = 2 + ACTOR_MAX_LEVEL;            //!< Number of rows in an actor stat (HP/MP/Attack/Defense/Intelligence/Agility) CSV file
//! Array of header strings for an actor stat (HP/MP/Attack/Defense/Intelligence/Agility) CSV file
std::string CSV_HEADER_ACTOR_STAT[] =
{
    "Notes (optional)",
    "ID",
    "Level "        // This is combined with a level number to form a full header
};

//! Array of header strings for an actor skill learning CSV file
std::string CSV_HEADER_ACTOR_SKILL_LEARNING[] =
{
    "Notes (optional)",
    "Actor ID",
    "Skill learning ID",
    "Level requirement",
    "Skill ID"
};

//! Array of header strings for an actor condition CSV file
std::string CSV_HEADER_ACTOR_CONDITION[] =
{
    "Notes (optional)",
    "Actor ID",
    "Condition ID",
    "Susceptibility rank"
};

//! Array of header strings for an actor attribute CSV file
std::string CSV_HEADER_ACTOR_ATTRIBUTE[] =
{
    "Notes (optional)",
    "Actor ID",
    "Attribute ID",
    "Susceptibility rank"
};

const int CSV_ROWS_CLASS = 18;                          //!< Number of rows in a Class CSV file
std::string CSV_HEADER_CLASS[] =
//! Array of header strings for a Class CSV file
{
    "Notes (optional)",
    "ID",
    "Name",
    "Two weapons",
    "AI control",
    "Lock equipment",
    "Mighty guard",
    "Exp primary",
    "Exp secondary",
    "Exp tertiary",
    "Battle graphic ID",
    "Battle command "       // This is combined with a battle command number to form a full header
};

const int CSV_ROWS_CLASS_STAT = 2 + CLASS_MAX_LEVEL;            //!< Number of rows in an class stat (HP/MP/Attack/Defense/Intelligence/Agility) CSV file
//! Array of header strings for a class stat (HP/MP/Attack/Defense/Intelligence/Agility) CSV file
std::string CSV_HEADER_CLASS_STAT[] =
{
    "Notes (optional)",
    "ID",
    "Level "        // This is combined with a level number to form a full header
};

//! Array of header strings for a class skill learning CSV file
std::string CSV_HEADER_CLASS_SKILL_LEARNING[] =
{
    "Notes (optional)",
    "Class ID",
    "Skill learning ID",
    "Level requirement",
    "Skill ID"
};

//! Array of header strings for a class condition CSV file
std::string CSV_HEADER_CLASS_CONDITION[] =
{
    "Notes (optional)",
    "Class ID",
    "Condition ID",
    "Susceptibility rank"
};

//! Array of header strings for a class attribute CSV file
std::string CSV_HEADER_CLASS_ATTRIBUTE[] =
{
    "Notes (optional)",
    "Class ID",
    "Attribute ID",
    "Susceptibility rank"
};

const int CSV_ROWS_SKILL = 31;                          //!< Number of rows in a Skill CSV file
std::string CSV_HEADER_SKILL[] =
//! Array of header strings for a Skill CSV file
{
    "Notes (optional)",
    "ID",
    "Name",
    "Description",
    "Use message 2K",
    "Use message 2K alt",
    "Evasion type",
    "Skill type",
    "MP consumed",
    "Is MP percent",
    "MP consumed percent",
    "Target",
    "Switch ID",
    "Battle animation ID",
    "Use in field",
    "Use in battle",
    "Attack influence",
    "Intelligence influence",
    "Variance",
    "Effect rating",
    "Success rate",
    "Affects HP",
    "Affects MP",
    "Affects attack",
    "Affects defense",
    "Affects intelligence",
    "Affects agility",
    "Absorb",
    "Ignore defense",
    "Attribute resist",
    "Inflict/remove condition (0 = inflict, 1 = remove)"
};

//! Array of header strings for a skill condition CSV file
std::string CSV_HEADER_SKILL_CONDITION[] =
{
    "Notes (optional)",
    "Skill ID",
    "Condition ID",
    "Affects condition"
};

//! Array of header strings for a skill attribute CSV file
std::string CSV_HEADER_SKILL_ATTRIBUTE[] =
{
    "Notes (optional)",
    "Skill ID",
    "Attribute ID",
    "Has attribute"
};

//! Array of header strings for a skill animation CSV file
std::string CSV_HEADER_SKILL_ANIMATION[] =
{
    "Notes (optional)",
    "Skill ID",
    "Actor ID",
    "Battler pose",
    "Movement",
    "Add after image"
};

const int CSV_ROWS_ITEM = 48;                          //!< Number of rows in an Item CSV file
std::string CSV_HEADER_ITEM[] =
//! Array of header strings for an Item CSV file
{
    "Notes (optional)",
    "ID",
    "Name",
    "Description",
    "Type",
    "Price",
    "Number of uses",
    "Attack",
    "Defense",
    "Intelligence",
    "Agility",
    "Two-handed",
    "MP consumed",
    "Hit probability",
    "Critical bonus",
    "Battle animation ID",
    "Always attack first",
    "Attack twice",
    "Attack all",
    "Ignore enemy evasion",
    "Prevent critical hits",
    "Increased evasion",
    "Half MP consumption",
    "No terrain damage",
    "Cursed",
    "Target",
    "HP restore percent",
    "HP restore units",
    "MP restore percent",
    "MP restore units",
    "Only usable on the field map",
    "Only usable on dead characters",
    "Add max HP",
    "Add max MP",
    "Add attack",
    "Add defense",
    "Add intelligence",
    "Add agility",
    "Message shown when used",
    "Invoke skill when used",
    "Skill ID",
    "Switch ID",
    "Activation allowed in field",
    "Activation allowed in battle",
    "Condition effectiveness",
    "Condition resist or inflict",
    "Weapon trajectory",
    "Weapon targetting"
};

//! Array of header strings for an item usable by actor CSV file
std::string CSV_HEADER_ITEM_USABLE_BY_ACTOR[] =
{
    "Notes (optional)",
    "Item ID",
    "Actor ID",
    "Actor can use"
};

//! Array of header strings for an item usable by class CSV file
std::string CSV_HEADER_ITEM_USABLE_BY_CLASS[] =
{
    "Notes (optional)",
    "Item ID",
    "Class ID",
    "Class can use"
};

//! Array of header strings for an item condition CSV file
std::string CSV_HEADER_ITEM_CONDITION[] =
{
    "Notes (optional)",
    "Item ID",
    "Condition ID",
    "Affects condition"
};

//! Array of header strings for an item attribute CSV file
std::string CSV_HEADER_ITEM_ATTRIBUTE[] =
{
    "Notes (optional)",
    "Item ID",
    "Attribute ID",
    "Affects attribute"
};

//! Array of header strings for an item weapon animation CSV file
std::string CSV_HEADER_ITEM_WEAPON_ANIMATION[] =
{
    "Notes (optional)",
    "Item ID",
    "Actor ID",
    "Battler pose",
    "Animation type (0 = weapon, 1 = battle)",
    "Weapon animation ID",
    "Movement before attack",
    "Add afterimage",
    "Number of attacks",
    "Ranged weapon",
    "Ranged weapon animation ID",
    "Ranged weapon animation speed"
};

const int CSV_ROWS_ENEMY = 20;                          //!< Number of rows in an enemy CSV file
std::string CSV_HEADER_ENEMY[] =
//! Array of header strings for an enemy CSV file
{
    "Notes (optional)",
    "ID",
    "Name",
    "File name",
    "Color alteration",
    "Is transparent",
    "Is flying",
    "Max HP",
    "Max MP",
    "Attack",
    "Defense",
    "Intelligence",
    "Agility",
    "Experience",
    "Money",
    "Item ID",
    "Item probability",
    "Allow critical hit",
    "Critical hit probability",
    "Attacks often miss"
};

//! Array of header strings for an enemy condition CSV file
std::string CSV_HEADER_ENEMY_CONDITION[] =
{
    "Notes (optional)",
    "Enemy ID",
    "Condition ID",
    "Susceptibility rank"
};

//! Array of header strings for an enemy attribute CSV file
std::string CSV_HEADER_ENEMY_ATTRIBUTE[] =
{
    "Notes (optional)",
    "Enemy ID",
    "Attribute ID",
    "Susceptibility rank"
};

//! Array of header strings for an enemy behavior CSV file
std::string CSV_HEADER_ENEMY_BEHAVIOR[] =
{
    "Notes (optional)",
    "Enemy ID",
    "Behavior ID",
    "Precondition",
    "Precondition param 1",
    "Precondition param 2",
    "Precondition switch ID",
    "Priority",
    "Switch on",
    "Switch on ID",
    "Switch off",
    "Switch off ID",
    "Action kind",
    "Basic action type",
    "Skill ID",
    "Transform enemy ID"
};

const int CSV_ROWS_ATTRIBUTE = 9;                          //!< Number of rows in an attribute CSV file
//! Array of header strings for an attribute CSV file
std::string CSV_HEADER_ATTRIBUTE[] =
{
    "Notes (optional)",
    "ID",
    "Name",
    "Type",
    "Damage multiplier A",
    "Damage multiplier B",
    "Damage multiplier C",
    "Damage multiplier D",
    "Damage multiplier E"
};

const int CSV_ROWS_CONDITION = 43;                          //!< Number of rows in a condition CSV file
//! Array of header strings for a condition CSV file
std::string CSV_HEADER_CONDITION[] =
{
    "Notes (optional)",
    "ID",
    "Name",
    "Persists after battle",
    "Color",
    "Priority",
    "Action restriction",
    "Hit percentage",
    "Evade all attacks",
    "Reflect skills",
    "Equipment cannot be changed",
    "Susceptibility A",
    "Susceptibility B",
    "Susceptibility C",
    "Susceptibility D",
    "Susceptibility E",
    "Recovery turns",
    "Recovery turns percent",
    "Recovery physical percent",
    "Base stat alteration",
    "Alters attack",
    "Alters defense",
    "Alters intelligence",
    "Alters agility",
    "Prevent skill by attack",
    "Prevent skill attack influence",
    "Prevent skill by intelligence",
    "Prevent skill intelligence influence",
    "HP alteration type",
    "HP damage percent",
    "HP damage units",
    "HP damage on step",
    "HP damage step number",
    "MP alteration type",
    "MP damage percent",
    "MP damage units",
    "MP damage on step",
    "MP damage step number",
    "Animation ID",
    "Message 2K ally state",
    "Message 2K enemy state",
    "Message 2K already in",
    "Message 2K regular",
    "Message 2K recovery"
};

//! Map of RPG::DamageMultiplier_T values to the associated token strings used in CSV files
std::map<RPG::DamageMultiplier_T, std::string> damageMultiplierToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::DamageMultiplier_T values
std::map<std::string, RPG::DamageMultiplier_T> csvTokenToDamageMultiplier;
//! Map of RPG::ActionKind_T values to the associated token strings used in CSV files
std::map<RPG::ActionKind_T, std::string> actionKindToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::ActionKind_T values
std::map<std::string, RPG::ActionKind_T> csvTokenToActionKind;
//! Map of RPG::BasicAction_T values to the associated token strings used in CSV files
std::map<RPG::BasicAction_T, std::string> basicActionToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::BasicAction_T values
std::map<std::string, RPG::BasicAction_T> csvTokenToBasicAction;
//! Map of index values to the associated token strings used in CSV files (would use RPG::SkillType for index, but custom poses go past the enum)
std::map<int, std::string> skillTypeToCsvToken;
//! Map of token strings used in CSV files to the associated index values (would use RPG::SkillType for value, but custom poses go past the enum)
std::map<std::string, int> csvTokenToSkillType;
//! Map of RPG::SkillTarget values to the associated token strings used in CSV files
std::map<RPG::SkillTarget, std::string> skillTargetToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::SkillTarget values
std::map<std::string, RPG::SkillTarget> csvTokenToSkillTarget;
//! Map of index values to the associated token strings used in CSV files (would use RPG::AnimationBattlerPose for index, but custom poses go past the enum)
std::map<int, std::string> animationBattlerPoseToCsvToken;
//! Map of token strings used in CSV files to the associated index values (would use RPG::AnimationBattlerPose for value, but custom poses go past the enum)
std::map<std::string, int> csvTokenToAnimationBattlerPose;
//! Map of RPG::AnimationMoveBeforeAttack values to the associated token strings used in CSV files
std::map<RPG::AnimationMoveBeforeAttack, std::string> animationMoveBeforeAttackToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::AnimationMoveBeforeAttack values
std::map<std::string, RPG::AnimationMoveBeforeAttack> csvTokenToAnimationMoveBeforeAttack;
//! Map of RPG::WeaponNumberOfAttacks values to the associated token strings used in CSV files
std::map<RPG::WeaponNumberOfAttacks, std::string> weaponNumberOfAttacksToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::WeaponNumberOfAttacks values
std::map<std::string, RPG::WeaponNumberOfAttacks> csvTokenToWeaponNumberOfAttacks;
////! Map of boolean values (for RPG::items::weaponAnims::isBattleAnim) to the associated token strings used in CSV files
//std::map<bool, std::string> itemWeaponAnimationIsBattleAnimationToCsvToken;
////! Map of token strings used in CSV files to the associated boolean values (for RPG::items::weaponAnims::isBattleAnim)
//std::map<std::string, bool> csvTokenToItemWeaponAnimationIsBattleAnimation;
//! Map of RPG::WeaponRangedAnimSpeed values to the associated token strings used in CSV files
std::map<RPG::WeaponRangedAnimSpeed, std::string> weaponRangedAnimSpeedToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::WeaponRangedAnimSpeed values
std::map<std::string, RPG::WeaponRangedAnimSpeed> csvTokenToWeaponRangedAnimSpeed;
//! Map of RPG::WeaponRangedTrajectory values to the associated token strings used in CSV files
std::map<RPG::WeaponRangedTrajectory, std::string> weaponRangedTrajectoryToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::WeaponRangedTrajectory values
std::map<std::string, RPG::WeaponRangedTrajectory> csvTokenToWeaponRangedTrajectory;
//! Map of RPG::WeaponRangedTargetting values to the associated token strings used in CSV files
std::map<RPG::WeaponRangedTargetting, std::string> weaponRangedTargettingToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::WeaponRangedTargetting values
std::map<std::string, RPG::WeaponRangedTargetting> csvTokenToWeaponRangedTargetting;
//! Map of RPG::ItemType values to the associated token strings used in CSV files
std::map<RPG::ItemType, std::string> itemTypeToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::ItemType values
std::map<std::string, RPG::ItemType> csvTokenToItemType;
//! Map of RPG::ItemTarget values to the associated token strings used in CSV files
std::map<RPG::ItemTarget, std::string> itemTargetToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::ItemTarget values
std::map<std::string, RPG::ItemTarget> csvTokenToItemTarget;
//! Map of RPG::ItemScrollMsgShown values to the associated token strings used in CSV files
std::map<RPG::ItemScrollMsgShown, std::string> itemScrollMsgShownToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::ItemScrollMsgShown values
std::map<std::string, RPG::ItemScrollMsgShown> csvTokenToItemScrollMsgShown;
//! Map of RPG::ItemConditionFlag values to the associated token strings used in CSV files
std::map<RPG::ItemConditionFlag, std::string> itemConditionFlagToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::ItemConditionFlag values
std::map<std::string, RPG::ItemConditionFlag> csvTokenToItemConditionFlag;
//! Map of RPG::MonsterPrecondition_T values to the associated token strings used in CSV files
std::map<RPG::MonsterPrecondition_T, std::string> monsterPreconditionToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::MonsterPrecondition_T values
std::map<std::string, RPG::MonsterPrecondition_T> csvTokenToMonsterPrecondition;
//! Map of RPG::AttributeType_T values to the associated token strings used in CSV files
std::map<RPG::AttributeType_T, std::string> attributeTypeToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::AttributeType_T values
std::map<std::string, RPG::AttributeType_T> csvTokenToAttributeType;
//! Map of RPG::ActionRestriction_T values to the associated token strings used in CSV files
std::map<RPG::ActionRestriction_T, std::string> actionRestrictionToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::ActionRestriction_T values
std::map<std::string, RPG::ActionRestriction_T> csvTokenToActionRestriction;
//! Map of RPG::BaseStatAlteration values to the associated token strings used in CSV files
std::map<RPG::BaseStatAlteration, std::string> baseStatAlterationToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::BaseStatAlteration values
std::map<std::string, RPG::BaseStatAlteration> csvTokenToBaseStatAlteration;
//! Map of RPG::AlterationType values to the associated token strings used in CSV files
std::map<RPG::AlterationType, std::string> alterationTypeToCsvToken;
//! Map of token strings used in CSV files to the associated RPG::AlterationType values
std::map<std::string, RPG::AlterationType> csvTokenToAlterationType;

std::string startingDatabaseFileNameBase = STARTING_DATABASE_FILE_NAME_BASE;    //!< File name base for starting database files
std::string saveStateFileNameBase = SAVE_STATE_FILE_NAME_BASE;  //!< File name base for save state files
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor CSV file
bool overriddenDataActor[CSV_ROWS_ACTOR - 2];                           // Disinclude "Notes (optional)" and "ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor HP CSV file
bool overriddenDataActorHp[CSV_ROWS_ACTOR_STAT - 2];                    // Disinclude "Notes (optional)" and "Actor ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor MP CSV file
bool overriddenDataActorMp[CSV_ROWS_ACTOR_STAT - 2];                    // Disinclude "Notes (optional)" and "Actor ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor attack CSV file
bool overriddenDataActorAttack[CSV_ROWS_ACTOR_STAT - 2];                // Disinclude "Notes (optional)" and "Actor ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor defense CSV file
bool overriddenDataActorDefense[CSV_ROWS_ACTOR_STAT - 2];               // Disinclude "Notes (optional)" and "Actor ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor intelligence CSV file
bool overriddenDataActorIntelligence[CSV_ROWS_ACTOR_STAT - 2];          // Disinclude "Notes (optional)" and "Actor ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor agility CSV file
bool overriddenDataActorAgility[CSV_ROWS_ACTOR_STAT - 2];               // Disinclude "Notes (optional)" and "Actor ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor skill learning CSV file
std::set< std::pair<int, int> > overriddenDataActorSkillLearning;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor conditions CSV file
std::set< std::pair<int, int> > overriddenDataActorConditions;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an actor attributes CSV file
std::set< std::pair<int, int> > overriddenDataActorAttributes;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class CSV file
bool overriddenDataClass[CSV_ROWS_CLASS - 2];                           // Disinclude "Notes (optional)" and "ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class HP CSV file
bool overriddenDataClassHp[CSV_ROWS_CLASS_STAT - 2];                    // Disinclude "Notes (optional)" and "Class ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class MP CSV file
bool overriddenDataClassMp[CSV_ROWS_CLASS_STAT - 2];                    // Disinclude "Notes (optional)" and "Class ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class attack CSV file
bool overriddenDataClassAttack[CSV_ROWS_CLASS_STAT - 2];                // Disinclude "Notes (optional)" and "Class ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class defense CSV file
bool overriddenDataClassDefense[CSV_ROWS_CLASS_STAT - 2];               // Disinclude "Notes (optional)" and "Class ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class intelligence CSV file
bool overriddenDataClassIntelligence[CSV_ROWS_CLASS_STAT - 2];          // Disinclude "Notes (optional)" and "Class ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class agility CSV file
bool overriddenDataClassAgility[CSV_ROWS_CLASS_STAT - 2];               // Disinclude "Notes (optional)" and "Class ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class skill learning CSV file
std::set< std::pair<int, int> > overriddenDataClassSkillLearning;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class conditions CSV file
std::set< std::pair<int, int> > overriddenDataClassConditions;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a class attributes CSV file
std::set< std::pair<int, int> > overriddenDataClassAttributes;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a skill CSV file
bool overriddenDataSkill[CSV_ROWS_SKILL - 2];                           // Disinclude "Notes (optional)" and "ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a skill conditions CSV file
std::set< std::pair<int, int> > overriddenDataSkillConditions;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a skill attributes CSV file
std::set< std::pair<int, int> > overriddenDataSkillAttributes;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a skill animations CSV file
std::set< std::pair<int, int> > overriddenDataSkillAnimations;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an item CSV file
bool overriddenDataItem[CSV_ROWS_ITEM - 2];                             // Disinclude "Notes (optional)" and "ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an item usable by actor CSV file
std::set< std::pair<int, int> > overriddenDataItemUsableByActors;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an item usable by class CSV file
std::set< std::pair<int, int> > overriddenDataItemUsableByClasses;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an item conditions CSV file
std::set< std::pair<int, int> > overriddenDataItemConditions;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an item attributes CSV file
std::set< std::pair<int, int> > overriddenDataItemAttributes;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an item weapon animations CSV file
std::set< std::pair<int, int> > overriddenDataItemWeaponAnimations;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an enemy CSV file
bool overriddenDataEnemy[CSV_ROWS_ENEMY - 2];                           // Disinclude "Notes (optional)" and "Enemy ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an enemy conditions CSV file
std::set< std::pair<int, int> > overriddenDataEnemyConditions;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an enemy attributes CSV file
std::set< std::pair<int, int> > overriddenDataEnemyAttributes;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an enemy behaviors CSV file
std::set< std::pair<int, int> > overriddenDataEnemyBehaviors;
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with an attributes CSV file
bool overriddenDataAttributes[CSV_ROWS_ATTRIBUTE - 2];                  // Disinclude "Notes (optional)" and "ID" rows
//! Flags indicating which data have been overridden by the plugin (NOT including overrides from starting file list) with a conditions CSV file
bool overriddenDataConditions[CSV_ROWS_CONDITION - 2];                  // Disinclude "Notes (optional)" and "ID" rows

#ifdef DYNDATABASEOVERRIDE_DEBUG
std::ofstream * ofDebug;                                //!< File output stream to record debug data
#endif // DYNDATABASEOVERRIDE_DEBUG

//! Sanitize a value for use in a CSV file
/*!
    sanitizeForCsv takes a std::string and converts it to a form that is appropriate for storage in
    a CSV file. This is necessary because an RPG::DStringPtr representing an empty string will cause
    issues if written directly to the file stream (and if other problem cases are found, they can be
    easily added to this function).

    \param r_Value: (std::string) The value to sanitize
    \return (std::string) The sanitized value
*/
std::string sanitizeForCsv( std::string r_Value ) {
    if( 0 == r_Value.length() ) {
        // Empty string
        return "";
    }
    // If no special cases are found, return the value as-is
    return r_Value;
}

//! Sanitizes a value for use in a CSV file
/*!
    sanitizeForCsv takes an int and converts it to a form that is appropriate for storage in a CSV
    file. This is necessary more for strings than anything else, but having an overloaded method for
    each potential data type allows the values to be fed into a "one-size-fits-all" process.

    \param r_Value: (int) The value to sanitize
    \return (int) The sanitized value
*/
int sanitizeForCsv( int r_Value ) {
    // int values work fine being written directly to file stream; pass it back as-is
    return r_Value;
}

//! Synchronizes the active game data to the database for a new game
/*!
    syncActiveToDatabaseForNewGame updates the active game data to match what is in the database
    in the context of starting a new game. For example, active actor information -- HP, MP, skills
    learned, etc. -- needs to be synced with the database when a new game is started.
*/
void syncActiveToDatabaseForNewGame() {
    // Sync all active data to database values
    // Actors
    for( int actorId = 1; actorId <= RPG::actors.count(); actorId++ ) {
        RPG::actors[actorId]->level = RPG::dbActors[actorId]->minLevel;
        int curLevel = RPG::actors[actorId]->level;
        RPG::actors[actorId]->hp = RPG::dbActors[actorId]->maxHp[curLevel];
        RPG::actors[actorId]->mp = RPG::dbActors[actorId]->maxMp[curLevel];
        RPG::actors[actorId]->weaponId = RPG::dbActors[actorId]->weaponId;
        RPG::actors[actorId]->shieldId = RPG::dbActors[actorId]->shieldId;
        RPG::actors[actorId]->armorId = RPG::dbActors[actorId]->armorId;
        RPG::actors[actorId]->helmetId = RPG::dbActors[actorId]->helmetId;
        RPG::actors[actorId]->accessoryId = RPG::dbActors[actorId]->accessoryId;
        for( int skillId = 1; skillId <= RPG::skills.count(); skillId++ ) {
            bool skillShouldBeKnown = false;
            bool skillIsKnown = RPG::actors[actorId]->isSkillKnown( skillId );
            for( int skillProgressionId = 0; skillProgressionId < RPG::dbActors[actorId]->skillLearning.count(); skillProgressionId++ ) {
                if( skillId == RPG::dbActors[actorId]->skillLearning[skillProgressionId]->skillId
                    && curLevel >= RPG::dbActors[actorId]->skillLearning[skillProgressionId]->levelReq ) {
                    // Skill should be known by actor at their current level
                    skillShouldBeKnown = true;
                }
            }
            if( skillShouldBeKnown && !skillIsKnown ) {
                // Skill should be known, but is not; add it
                RPG::actors[actorId]->addSkill( skillId );
            }
            if( !skillShouldBeKnown && skillIsKnown ) {
                // Skill should not be known, but is; remove it
                RPG::actors[actorId]->removeSkill( skillId );
            }
        }
    }
}

// Export macros
/*
    SETUP_EXPORT_FUNCTION_SIMPLE is a macro for setting up an export function in context of a simple
    data value. For example, in the exportActors function, each line of the CSV file holds fairly
    simple values -- int, bool, or string -- while more complicated data structures nested within
    the DBActor class are handled by other functions.

    It is expected that the following variables will be defined and initialized in the function
    before this macro is used (the values used here are from exportActors and are just examples):

    std::string fileName = r_FileNameBase + ACTOR_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActor;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR;   // Pointer to the array of header strings to use

    ***** Human-readable representation *****

    std::ofstream outputStream;    // Output file stream pointer
    int rowIndex, primaryKeyIndex, overriddenDataIndex;     // Index variables

    // If the overridden data structure is empty and r_ExportAll is false, skip doing anything
    if( !r_ExportAll ) {
        bool overriddenDataFound = false;
        for( int i = 0; !overriddenDataFound && i < overriddenDataStructureSize; i++ ) {
            overriddenDataFound = overriddenDataStructure[i];
        }
        if( !overriddenDataFound ) {
            // No overridden data found
            return;
        }
    }
    // Open output file
    outputStream.open( fileName.c_str() );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Notes (optional)
    rowIndex = 0;
    // Output row header
    outputStream << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR;
    // Output notes
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
        outputStream << r_NoteValue << CSV_CELL_SEPARATOR;
    }
    // ID
    // ID is always required, so do not check whether it is desired
    rowIndex++;
    // Output row header
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR;
    // Output record values
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
        outputStream << r_IdValue << CSV_CELL_SEPARATOR;
    }
    // To offset the fact that WRITE_CSV_ROW_* macros start by incrementing overriddenDataIndex,
    // initialize it to -1
    overriddenDataIndex = -1;
*/
#define SETUP_EXPORT_FUNCTION_SIMPLE(r_NoteValue, r_IdValue) \
    std::ofstream outputStream; \
    int rowIndex, primaryKeyIndex, overriddenDataIndex; \
    if( !r_ExportAll ) { \
        bool overriddenDataFound = false; \
        for( int i = 0; !overriddenDataFound && i < overriddenDataStructureSize; i++ ) { \
            overriddenDataFound = overriddenDataStructure[i]; \
        } \
        if( !overriddenDataFound ) { \
            return; \
        } \
    } \
    outputStream.open( fileName.c_str() ); \
    rowIndex = 0; \
    outputStream << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR; \
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
        outputStream << r_NoteValue << CSV_CELL_SEPARATOR; \
    } \
    rowIndex++; \
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR; \
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
        outputStream << r_IdValue << CSV_CELL_SEPARATOR; \
    } \
    overriddenDataIndex = -1;

/*
    SETUP_EXPORT_FUNCTION_DUAL_KEY is a macro for setting up an export function in context of a
    dual-key data structure. For example, in the exportActorsSkillLearning function, we are dealing
    with individual SkillProgression data structures within each DBActor, and the SkillProgression
    data structures have varying lengths. Hence we need to accommodate both the DBActor (primary
    key) and SkillProgression (secondary key) data structures.

    It is expected that the following variables will be defined and initialized in the function
    before this macro is used (the values used here are from exportActorsSkillLearning and are just
    examples):

    std::string fileName = r_FileNameBase + ACTOR_SKILL_LEARNING_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 0;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorSkillLearning;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_SKILL_LEARNING;   // Pointer to the array of header strings to use

    ***** Human-readable representation *****

    std::ofstream outputStream;     // Output file stream pointer
    int rowIndex, primaryKeyIndex, secondaryKeyIndex;      // Index variables
    std::pair<int, int> overriddenDataIndex;     // Index for use in accessing overridden data structure

    // If the overridden data structure is empty and r_ExportAll is false, skip doing anything
    if( !r_ExportAll && overriddenDataStructure->empty() ) {
        return;
    }
    // Open output file
    outputStream.open( fileName.c_str() );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Notes (optional)
    rowIndex = 0;
    // Output row header
    outputStream << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR;
    // Output notes
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
        for( secondaryKeyIndex = secondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) {
            overriddenDataIndex.first = primaryKeyIndex;
            overriddenDataIndex.second = secondaryKeyIndex;
            outputStream << r_NoteValue << CSV_CELL_SEPARATOR;
        }
    }
    // Primary data structure ID
    rowIndex++;
    // Output row header
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR;
    // Output record values
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
        // The primary key index must be repeated for each element in the secondary data structure
        for( secondaryKeyIndex = secondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) {
            outputStream << primaryKeyIndex << CSV_CELL_SEPARATOR;
        }
    }
    // Secondary data structure ID
    rowIndex++;
    // Output row header
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR;
    // Output record values
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
        for( secondaryKeyIndex = secondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) {
            outputStream << secondaryKeyIndex << CSV_CELL_SEPARATOR;
        }
    }
*/
#define SETUP_EXPORT_FUNCTION_DUAL_KEY(r_SecondaryKeyLimit, r_NoteValue ) \
    std::ofstream outputStream; \
    int rowIndex, primaryKeyIndex, secondaryKeyIndex; \
    std::pair<int, int> overriddenDataIndex; \
    if( !r_ExportAll && overriddenDataStructure->empty() ) { \
        return; \
    } \
    outputStream.open( fileName.c_str() ); \
    rowIndex = 0; \
    outputStream << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR; \
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
        for( secondaryKeyIndex = secondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) { \
            overriddenDataIndex.first = primaryKeyIndex; \
            overriddenDataIndex.second = secondaryKeyIndex; \
            outputStream << r_NoteValue << CSV_CELL_SEPARATOR; \
        } \
    } \
    rowIndex++; \
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR; \
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
        for( secondaryKeyIndex = secondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) { \
            outputStream << primaryKeyIndex << CSV_CELL_SEPARATOR; \
        } \
    } \
    rowIndex++; \
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR; \
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
        for( secondaryKeyIndex = secondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) { \
            outputStream << secondaryKeyIndex << CSV_CELL_SEPARATOR; \
        } \
    }

/*
    WRITE_CSV_ROW_SIMPLE is a macro for writing a line to a CSV file in context of a simple data
    value. For example, in the exportActors function, each line of the CSV file holds fairly simple
    values -- int, bool, or string -- while more complicated data structures nested within the
    DBActor class are handled by other functions.

    It is expected that the SETUP_EXPORT_FUNCTION_SIMPLE macro will be used in the function before
    this macro is used.

    ***** Human-readable representation *****

    rowIndex++;
    overriddenDataIndex++;
    if( r_ExportAll || overriddenDataStructure[overriddenDataIndex] ) {
        // Output row header
        outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR;
        // Output record values
        for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
            outputStream << sanitizeForCsv( r_CellValue ) << CSV_CELL_SEPARATOR;
        }
    }
*/
#define WRITE_CSV_ROW_SIMPLE(r_CellValue) \
    rowIndex++; \
    overriddenDataIndex++; \
    if( r_ExportAll || overriddenDataStructure[overriddenDataIndex] ) { \
        outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR; \
        for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
            outputStream << sanitizeForCsv( r_CellValue ) << CSV_CELL_SEPARATOR; \
        } \
    }

/*
    WRITE_CSV_ROWS_ARRAY is a macro for writing multiple lines to a CSV file in context of an array
    of simple data values. For example, in the exportActorsHp function, each line of the CSV file
    holds int values associated with a given Actor level across all Actors.

    It is expected that the SETUP_EXPORT_FUNCTION_SIMPLE macro will be used in the function before
    this macro is used. Also, the extra variable int secondaryKeyIndex must be defined.

    ***** Human-readable representation *****

    overriddenDataIndex = 0;
    for( secondaryKeyIndex = r_SecondaryKeyBase; secondaryKeyIndex <= r_RowLimit; secondaryKeyIndex++ ) {
        if( r_ExportAll || overriddenDataStructure[overriddenDataIndex] ) {
            // Output row header
            outputStream << CSV_LINE_SEPARATOR << r_RowHeader << CSV_CELL_SEPARATOR;
            // Output record values
            for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
                outputStream << sanitizeForCsv( r_CellValue ) << CSV_CELL_SEPARATOR;
            }
        }
        overriddenDataIndex++;
    }
    // Add a blank line at the end to ensure the last line of data does not get skipped when importing
    outputStream << CSV_LINE_SEPARATOR;

*/
#define WRITE_CSV_ROWS_ARRAY(r_CellValue, r_RowHeader, r_SecondaryKeyBase, r_SecondaryKeyLimit) \
    overriddenDataIndex = 0; \
    for( secondaryKeyIndex = r_SecondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) { \
        if( r_ExportAll || overriddenDataStructure[overriddenDataIndex] ) { \
            outputStream << CSV_LINE_SEPARATOR << r_RowHeader << CSV_CELL_SEPARATOR; \
            for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
                outputStream << sanitizeForCsv( r_CellValue ) << CSV_CELL_SEPARATOR; \
            } \
        } \
        overriddenDataIndex++; \
    } \
    outputStream << CSV_LINE_SEPARATOR;

/*
    WRITE_CSV_ROW_DUAL_KEY is a macro for writing a line to a CSV file in context of a dual-key data
    structure. For example, in the exportActorsSkillLearning function, we are dealing with
    individual SkillProgression data structures within each DBActor, and the SkillProgression data
    structures have varying lengths. Hence we need to accommodate both the DBActor (primary key) and
    SkillProgression (secondary key) data structures.

    It is expected that the SETUP_EXPORT_FUNCTION_DUAL_KEY macro will be used in the function before
    this macro is used.

    ***** Human-readable representation *****

    rowIndex++;
    // Output row header
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR;
    // Output record values
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) {
        for( secondaryKeyIndex = secondaryKeybase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) {
            overriddenDataIndex.first = primaryKeyIndex;
            overriddenDataIndex.second = secondaryKeyIndex;
            if( r_ExportAll || overriddenDataStructure[overriddenDataIndex] ) {
                outputStream << sanitizeForCsv( r_CellValue ) << CSV_CELL_SEPARATOR;
            }
        }
    }
*/
#define WRITE_CSV_ROW_DUAL_KEY(r_CellValue, r_SecondaryKeyLimit) \
    rowIndex++; \
    outputStream << CSV_LINE_SEPARATOR << csvHeaderArray[rowIndex] << CSV_CELL_SEPARATOR; \
    for( primaryKeyIndex = primaryKeyBase; primaryKeyIndex <= primaryKeyLimit; primaryKeyIndex++ ) { \
        for( secondaryKeyIndex = secondaryKeyBase; secondaryKeyIndex <= r_SecondaryKeyLimit; secondaryKeyIndex++ ) { \
            overriddenDataIndex.first = primaryKeyIndex; \
            overriddenDataIndex.second = secondaryKeyIndex; \
            if( r_ExportAll || overriddenDataStructure->count( overriddenDataIndex ) ) { \
                outputStream << sanitizeForCsv( r_CellValue ) << CSV_CELL_SEPARATOR; \
            } \
        } \
    }

//! Export the current Actor database values to a .csv file
/*!
    exportActors exports the current Actor database values (not including array values like HP, MP,
    etc.) to a .csv file, using a given file name base for the file name. This function is used as a
    subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActors( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActors( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActor;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbActors[primaryKeyIndex]->name, RPG::dbActors[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Name
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->name );
    // Degree
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->degree );
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    // Charset file name
//    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->charsetFilename );
//    // Charset ID
//    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->charsetId );
//    // Charset transparent
//    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->charsetTransparent );
    // Min level
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->minLevel );
    // Max level
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->maxLevel );
    // Allow critical hit
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->allowCriticalHit );
    // Critical hit probability
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->criticalHitProbability );
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    // Faceset file name
//    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->facesetFilename );
//    // Faceset ID
//    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->facesetId );
    // Two weapons
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->twoWeapons );
    // AI control
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->aiControl );
    // Lock equipment
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->lockEquipment );
    // Mighty guard
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->mightyGuard );
    // Exp primary
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->expPrimary );
    // Exp secondary
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->expSecondary );
    // Exp tertiary
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->expTertiary );
    // Weapon ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->weaponId );
    // Shield ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->shieldId );
    // Armor ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->armorId );
    // Helmet ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->helmetId );
    // Accessory ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->accessoryId );
    // Unarmed battle animation ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->unarmedBattleAnimationId );
    // Class ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->classId );
    // Battle graphic ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->battleGraphicId );
    // X pos in battle
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->xPosInBattle );
    // Y pos in battle
    WRITE_CSV_ROW_SIMPLE( RPG::dbActors[primaryKeyIndex]->yPosInBattle );
    // Close output file
    outputStream.close();
}

//! Export the current Actor HP database values to a .csv file
/*!
    exportActorsHp exports the current Actor HP database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActorsHp( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsHp( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_HP_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorHp;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbActors[primaryKeyIndex]->name, RPG::dbActors[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbActors[primaryKeyIndex]->maxHp[secondaryKeyIndex], CSV_HEADER_ACTOR_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, ACTOR_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Actor MP database values to a .csv file
/*!
    exportActorsMp exports the current Actor MP database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActorsMp( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsMp( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_MP_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorMp;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbActors[primaryKeyIndex]->name, RPG::dbActors[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbActors[primaryKeyIndex]->maxMp[secondaryKeyIndex], CSV_HEADER_ACTOR_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, ACTOR_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Actor Attack database values to a .csv file
/*!
    exportActorsAttack exports the current Actor Attack database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActorsAttack( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsAttack( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_ATTACK_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorAttack;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbActors[primaryKeyIndex]->name, RPG::dbActors[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbActors[primaryKeyIndex]->attack[secondaryKeyIndex], CSV_HEADER_ACTOR_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, ACTOR_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Actor Defense database values to a .csv file
/*!
    exportActorsDefense exports the current Actor Defense database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void exportActorsDefense( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsDefense( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_DEFENSE_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorDefense;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbActors[primaryKeyIndex]->name, RPG::dbActors[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbActors[primaryKeyIndex]->defense[secondaryKeyIndex], CSV_HEADER_ACTOR_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, ACTOR_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Actor Intelligence database values to a .csv file
/*!
    exportActorsIntelligence exports the current Actor Intelligence database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActorsIntelligence( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsIntelligence( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_INTELLIGENCE_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorIntelligence;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbActors[primaryKeyIndex]->name, RPG::dbActors[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbActors[primaryKeyIndex]->intelligence[secondaryKeyIndex], CSV_HEADER_ACTOR_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, ACTOR_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Actor Agility database values to a .csv file
/*!
    exportActorsAgility exports the current Actor Agility database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActorsAgility( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsAgility( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_AGILITY_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorAgility;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ACTOR_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbActors[primaryKeyIndex]->name, RPG::dbActors[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbActors[primaryKeyIndex]->agility[secondaryKeyIndex], CSV_HEADER_ACTOR_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, ACTOR_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Actor skill learning database values to a .csv file
/*!
    exportActorsSkillLearning exports the current Actor skill learning database values to a .csv
    file, using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActorsSkillLearning( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsSkillLearning( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_SKILL_LEARNING_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 0;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;        // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorSkillLearning;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_SKILL_LEARNING;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbActors[primaryKeyIndex]->skillLearning.count() - 1,
        RPG::dbActors[primaryKeyIndex]->name << " - " << RPG::skills[RPG::dbActors[primaryKeyIndex]->skillLearning[secondaryKeyIndex]->skillId]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Level requirement
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbActors[primaryKeyIndex]->skillLearning[secondaryKeyIndex]->levelReq, RPG::dbActors[primaryKeyIndex]->skillLearning.count() - 1 + secondaryKeyBase );
    // Skill ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbActors[primaryKeyIndex]->skillLearning[secondaryKeyIndex]->skillId, RPG::dbActors[primaryKeyIndex]->skillLearning.count() - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Actor Condition database values to a .csv file
/*!
    exportActorsConditions exports the current Actor Conditions database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void exportActorsConditions( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsConditions( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_CONDITIONS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_CONDITION;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbActors[primaryKeyIndex]->conditions.size,
        RPG::dbActors[primaryKeyIndex]->name << " - " << RPG::conditions[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Condition susceptibility rank
    WRITE_CSV_ROW_DUAL_KEY( damageMultiplierToCsvToken[RPG::dbActors[primaryKeyIndex]->conditions[secondaryKeyIndex]], RPG::dbActors[primaryKeyIndex]->conditions.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Actor Condition database values to a .csv file
/*!
    exportActorsAttributes exports the current Actor Attributes database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportActorsAttributes( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportActorsAttributes( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ACTOR_ATTRIBUTES_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbActors.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_ATTRIBUTE;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbActors[primaryKeyIndex]->attributes.size,
        RPG::dbActors[primaryKeyIndex]->name << " - " << RPG::attributes[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Attribute susceptibility rank
    WRITE_CSV_ROW_DUAL_KEY( damageMultiplierToCsvToken[RPG::dbActors[primaryKeyIndex]->attributes[secondaryKeyIndex]], RPG::dbActors[primaryKeyIndex]->attributes.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Class database values to a .csv file
/*!
    exportClasses exports the current Class database values (not including array values like HP, MP,
    etc.) to a .csv file, using a given file name base for the file name. This function is used as a
    subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClasses( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClasses( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataClass;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CLASS - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CLASS;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name, RPG::dbClasses[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Name
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name );
    // Two weapons
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->twoWeapons );
    // AI control
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->aiControl );
    // Lock equipment
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->lockEquipment );
    // Mighty guard
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->mightyGuard );
    // Exp primary
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->expPrimary );
    // Exp secondary
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->expSecondary );
    // Exp tertiary
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->expTertiary );
    // Battle graphic ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbClasses[primaryKeyIndex]->battleGraphicId );
    // Battle commands
    WRITE_CSV_ROWS_ARRAY( RPG::dbClasses[primaryKeyIndex]->battleCommands[secondaryKeyIndex], CSV_HEADER_CLASS[CSV_INDEX_CLASS_BATTLE_COMMAND_0] << secondaryKeyIndex, 0, CLASS_BATTLE_COMMANDS - 1 );
    // Close output file
    outputStream.close();
}

//! Export the current Class HP database values to a .csv file
/*!
    exportClassesHp exports the current Class HP database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesHp( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesHp( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_HP_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorHp;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CLASS_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name, RPG::dbClasses[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbClasses[primaryKeyIndex]->maxHp[secondaryKeyIndex], CSV_HEADER_CLASS_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, CLASS_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Class MP database values to a .csv file
/*!
    exportClassesMp exports the current Class MP database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesMp( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesMp( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_MP_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorMp;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CLASS_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name, RPG::dbClasses[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbClasses[primaryKeyIndex]->maxMp[secondaryKeyIndex], CSV_HEADER_CLASS_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, CLASS_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Class Attack database values to a .csv file
/*!
    exportClassesAttack exports the current Class Attack database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesAttack( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesAttack( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_ATTACK_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorAttack;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CLASS_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name, RPG::dbClasses[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbClasses[primaryKeyIndex]->attack[secondaryKeyIndex], CSV_HEADER_CLASS_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, CLASS_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Class Defense database values to a .csv file
/*!
    exportClassesDefense exports the current Class Defense database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesDefense( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesDefense( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_DEFENSE_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorDefense;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CLASS_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name, RPG::dbClasses[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbClasses[primaryKeyIndex]->defense[secondaryKeyIndex], CSV_HEADER_CLASS_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, CLASS_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Class Intelligence database values to a .csv file
/*!
    exportClassesIntelligence exports the current Class Intelligence database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesIntelligence( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesIntelligence( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_INTELLIGENCE_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorIntelligence;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CLASS_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name, RPG::dbClasses[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbClasses[primaryKeyIndex]->intelligence[secondaryKeyIndex], CSV_HEADER_CLASS_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, CLASS_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Class Agility database values to a .csv file
/*!
    exportClassesAgility exports the current Class Agility database values to a .csv file, using a given file
    name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesAgility( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesAgility( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_AGILITY_DATABASE_FILE_NAME_END;          // Output file name
    int secondaryKeyIndex;
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataActorAgility;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CLASS_STAT - 2;  // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbClasses[primaryKeyIndex]->name, RPG::dbClasses[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Stats at each level
    WRITE_CSV_ROWS_ARRAY( RPG::dbClasses[primaryKeyIndex]->agility[secondaryKeyIndex], CSV_HEADER_CLASS_STAT[CSV_INDEX_ACTOR_STAT_LEVEL_1] << secondaryKeyIndex, 1, CLASS_MAX_LEVEL );
    // Close output file
    outputStream.close();
}

//! Export the current Class skill learning database values to a .csv file
/*!
    exportClassesSkillLearning exports the current Class skill learning database values to a .csv
    file, using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesSkillLearning( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesSkillLearning( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_SKILL_LEARNING_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 0;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;        // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataClassSkillLearning;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_CLASS_SKILL_LEARNING;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbClasses[primaryKeyIndex]->skillLearning.count() - 1,
        RPG::dbClasses[primaryKeyIndex]->name << " - " << RPG::skills[RPG::dbClasses[primaryKeyIndex]->skillLearning[secondaryKeyIndex]->skillId]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Level requirement
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbClasses[primaryKeyIndex]->skillLearning[secondaryKeyIndex]->levelReq, RPG::dbClasses[primaryKeyIndex]->skillLearning.count() - 1 + secondaryKeyBase );
    // Skill ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbClasses[primaryKeyIndex]->skillLearning[secondaryKeyIndex]->skillId, RPG::dbClasses[primaryKeyIndex]->skillLearning.count() - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Class Condition database values to a .csv file
/*!
    exportClassesConditions exports the current Class Conditions database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesConditions( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesConditions( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_CONDITIONS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataClassConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_CLASS_CONDITION;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbClasses[primaryKeyIndex]->conditions.size,
        RPG::dbClasses[primaryKeyIndex]->name << " - " << RPG::conditions[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Condition susceptibility rank
    WRITE_CSV_ROW_DUAL_KEY( damageMultiplierToCsvToken[RPG::dbClasses[primaryKeyIndex]->conditions[secondaryKeyIndex]], RPG::dbClasses[primaryKeyIndex]->conditions.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Class Condition database values to a .csv file
/*!
    exportClassesAttributes exports the current Class Attributes database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportClassesAttributes( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportClassesAttributes( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CLASS_ATTRIBUTES_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbClasses.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataClassAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_CLASS_ATTRIBUTE;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbClasses[primaryKeyIndex]->attributes.size,
        RPG::dbClasses[primaryKeyIndex]->name << " - " << RPG::attributes[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Condition susceptibility rank
    WRITE_CSV_ROW_DUAL_KEY( damageMultiplierToCsvToken[RPG::dbClasses[primaryKeyIndex]->attributes[secondaryKeyIndex]], RPG::dbClasses[primaryKeyIndex]->attributes.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Skill database values to a .csv file
/*!
    exportSkills exports the current Skill database values (not including array values like
    conditions, attributes, etc.) to a .csv file, using a given file name base for the file name.
    This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportSkills( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportSkills( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + SKILL_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::skills.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataSkill;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_SKILL - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_SKILL;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::skills[primaryKeyIndex]->name, RPG::skills[primaryKeyIndex]->id )
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Name
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->name );
    // Description
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->description );
    // Use message 2K
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->useMsg2k );
    // Use message 2K alt
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->useMsg2kAlt );
    // Evasion type
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->evasionType );
    // Type
    WRITE_CSV_ROW_SIMPLE( skillTypeToCsvToken[RPG::skills[primaryKeyIndex]->type] );
    // MP consumed
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->mpCost );
    // Is MP percent
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->isMpPercent );
    // MP consumed percent
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->mpCostPercent );
    // Target
    WRITE_CSV_ROW_SIMPLE( skillTargetToCsvToken[RPG::skills[primaryKeyIndex]->target] );
    // Switch ID
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->switchId );
    // Battle animation ID
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->battleAnimId );
    // Use in field
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->useInField );
    // Use in battle
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->useInBattle );
    // Attack influence
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->atkInfluence );
    // Intelligence influence
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->intInfluence );
    // Variance
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->variance );
    // Effect rating
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->effectRating );
    // Success rate
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->successRate );
    // Affect HP
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->affectHp );
    // Affect MP
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->affectMp );
    // Affect strength
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->affectStr );
    // Affect defense
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->affectDef );
    // Affect intelligence
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->affectInt );
    // Affect agility
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->affectAgl );
    // Absorb
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->absorb );
    // Ignore defense
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->ignoreDef );
    // Attribute resist
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->attributeResist );
    // Condition flag
    WRITE_CSV_ROW_SIMPLE( RPG::skills[primaryKeyIndex]->conditionFlag );
    // Close output file
    outputStream.close();
}

//! Export the current Skill Condition database values to a .csv file
/*!
    exportSkillsConditions exports the current Skill Conditions database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportSkillsConditions( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportSkillsConditions( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + SKILL_CONDITIONS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::skills.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataSkillConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_SKILL_CONDITION;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::skills[primaryKeyIndex]->conditions.size,
        RPG::skills[primaryKeyIndex]->name << " - " << RPG::conditions[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Condition flag
    // RPG::Skill::conditions is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_DUAL_KEY( RPG::skills[primaryKeyIndex]->conditions[secondaryKeyIndex - 1], RPG::skills[primaryKeyIndex]->conditions.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Skill Attribute database values to a .csv file
/*!
    exportSkillsAttributes exports the current Skill Attributes database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportSkillsAttributes( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportSkillsAttributes( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + SKILL_ATTRIBUTES_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::skills.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataSkillAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_SKILL_ATTRIBUTE;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::skills[primaryKeyIndex]->attributes.size,
        RPG::skills[primaryKeyIndex]->name << " - " << RPG::attributes[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Attribute flag
    // RPG::Skill::attributes is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_DUAL_KEY( RPG::skills[primaryKeyIndex]->attributes[secondaryKeyIndex - 1], RPG::skills[primaryKeyIndex]->attributes.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Skill Animation database values to a .csv file
/*!
    exportSkillsAnimations exports the current Skill Animations database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportSkillsAnimations( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportSkillsAnimations( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + SKILL_ANIMATIONS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::skills.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataSkillAnimations;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_SKILL_ANIMATION;   // Pointer to the array of header strings to use

    // NOTE: The RPG::Skill::skillAnims structure is supposed to hold information about skill
    // animations for each actor. However, for some reason some of the skillAnims structures contain
    // a different number of entries than there are actors. The number of entries does seem to be
    // somehow linked to the number of actors at some time -- for example, after I increased the
    // number of actors to 35 AND fiddled around in the skill animations in the RM2K3 database
    // editor, some of the skillAnims changed to 35 entries. However, I haven't been able to reverse
    // that change by switching the numbers back and doing more fiddling, and I don't understand how
    // it really works. So I decided to use the minimum between RPG::dbActors.count() and
    // RPG::skills[primaryKeyIndex]->skillAnims.count() as the secondary key limit in this function,
    // even though that means some of the skillAnims entries (which I believe are empty data) are
    // skipped that way.
    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        std::min( RPG::skills[primaryKeyIndex]->skillAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase,
        RPG::skills[primaryKeyIndex]->name << " - " << RPG::dbActors[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Battler pose
    WRITE_CSV_ROW_DUAL_KEY( animationBattlerPoseToCsvToken[RPG::skills[primaryKeyIndex]->skillAnims[secondaryKeyIndex]->battlerPose], std::min( RPG::skills[primaryKeyIndex]->skillAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Movement
    WRITE_CSV_ROW_DUAL_KEY( animationMoveBeforeAttackToCsvToken[RPG::skills[primaryKeyIndex]->skillAnims[secondaryKeyIndex]->movement], std::min( RPG::skills[primaryKeyIndex]->skillAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Add after image
    WRITE_CSV_ROW_DUAL_KEY( RPG::skills[primaryKeyIndex]->skillAnims[secondaryKeyIndex]->addAfterImage, std::min( RPG::skills[primaryKeyIndex]->skillAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Item database values to a .csv file
/*!
    exportItems exports the current Item database values (not including array values like Attributes, Conditions,
    etc.) to a .csv file, using a given file name base for the file name. This function is used as a
    subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportItems( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportItems( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ITEM_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::items.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataItem;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ITEM - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ITEM;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::items[primaryKeyIndex]->name, RPG::items[primaryKeyIndex]->id );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Name
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->name );
    // Description
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->description );
    // Type
    WRITE_CSV_ROW_SIMPLE( itemTypeToCsvToken[RPG::items[primaryKeyIndex]->type] );
    // Price
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->price );
    // Number of uses
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->numUses );
    // Attack
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->str );
    // Defense
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->def );
    // Intelligence
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->intl );
    // Agility
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->agl );
    // Two-handed
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->twoHanded );
    // MP consumed
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->mpCost );
    // Hit probability
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->hitProb );
    // Critical bonus
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->critBonus );
    // Battle animation ID
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->battleAnimId );
    // Always attack first
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->alwaysAttackFirst );
    // Attack twice
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->attackTwice );
    // Attack all
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->attackAll );
    // Ignore evasion
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->ignoreEvasion );
    // Prevent critical hits
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->preventCrit );
    // Increased evasion
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->increaseEvasion );
    // Half MP consumption
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->halfMpCost );
    // No terrain damage
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->noTerrainDmg );
    // Cursed
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->isCursed );
    // Target
    WRITE_CSV_ROW_SIMPLE( itemTargetToCsvToken[RPG::items[primaryKeyIndex]->target] );
    // HP restore percent
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->hpRestorePercent );
    // HP restore units
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->hpRestoreUnits );
    // MP restore percent
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->mpRestorePercent );
    // MP restore units
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->mpRestoreUnits );
    // Only usable on the field map
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->onlyInField );
    // Only usable on dead characters
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->onlyOnDead );
    // Add max HP
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->addMaxHp );
    // Add max MP
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->addMaxMp );
    // Add attack
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->addStr );
    // Add defense
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->addDef );
    // Add intelligence
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->addInt );
    // Add agility
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->addAgl );
    // Message shown when used
    WRITE_CSV_ROW_SIMPLE( itemScrollMsgShownToCsvToken[RPG::items[primaryKeyIndex]->messageShown] );
    // Invoke skill when used
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->invokeSkill );
    // Skill ID
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->skillId );
    // Switch ID
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->switchId );
    // Activation allowed in field
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->useInField );
    // Activation allowed in battle
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->useInBattle );
    // Condition effectiveness
    WRITE_CSV_ROW_SIMPLE( RPG::items[primaryKeyIndex]->conditionPercentage );
    // Condition resist or inflict
    WRITE_CSV_ROW_SIMPLE( itemConditionFlagToCsvToken[RPG::items[primaryKeyIndex]->conditionFlag] );
    // Weapon trajectory
    WRITE_CSV_ROW_SIMPLE( weaponRangedTrajectoryToCsvToken[RPG::items[primaryKeyIndex]->weaponTrajectory] );
    // Weapon targetting
    WRITE_CSV_ROW_SIMPLE( weaponRangedTargettingToCsvToken[RPG::items[primaryKeyIndex]->weaponTargetting] );
    // Close output file
    outputStream.close();
}

//! Export the current Item Usable By Actor database values to a .csv file
/*!
    exportItemsUsableByActors exports the current Item Usable By Actor database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportItemsUsableByActors( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportItemsUsableByActors( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ITEM_USABLE_BY_ACTORS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::items.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemUsableByActors;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_USABLE_BY_ACTOR;   // Pointer to the array of header strings to use

    // NOTE: The number of entries in a RPG::Item::usableByHero structure does not always match the
    // size of RPG::dbActors.count(). Hence we use the minimum between them as a secondary key limit
    // in order to avoid violations, even though this means some data (probably empty) is skipped.
    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        std::min( RPG::items[primaryKeyIndex]->usableByHero.size, RPG::dbActors.count() ),
        RPG::items[primaryKeyIndex]->name << " - " << RPG::dbActors[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Actor can use
    // RPG::Item::usableByHero is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->usableByHero[secondaryKeyIndex - 1], std::min( RPG::items[primaryKeyIndex]->usableByHero.size, RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Item Usable By Class database values to a .csv file
/*!
    exportItemsUsableByClasses exports the current Item Usable By Class database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportItemsUsableByClasses( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportItemsUsableByClasses( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ITEM_USABLE_BY_CLASSES_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 0;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::items.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemUsableByClasses;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_USABLE_BY_CLASS;   // Pointer to the array of header strings to use

    // NOTE: The 0th entry in RPG::Item::usableByClasses indicates whether an actor with no class is
    // allowed to use the item. Hence the note value for entries with secondaryKeyIndex of 0 should
    // indicate "(None)", and the potential size of the array is 1 greater than the size of
    // RPG::dbClasses.
    // NOTE: The number of entries in a RPG::Item::usableByClasses structure does not always match
    // the size of RPG::dbClasses.count() + 1. Hence we use the minimum between them as a secondary
    // key limit in order to avoid violations, even though this means some data (probably empty) is
    // skipped.
    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        std::min( RPG::items[primaryKeyIndex]->usableByClass.size, RPG::dbClasses.count() + 1 ) - 1 + secondaryKeyBase,
        RPG::items[primaryKeyIndex]->name << " - " << ( 0 == secondaryKeyIndex ? "(None)" : RPG::dbClasses[secondaryKeyIndex]->name ) );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Class can use
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->usableByClass[secondaryKeyIndex], std::min( RPG::items[primaryKeyIndex]->usableByClass.size, RPG::dbClasses.count() + 1 ) - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Item Condition database values to a .csv file
/*!
    exportItemsConditions exports the current Item Conditions database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportItemsConditions( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportItemsConditions( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ITEM_CONDITIONS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::items.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_CONDITION;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::items[primaryKeyIndex]->conditions.size,
        RPG::items[primaryKeyIndex]->name << " - " << RPG::conditions[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Affects condition
    // RPG::Item::conditions is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->conditions[secondaryKeyIndex - 1], RPG::items[primaryKeyIndex]->conditions.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Item Attribute database values to a .csv file
/*!
    exportItemsAttributes exports the current Item Attributes database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportItemsAttributes( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportItemsAttributes( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ITEM_ATTRIBUTES_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::items.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_ATTRIBUTE;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::items[primaryKeyIndex]->attributes.size,
        RPG::items[primaryKeyIndex]->name << " - " << RPG::attributes[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Affects condition
    // RPG::Item::attributes is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->attributes[secondaryKeyIndex - 1], RPG::items[primaryKeyIndex]->attributes.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Item Weapon Animation database values to a .csv file
/*!
    exportItemsWeaponAnimations exports the current Item Weapon Animations database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportItemsWeaponAnimations( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportItemsWeaponAnimations( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ITEM_WEAPON_ANIMATIONS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::items.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemWeaponAnimations;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_WEAPON_ANIMATION;   // Pointer to the array of header strings to use

    // NOTE: The number of entries in a RPG::Item::weaponAnims structure does not always match the
    // size of RPG::dbActors.count(). Hence we use the minimum between them as a secondary key limit
    // in order to avoid violations, even though this means some data (probably empty) is skipped.
    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase,
        RPG::items[primaryKeyIndex]->name << " - " << RPG::dbActors[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Battler pose
    WRITE_CSV_ROW_DUAL_KEY( animationBattlerPoseToCsvToken[RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->battlerPose], std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Animation type
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->isBattleAnim, std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Weapon animation ID
    // RPG::AnimationWeapon::battleAnimId is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->battleAnimId + 1, std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Movement before attack
    WRITE_CSV_ROW_DUAL_KEY( animationMoveBeforeAttackToCsvToken[RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->movement], std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Add afterimage
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->addAfterImage, std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Number of attacks
    WRITE_CSV_ROW_DUAL_KEY( weaponNumberOfAttacksToCsvToken[RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->numAttacks], std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Ranged weapon
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->isRangedWeapon, std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Ranged weapon animation ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->rangedAnimId, std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Ranged weapon animation speed
    WRITE_CSV_ROW_DUAL_KEY( weaponRangedAnimSpeedToCsvToken[RPG::items[primaryKeyIndex]->weaponAnims[secondaryKeyIndex]->rangedAnimSpeed], std::min( RPG::items[primaryKeyIndex]->weaponAnims.count(), RPG::dbActors.count() ) - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Enemy database values to a .csv file
/*!
    exportEnemies exports the current Enemy database values (not including array values like Attributes, Conditions,
    etc.) to a .csv file, using a given file name base for the file name. This function is used as a
    subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportEnemies( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportEnemies( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ENEMY_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::dbMonsters.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataEnemy;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ENEMY - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ENEMY;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->name, RPG::dbMonsters[primaryKeyIndex]->id );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Name
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->name );
    // File name
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->filename );
    // Color alteration
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->colorAlteration );
    // Is transparent
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->isTransparent );
    // Is flying
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->isFlying );
    // Max HP
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->maxHp );
    // Max MP
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->maxMp );
    // Attack
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->attack );
    // Defense
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->defense );
    // Intelligence
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->intelligence );
    // Agility
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->agility );
    // Experience
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->exp );
    // Money
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->gold );
    // Item ID
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->itemId );
    // Item probability
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->itemProbability );
    // Allow critical hit
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->critFlag );
    // Critical hit probability
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->critPercentage );
    // Attacks often miss
    WRITE_CSV_ROW_SIMPLE( RPG::dbMonsters[primaryKeyIndex]->oftenMiss );
    // Close output file
    outputStream.close();
}

//! Export the current Monster Condition database values to a .csv file
/*!
    exportEnemiesConditions exports the current Monster Conditions database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportEnemiesConditions( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportEnemiesConditions( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ENEMY_CONDITIONS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbMonsters.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataEnemyConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ENEMY_CONDITION;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbMonsters[primaryKeyIndex]->conditions.size,
        RPG::dbMonsters[primaryKeyIndex]->name << " - " << RPG::conditions[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Condition susceptibility rank
    WRITE_CSV_ROW_DUAL_KEY( damageMultiplierToCsvToken[RPG::dbMonsters[primaryKeyIndex]->conditions[secondaryKeyIndex]], RPG::dbMonsters[primaryKeyIndex]->conditions.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Monster Condition database values to a .csv file
/*!
    exportEnemiesAttributes exports the current Monster Attributes database values to a .csv file,
    using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportEnemiesAttributes( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportEnemiesAttributes( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ENEMY_ATTRIBUTES_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 1;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbMonsters.count() - 1 + primaryKeyBase;    // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataEnemyAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ENEMY_ATTRIBUTE;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbMonsters[primaryKeyIndex]->attributes.size,
        RPG::dbMonsters[primaryKeyIndex]->name << " - " << RPG::attributes[secondaryKeyIndex]->name );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Attribute susceptibility rank
    WRITE_CSV_ROW_DUAL_KEY( damageMultiplierToCsvToken[RPG::dbMonsters[primaryKeyIndex]->attributes[secondaryKeyIndex]], RPG::dbMonsters[primaryKeyIndex]->attributes.size - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Monster behavior database values to a .csv file
/*!
    exportEnemiesBehaviors exports the current Monster behavior database values to a .csv
    file, using a given file name base for the file name. This function is used as a subroutine of
    exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportEnemiesBehaviors( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportEnemiesBehaviors( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ENEMY_BEHAVIORS_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1, secondaryKeyBase = 0;       // Base values (i.e. 0-based or 1-based) of the key indexes
    int primaryKeyLimit = RPG::dbMonsters.count() - 1 + primaryKeyBase;        // Limit value for primary key index (secondary key index limit is dependent on current primary key, so cannot be set to a static value)
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataEnemyBehaviors;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ENEMY_BEHAVIOR;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_DUAL_KEY(
        RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1,
        RPG::dbMonsters[primaryKeyIndex]->name << " - Behavior " << secondaryKeyIndex );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Monster precondition
    WRITE_CSV_ROW_DUAL_KEY( monsterPreconditionToCsvToken[RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->precondition], RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Precondition param 1
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->preconditionParam[0], RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Precondition param 2
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->preconditionParam[1], RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Precondition switch ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->preconditionSwitchId, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Priority
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->priority, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Switch on
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->switchOn, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Switch on ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->switchOnId, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Switch off
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->switchOff, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Switch off ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->switchOffId, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Action kind
    WRITE_CSV_ROW_DUAL_KEY( actionKindToCsvToken[RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->kind], RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Basic action type
    WRITE_CSV_ROW_DUAL_KEY( basicActionToCsvToken[RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->basicActionId], RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Skill ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->skillId, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Transform enemy ID
    WRITE_CSV_ROW_DUAL_KEY( RPG::dbMonsters[primaryKeyIndex]->behavior[secondaryKeyIndex]->monsterId, RPG::dbMonsters[primaryKeyIndex]->behavior.count() - 1 + secondaryKeyBase );
    // Close output file
    outputStream.close();
}

//! Export the current Attribute database values to a .csv file
/*!
    exportAttributes exports the current Attribute database values to a .csv file, using a given
    file name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportAttributes( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportAttributes( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + ATTRIBUTE_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::attributes.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataAttributes;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_ATTRIBUTE - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_ATTRIBUTE;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::attributes[primaryKeyIndex]->name, RPG::attributes[primaryKeyIndex]->id );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Name
    WRITE_CSV_ROW_SIMPLE( RPG::attributes[primaryKeyIndex]->name );
    // Type
    WRITE_CSV_ROW_SIMPLE( attributeTypeToCsvToken[RPG::attributes[primaryKeyIndex]->type] );
    // Damage multiplier A
    WRITE_CSV_ROW_SIMPLE( RPG::attributes[primaryKeyIndex]->dmgA );
    // Damage multiplier B
    WRITE_CSV_ROW_SIMPLE( RPG::attributes[primaryKeyIndex]->dmgB );
    // Damage multiplier C
    WRITE_CSV_ROW_SIMPLE( RPG::attributes[primaryKeyIndex]->dmgC );
    // Damage multiplier D
    WRITE_CSV_ROW_SIMPLE( RPG::attributes[primaryKeyIndex]->dmgD );
    // Damage multiplier E
    WRITE_CSV_ROW_SIMPLE( RPG::attributes[primaryKeyIndex]->dmgE );
    // Close output file
    outputStream.close();
}

//! Export the current Condition database values to a .csv file
/*!
    exportConditions exports the current Condition database values to a .csv file, using a given
    file name base for the file name. This function is used as a subroutine of exportDatabase.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportConditions( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportConditions( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::string fileName = r_FileNameBase + CONDITION_DATABASE_FILE_NAME_END;          // Output file name
    int primaryKeyBase = 1;       // Base value (i.e. 0-based or 1-based) of the key index
    int primaryKeyLimit = RPG::conditions.count() - 1 + primaryKeyBase;        // Limit value for primary key index
    bool * overriddenDataStructure = overriddenDataConditions;   // Pointer to the overridden data structure to use
    int overriddenDataStructureSize = CSV_ROWS_CONDITION - 2;   // Size of the overridden data structure (disincludes the "Notes (optional)" and "ID" rows)
    std::string * csvHeaderArray = CSV_HEADER_CONDITION;   // Pointer to the array of header strings to use

    SETUP_EXPORT_FUNCTION_SIMPLE( RPG::conditions[primaryKeyIndex]->name, RPG::conditions[primaryKeyIndex]->id );
    // Go through all data for this record type and export the desired ones (everything if r_ExportAll is true, only modified data otherwise)
    // Name
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->name );
    // Persists after battle
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->persistsAfterBattle );
    // Color
    // RPG::Condition::color is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->color + 1 );
    // Priority
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->priority );
    // Action restriction
    WRITE_CSV_ROW_SIMPLE( actionRestrictionToCsvToken[RPG::conditions[primaryKeyIndex]->actionRestriction] );
    // Hit percentage
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->hitPercentage );
    // Evade all attacks
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->evadeAttacks );
    // Reflect skills
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->reflectSkills );
    // Equipment cannot be changed
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->lockEquipment );
    // Susceptibility A
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->susA );
    // Susceptibility B
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->susB );
    // Susceptibility C
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->susC );
    // Susceptibility D
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->susD );
    // Susceptibility E
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->susE );
    // Recovery turns
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->recoveryTurns );
    // Recovery turns percent
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->recoveryTurnsPercent );
    // Recovery physical percent
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->recoveryPhysicalPercent );
    // Base stat alteration
    WRITE_CSV_ROW_SIMPLE( baseStatAlterationToCsvToken[RPG::conditions[primaryKeyIndex]->statAlteration] );
    // Alters attack
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->alterStr );
    // Alters defense
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->alterDef );
    // Alters intelligence
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->alterInt );
    // Alters agility
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->alterAgi );
    // Prevent skill by attack
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->preventSkillByAtk );
    // Prevent skill attack influence
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->atkInfluence );
    // Prevent skill by intelligence
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->preventSkillByInt );
    // Prevent skill intelligence influence
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->intInfluence );
    // HP alteration type
    WRITE_CSV_ROW_SIMPLE( alterationTypeToCsvToken[RPG::conditions[primaryKeyIndex]->hpAlterationType] );
    // HP damage percent
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->hpDamagePercentage );
    // HP damage units
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->hpDamageUnits );
    // HP damage on step
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->hpDamageEachStep );
    // HP damage step number
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->hpDamageStepNum );
    // MP alteration type
    WRITE_CSV_ROW_SIMPLE( alterationTypeToCsvToken[RPG::conditions[primaryKeyIndex]->mpAlterationType] );
    // MP damage percent
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->mpDamagePercentage );
    // MP damage units
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->mpDamageUnits );
    // MP damage on step
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->mpDamageEachStep );
    // MP damage step number
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->mpDamageStepNum );
    // Animation ID
    // RPG::Condition::poseId is 0-based even though the thing it refers to conceptually has 1-based indexing
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->poseId + 1 );
    // Message 2K ally state
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->msg2kAllyState );
    // Message 2K enemy state
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->msg2kEnemyState );
    // Message 2K already in
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->msg2kAlreadyIn );
    // Message 2K regular
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->msg2kRegular );
    // Message 2K recovery
    WRITE_CSV_ROW_SIMPLE( RPG::conditions[primaryKeyIndex]->msg2kRecovery );
    // Close output file
    outputStream.close();
}

//! Export the current database values to .csv files
/*!
    exportDatabase exports the current database values to a collection of .csv files, using a given
    file name base for the file names. This function is used to record the current database
    associated with save states, and can also be called with the
    @dyndatabaseoverride_export_database comment command.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void exportDatabase( std::string r_FileNameBase, bool r_ExportAll )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "exportDatabase( " << r_FileNameBase << ", " << r_ExportAll << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    // Call all relevant database export functions
    exportActors( r_FileNameBase, r_ExportAll );
    exportActorsHp( r_FileNameBase, r_ExportAll );
    exportActorsMp( r_FileNameBase, r_ExportAll );
    exportActorsAttack( r_FileNameBase, r_ExportAll );
    exportActorsDefense( r_FileNameBase, r_ExportAll );
    exportActorsIntelligence( r_FileNameBase, r_ExportAll );
    exportActorsAgility( r_FileNameBase, r_ExportAll );
    exportActorsSkillLearning( r_FileNameBase, r_ExportAll );
    exportActorsConditions( r_FileNameBase, r_ExportAll );
    exportActorsAttributes( r_FileNameBase, r_ExportAll );
    exportClasses( r_FileNameBase, r_ExportAll );
    exportClassesHp( r_FileNameBase, r_ExportAll );
    exportClassesMp( r_FileNameBase, r_ExportAll );
    exportClassesAttack( r_FileNameBase, r_ExportAll );
    exportClassesDefense( r_FileNameBase, r_ExportAll );
    exportClassesIntelligence( r_FileNameBase, r_ExportAll );
    exportClassesAgility( r_FileNameBase, r_ExportAll );
    exportClassesSkillLearning( r_FileNameBase, r_ExportAll );
    exportClassesConditions( r_FileNameBase, r_ExportAll );
    exportClassesAttributes( r_FileNameBase, r_ExportAll );
    exportSkills( r_FileNameBase, r_ExportAll );
    exportSkillsConditions( r_FileNameBase, r_ExportAll );
    exportSkillsAttributes( r_FileNameBase, r_ExportAll );
    exportSkillsAnimations( r_FileNameBase, r_ExportAll );
    exportItems( r_FileNameBase, r_ExportAll );
    exportItemsUsableByActors( r_FileNameBase, r_ExportAll );
    exportItemsUsableByClasses( r_FileNameBase, r_ExportAll );
    exportItemsConditions( r_FileNameBase, r_ExportAll );
    exportItemsAttributes( r_FileNameBase, r_ExportAll );
    exportItemsWeaponAnimations( r_FileNameBase, r_ExportAll );
    exportEnemies( r_FileNameBase, r_ExportAll );
    exportEnemiesConditions( r_FileNameBase, r_ExportAll );
    exportEnemiesAttributes( r_FileNameBase, r_ExportAll );
    exportEnemiesBehaviors( r_FileNameBase, r_ExportAll );
    exportAttributes( r_FileNameBase, r_ExportAll );
    exportConditions( r_FileNameBase, r_ExportAll );
}

// Import macros
/*
    SETUP_IMPORT_FUNCTION_SIMPLE is a macro for setting up an import function in context of a simple
    data value. For example, in the importActors function, each line of the CSV file holds fairly
    simple values -- int, bool, or string -- while more complicated data structures nested within
    the DBActor class are handled by other functions.

    It is expected that the following variables will be defined and initialized in the function
    before this macro is used (the values used here are from importActors and are just examples):

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataStructure = overriddenDataActor;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR;    // Pointer to the array of header strings to use

    ***** Human-readable representation *****

    CSVIterator lineTokens( inputStream );      // Iterator for traversing input data
    std::string curToken;                       // Temporary token string
    int rowIndex, overriddenDataIndex;          // Index variables
    unsigned int columnIndex;                   // Index of the currently considered column from the .csv file
    std::list<int> primaryKeyList;                      // List of ID values
    std::list<int>::iterator primaryKeyListIterator;    // Iterator for primaryKeyList

    // Skip the Notes (optional) row if present
    curToken = (*lineTokens)[0];
    if( csvHeaderArray[0] == curToken ) {
        ++lineTokens;
        curToken = (*lineTokens)[0];
    }
    // Get the list of ID values
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) {
        curToken = (*lineTokens)[columnIndex];
        primaryKeyList.push_back( std::atoi( curToken.c_str() ) );
    }
    ++lineTokens;
    curToken = (*lineTokens)[0];
    // To offset the fact that READ_CSV_ROW_SIMPLE macros start by incrementing rowIndex and
    // overriddenDataIndex, initialize them to 1 less than actual starting index
    rowIndex = 1;
    overriddenDataIndex = -1;
*/
#define SETUP_IMPORT_FUNCTION_SIMPLE \
    CSVIterator lineTokens( inputStream ); \
    std::string curToken; \
    int rowIndex, overriddenDataIndex; \
    unsigned int columnIndex; \
    std::list<int> primaryKeyList; \
    std::list<int>::iterator primaryKeyListIterator; \
    curToken = (*lineTokens)[0]; \
    if( csvHeaderArray[0] == curToken ) { \
        ++lineTokens; \
        curToken = (*lineTokens)[0]; \
    } \
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) { \
        curToken = (*lineTokens)[columnIndex]; \
        primaryKeyList.push_back( std::atoi( curToken.c_str() ) ); \
    } \
    ++lineTokens; \
    curToken = (*lineTokens)[0]; \
    rowIndex = 1; \
    overriddenDataIndex = -1;

/*
    SETUP_IMPORT_FUNCTION_ARRAY is a macro for setting up an import function in context of an array
    of simple data values. For example, in the exportActorsHp function, each line of the CSV file
    holds int values associated with a given Actor level across all Actors.

    It is expected that the following variables will be defined and initialized in the function
    before this macro is used (the values used here are from importActorsHp and are just examples):

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataStructure = overriddenDataActor;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    ***** Human-readable representation *****

    CSVIterator lineTokens( inputStream );              // Iterator for traversing input data
    std::string curToken;                               // Temporary token string
    int secondaryKeyIndex, overriddenDataIndex;         // Index variables
    unsigned int columnIndex;                           // Index of the currently considered column from the .csv file
    std::list<int> primaryKeyList;                      // List of ID values
    std::list<int>::iterator primaryKeyListIterator;    // Iterator for primaryKeyList

    // Skip the Notes (optional) row if present
    curToken = (*lineTokens)[0];
    if( csvHeaderArray[0] == curToken ) {
        ++lineTokens;
        curToken = (*lineTokens)[0];
    }
    // Get the list of ID values
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) {
        curToken = (*lineTokens)[columnIndex];
        primaryKeyList.push_back( std::atoi( curToken.c_str() ) );
    }
    ++lineTokens;
    curToken = (*lineTokens)[0];
*/
#define SETUP_IMPORT_FUNCTION_ARRAY \
    CSVIterator lineTokens( inputStream ); \
    std::string curToken; \
    int secondaryKeyIndex, overriddenDataIndex; \
    unsigned int columnIndex; \
    std::list<int> primaryKeyList; \
    std::list<int>::iterator primaryKeyListIterator; \
    curToken = (*lineTokens)[0]; \
    if( csvHeaderArray[0] == curToken ) { \
        ++lineTokens; \
        curToken = (*lineTokens)[0]; \
    } \
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) { \
        curToken = (*lineTokens)[columnIndex]; \
        primaryKeyList.push_back( std::atoi( curToken.c_str() ) ); \
    } \
    ++lineTokens; \
    curToken = (*lineTokens)[0];

/*
    SETUP_IMPORT_FUNCTION_DUAL_KEY is a macro for setting up an import function in context of a
    dual-key data structure. For example, in the importActorsSkillLearning function, we are dealing
    with individual SkillProgression data structures within each DBActor, and the SkillProgression
    data structures have varying lengths. Hence we need to accommodate both the DBActor (primary
    key) and SkillProgression (secondary key) data structures.

    It is expected that the following variables will be defined and initialized in the function
    before this macro is used (the values used here are from importActorsSkillLearning and are just
    examples):

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorSkillLearning;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_SKILL_LEARNING;   // Pointer to the array of header strings to use

    ***** Human-readable representation *****

    CSVIterator lineTokens( inputStream );      // Iterator for traversing input data
    std::string curToken;                       // Temporary token string
    int rowIndex;                               // Current CSV row index
    unsigned int columnIndex;                   // Current CSV column index
    std::list<int> primaryKeyIdList, secondaryKeyIdList;                 // Lists of key ID values
    std::list<int>::iterator primaryKeyIdListIterator, secondaryKeyIdListIterator;    // Iterators for key ID lists

    // Skip the Notes (optional) row if present
    curToken = (*lineTokens)[0];
    if( csvHeaderArray[0] == curToken ) {
        ++lineTokens;
        curToken = (*lineTokens)[0];
    }
    // Get the list of primary key ID values
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) {
        curToken = (*lineTokens)[columnIndex];
        primaryKeyIdList.push_back( std::atoi( curToken.c_str() ) );
    }
    ++lineTokens;
    // Get the list of secondary key ID values
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) {
        curToken = (*lineTokens)[columnIndex];
        secondaryKeyIdList.push_back( std::atoi( curToken.c_str() ) );
    }
    // Record overridden flags if necessary
    if( r_boolSetOverriddenFlags ) {
        primaryKeyIdListIterator = primaryKeyIdList.begin();
        for( secondaryKeyIdListIterator = secondaryKeyIdList.begin(); secondaryKeyIdList.end() != secondaryKeyIdListIterator; ++secondaryKeyIdListIterator ) {
            overriddenDataStructure->insert( std::pair<int, int>( *primaryKeyIdListIterator, *secondaryKeyIdListIterator ) );
            ++primaryKeyIdListIterator;
        }
    }
    ++lineTokens;
    curToken = (*lineTokens)[0];
    // Start retrieving other data at row index 3 (after Notes (0), primary key (1), and secondary
    // key (2) rows), bearing in mind that each call of READ_CSV_ROW_DUAL_KEY will increment
    // rowIndex by 1
    rowIndex = 2;

*/
#define SETUP_IMPORT_FUNCTION_DUAL_KEY \
    CSVIterator lineTokens( inputStream ); \
    std::string curToken; \
    int rowIndex; \
    unsigned int columnIndex; \
    std::list<int> primaryKeyIdList, secondaryKeyIdList; \
    std::list<int>::iterator primaryKeyIdListIterator, secondaryKeyIdListIterator; \
    curToken = (*lineTokens)[0]; \
    if( csvHeaderArray[0] == curToken ) { \
        ++lineTokens; \
        curToken = (*lineTokens)[0]; \
    } \
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) { \
        curToken = (*lineTokens)[columnIndex]; \
        primaryKeyIdList.push_back( std::atoi( curToken.c_str() ) ); \
    } \
    ++lineTokens; \
    for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) { \
        curToken = (*lineTokens)[columnIndex]; \
        secondaryKeyIdList.push_back( std::atoi( curToken.c_str() ) ); \
    }\
    if( r_boolSetOverriddenFlags ) { \
        primaryKeyIdListIterator = primaryKeyIdList.begin(); \
        for( secondaryKeyIdListIterator = secondaryKeyIdList.begin(); secondaryKeyIdList.end() != secondaryKeyIdListIterator; ++secondaryKeyIdListIterator ) { \
            overriddenDataStructure->insert( std::pair<int, int>( *primaryKeyIdListIterator, *secondaryKeyIdListIterator ) ); \
            ++primaryKeyIdListIterator; \
        } \
    } \
    ++lineTokens; \
    curToken = (*lineTokens)[0]; \
    rowIndex = 2;

/*
    READ_CSV_ROW_SIMPLE is a macro for reading a line from a CSV file in context of a simple data
    value. For example, in the importActors function, each line of the CSV file holds fairly simple
    values -- int, bool, or string -- while more complicated data structures nested within the
    DBActor class are handled by other functions.

    It is expected that the SETUP_IMPORT_FUNCTION_SIMPLE macro will be used in the function before
    this macro is used.

    ***** Human-readable representation *****

    rowIndex++;
    overriddenDataIndex++;
    if( csvHeaderArray[rowIndex] == curToken ) {
        if( r_boolSetOverriddenFlags ) {
            overriddenDataArray[overriddenDataIndex] = true;
        }
        primaryKeyListIterator = primaryKeyList.begin();
        for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) {
            curToken = (*lineTokens)[columnIndex];
            r_DatabaseTarget = r_Value;
            ++primaryKeyListIterator;
        }
        ++lineTokens;
        curToken = (*lineTokens)[0];
    }
*/
#define READ_CSV_ROW_SIMPLE(r_DatabaseTarget, r_Value) \
    rowIndex++; \
    overriddenDataIndex++; \
    if( csvHeaderArray[rowIndex] == curToken ) { \
        if( r_boolSetOverriddenFlags ) { \
            overriddenDataArray[overriddenDataIndex] = true; \
        } \
        primaryKeyListIterator = primaryKeyList.begin(); \
        for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) { \
            curToken = (*lineTokens)[columnIndex]; \
            r_DatabaseTarget = r_Value; \
            ++primaryKeyListIterator; \
        } \
        ++lineTokens; \
        curToken = (*lineTokens)[0]; \
    }

/*
    READ_CSV_ROWS_ARRAY is a macro for reading multiple lines from a CSV file in context of an array
    of simple data values. For example, in the exportActorsHp function, each line of the CSV file
    holds int values associated with a given Actor level across all Actors.

    It is expected that the SETUP_IMPORT_FUNCTION_ARRAY macro will be used in the function before
    this macro is used.

    ***** Human-readable representation *****

    overriddenDataIndex = 0;
    while( lineTokens != CSVIterator() ) {
        if( r_boolSetOverriddenFlags ) {
            overriddenDataArray[overriddenDataIndex] = true;
        }
        // It's expected that secondaryKeyIndex will be used in r_DatabaseTarget
        secondaryKeyIndex = r_SecondaryKeyIndex;
        primaryKeyListIterator = primaryKeyList.begin();
        for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) {
            curToken = (*lineTokens)[columnIndex];
            r_DatabaseTarget = r_Value;
            ++primaryKeyListIterator;
        }
        ++lineTokens;
        curToken = (*lineTokens)[0];
        overriddenDataIndex++;
    }

*/
#define READ_CSV_ROWS_ARRAY(r_DatabaseTarget, r_Value, r_SecondaryKeyIndex) \
    overriddenDataIndex = 0; \
    while( lineTokens != CSVIterator() ) { \
        if( r_boolSetOverriddenFlags ) { \
            overriddenDataArray[overriddenDataIndex] = true; \
        } \
        secondaryKeyIndex = r_SecondaryKeyIndex; \
        primaryKeyListIterator = primaryKeyList.begin(); \
        for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) { \
            curToken = (*lineTokens)[columnIndex]; \
            r_DatabaseTarget = r_Value; \
            ++primaryKeyListIterator; \
        } \
        ++lineTokens; \
        curToken = (*lineTokens)[0]; \
        overriddenDataIndex++; \
    }

/*
    READ_CSV_ROW_DUAL_KEY is a macro for reading a line from a CSV file in context of a dual-key data
    structure. For example, in the importActorsSkillLearning function, we are dealing with
    individual SkillProgression data structures within each DBActor, and the SkillProgression data
    structures have varying lengths. Hence we need to accommodate both the DBActor (primary key) and
    SkillProgression (secondary key) data structures.

    It is expected that the SETUP_IMPORT_FUNCTION_DUAL_KEY macro will be used in the function before
    this macro is used.

    ***** Human-readable representation *****

    rowIndex++;
    if( csvHeaderArray[rowIndex] == curToken ) {
        primaryKeyIdListIterator = primaryKeyIdList.begin();
        secondaryKeyIdListIterator = secondaryKeyIdList.begin();
        for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) {
            curToken = (*lineTokens)[columnIndex];
            r_DatabaseTarget = r_Value;
            ++primaryKeyIdListIterator;
            ++secondaryKeyIdListIterator;
        }
        ++lineTokens;
        curToken = (*lineTokens)[0];
    }
*/
#define READ_CSV_ROW_DUAL_KEY(r_DatabaseTarget, r_Value) \
    rowIndex++; \
    if( csvHeaderArray[rowIndex] == curToken ) { \
        primaryKeyIdListIterator = primaryKeyIdList.begin(); \
        secondaryKeyIdListIterator = secondaryKeyIdList.begin(); \
        for( columnIndex = 1; (*lineTokens).size() > columnIndex; columnIndex++ ) { \
            curToken = (*lineTokens)[columnIndex]; \
            r_DatabaseTarget = r_Value; \
            ++primaryKeyIdListIterator; \
            ++secondaryKeyIdListIterator; \
        } \
        ++lineTokens; \
        curToken = (*lineTokens)[0]; \
    }

//! Import the data in a .csv file as Actor database values
/*!
    importActors imports Actor database values (not including array values like HP, MP, etc.) from
    a .csv file, using a given file name. This function is used as a subroutine of importDatabase,
    and can also be called with the @dyndatabaseoverride_import_actors comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActors( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActors( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataArray = overriddenDataActor;       // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ACTOR;    // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_SIMPLE;
    // Get record values
    // Name
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->name, curToken.c_str() );
    // Degree
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->degree, curToken.c_str() );
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    // Charset file name
//    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->charsetFilename, curToken.c_str() );
//    // Charset ID
//    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->charsetId, std::atoi( curToken.c_str() ) );
//    // Charset transparent
//    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->charsetTransparent, std::atoi( curToken.c_str() ) );
    // Min level
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->minLevel, std::atoi( curToken.c_str() ) );
    // Max level
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->maxLevel, std::atoi( curToken.c_str() ) );
    // Allow critical hit
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->allowCriticalHit, std::atoi( curToken.c_str() ) );
    // Critical hit probability
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->criticalHitProbability, std::atoi( curToken.c_str() ) );
    // Removed because direct manipulation of these values in RM2K3 causes issues
//    // Faceset file name
//    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->facesetFilename, curToken );
//    // Faceset ID
//    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->facesetId, std::atoi( curToken.c_str() ) );
    // Two weapons
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->twoWeapons, std::atoi( curToken.c_str() ) );
    // AI control
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->aiControl, std::atoi( curToken.c_str() ) );
    // Lock equipment
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->lockEquipment, std::atoi( curToken.c_str() ) );
    // Mighty guard
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->mightyGuard, std::atoi( curToken.c_str() ) );
    // Exp primary
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->expPrimary, std::atoi( curToken.c_str() ) );
    // Exp secondary
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->expSecondary, std::atoi( curToken.c_str() ) );
    // Exp tertiary
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->expTertiary, std::atoi( curToken.c_str() ) );
    // Weapon ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->weaponId, std::atoi( curToken.c_str() ) );
    // Shield ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->shieldId, std::atoi( curToken.c_str() ) );
    // Armor ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->armorId, std::atoi( curToken.c_str() ) );
    // Helmet ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->helmetId, std::atoi( curToken.c_str() ) );
    // Accessory ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->accessoryId, std::atoi( curToken.c_str() ) );
    // Unarmed battle animation ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->unarmedBattleAnimationId, std::atoi( curToken.c_str() ) );
    // Class ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->classId, std::atoi( curToken.c_str() ) );
    // Battle graphic ID
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->battleGraphicId, std::atoi( curToken.c_str() ) );
    // X pos in battle
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->xPosInBattle, std::atoi( curToken.c_str() ) );
    // Y pos in battle
    READ_CSV_ROW_SIMPLE( RPG::dbActors[*primaryKeyListIterator]->yPosInBattle, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor HP database values
/*!
    importActorsHp imports Actor HP database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_actors_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsHp( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsHp( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataActorHp;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbActors[*primaryKeyListIterator]->maxHp[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor MP database values
/*!
    importActorsMp imports Actor MP database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_actors_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsMp( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsMp( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataActorMp;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbActors[*primaryKeyListIterator]->maxMp[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor Attack database values
/*!
    importActorsAttack imports Actor Attack database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_actors_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsAttack( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsAttack( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataActorAttack;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbActors[*primaryKeyListIterator]->attack[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor Defense database values
/*!
    importActorsDefense imports Actor Defense database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_actors_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsDefense( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsDefense( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataActorDefense;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbActors[*primaryKeyListIterator]->defense[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor Intelligence database values
/*!
    importActorsIntelligence imports Actor Intelligence database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_actors_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsIntelligence( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsIntelligence( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataActorIntelligence;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbActors[*primaryKeyListIterator]->intelligence[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor Agility database values
/*!
    importActorsAgility imports Actor Agility database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_actors_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsAgility( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsAgility( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataActorAgility;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbActors[*primaryKeyListIterator]->agility[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_ACTOR_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor Skill Learning database values
/*!
    importActors imports Actor Skill Learning database values from a .csv file, using a given file
    name. This function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_actors_skill_learning comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsSkillLearning( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsSkillLearning( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorSkillLearning;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_SKILL_LEARNING;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Level requirement
    READ_CSV_ROW_DUAL_KEY( RPG::dbActors[*primaryKeyIdListIterator]->skillLearning[*secondaryKeyIdListIterator]->levelReq, std::atoi( curToken.c_str() ) );
    // Skill ID
    READ_CSV_ROW_DUAL_KEY( RPG::dbActors[*primaryKeyIdListIterator]->skillLearning[*secondaryKeyIdListIterator]->skillId, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor Conditions database values
/*!
    importActorsConditions imports Actor Conditions database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_actors_conditions comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsConditions( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsConditions( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_CONDITION;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Condition susceptibility rank
    READ_CSV_ROW_DUAL_KEY( RPG::dbActors[*primaryKeyIdListIterator]->conditions[*secondaryKeyIdListIterator], csvTokenToDamageMultiplier[curToken] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Actor Attributes database values
/*!
    importActorsAttributes imports Actor Attributes database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_actors_attributes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importActorsAttributes( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importActorsAttributes( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataActorAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ACTOR_ATTRIBUTE;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Attribute susceptibility rank
    READ_CSV_ROW_DUAL_KEY( RPG::dbActors[*primaryKeyIdListIterator]->attributes[*secondaryKeyIdListIterator], csvTokenToDamageMultiplier[curToken] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class database values
/*!
    importClasses imports Class database values (not including array values like HP, MP, etc.) from
    a .csv file, using a given file name. This function is used as a subroutine of importDatabase,
    and can also be called with the @dyndatabaseoverride_import_classes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClasses( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClasses( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataArray = overriddenDataClass;       // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CLASS;    // Pointer to the array of header strings to use
    int secondaryKeyIndex;

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_SIMPLE;
    // Get record values
    // Name
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->name, curToken.c_str() );
    // Two weapons
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->twoWeapons, std::atoi( curToken.c_str() ) );
    // AI control
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->aiControl, std::atoi( curToken.c_str() ) );
    // Lock equipment
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->lockEquipment, std::atoi( curToken.c_str() ) );
    // Mighty guard
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->mightyGuard, std::atoi( curToken.c_str() ) );
    // Exp primary
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->expPrimary, std::atoi( curToken.c_str() ) );
    // Exp secondary
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->expSecondary, std::atoi( curToken.c_str() ) );
    // Exp tertiary
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->expTertiary, std::atoi( curToken.c_str() ) );
    // Battle graphic ID
    READ_CSV_ROW_SIMPLE( RPG::dbClasses[*primaryKeyListIterator]->battleGraphicId, std::atoi( curToken.c_str() ) );
    // Battle commands
    READ_CSV_ROWS_ARRAY(
        RPG::dbClasses[*primaryKeyListIterator]->battleCommands[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_CLASS_BATTLE_COMMAND_0].length(), curToken.length() - csvHeaderArray[CSV_INDEX_CLASS_BATTLE_COMMAND_0].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class HP database values
/*!
    importClassesHp imports Class HP database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_classes_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesHp( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesHp( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataClassHp;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbClasses[*primaryKeyListIterator]->maxHp[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class MP database values
/*!
    importClassesMp imports Class MP database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_classes_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesMp( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesMp( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataClassMp;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbClasses[*primaryKeyListIterator]->maxMp[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class Attack database values
/*!
    importClassesAttack imports Class Attack database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_classes_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesAttack( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesAttack( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataClassAttack;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbClasses[*primaryKeyListIterator]->attack[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class Defense database values
/*!
    importClassesDefense imports Class Defense database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_classes_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesDefense( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesDefense( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataClassDefense;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbClasses[*primaryKeyListIterator]->defense[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class Intelligence database values
/*!
    importClassesIntelligence imports Class Intelligence database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_classes_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesIntelligence( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesIntelligence( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataClassIntelligence;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbClasses[*primaryKeyListIterator]->intelligence[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class Agility database values
/*!
    importClassesAgility imports Class Agility database values from a .csv file, using a given file name. This
    function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_classes_hp comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesAgility( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesAgility( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );        // File stream used to get input data
    bool * overriddenDataArray = overriddenDataClassAgility;     // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CLASS_STAT;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_ARRAY;
    // Get record values
    // Stats at each level
    READ_CSV_ROWS_ARRAY(
        RPG::dbClasses[*primaryKeyListIterator]->agility[secondaryKeyIndex],
        std::atoi( curToken.c_str() ),
        std::atoi( curToken.substr( csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length(), curToken.length() - csvHeaderArray[CSV_INDEX_CLASS_STAT_LEVEL_1].length() ).c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class Skill Learning database values
/*!
    importClasses imports Class Skill Learning database values from a .csv file, using a given file
    name. This function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_classes_skill_learning comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesSkillLearning( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesSkillLearning( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataClassSkillLearning;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_CLASS_SKILL_LEARNING;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Level requirement
    READ_CSV_ROW_DUAL_KEY( RPG::dbClasses[*primaryKeyIdListIterator]->skillLearning[*secondaryKeyIdListIterator]->levelReq, std::atoi( curToken.c_str() ) );
    // Skill ID
    READ_CSV_ROW_DUAL_KEY( RPG::dbClasses[*primaryKeyIdListIterator]->skillLearning[*secondaryKeyIdListIterator]->skillId, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class Conditions database values
/*!
    importClassesConditions imports Class Conditions database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_classes_conditions comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesConditions( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesConditions( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataClassConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_CLASS_CONDITION;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Condition susceptibility rank
    READ_CSV_ROW_DUAL_KEY( RPG::dbClasses[*primaryKeyIdListIterator]->conditions[*secondaryKeyIdListIterator], csvTokenToDamageMultiplier[curToken] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Class Attributes database values
/*!
    importClassesAttributes imports Class Attributes database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_classes_attributes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importClassesAttributes( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importClassesAttributes( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataClassAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_CLASS_ATTRIBUTE;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Attribute susceptibility rank
    READ_CSV_ROW_DUAL_KEY( RPG::dbClasses[*primaryKeyIdListIterator]->attributes[*secondaryKeyIdListIterator], csvTokenToDamageMultiplier[curToken] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Skill database values
/*!
    importSkills imports Skill database values (not including array values like HP, MP, etc.) from
    a .csv file, using a given file name. This function is used as a subroutine of importDatabase,
    and can also be called with the @dyndatabaseoverride_import_skills comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importSkills( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importSkills( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataArray = overriddenDataSkill;       // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_SKILL;    // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_SIMPLE;
    // Get record values
    // Name
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->name, curToken.c_str() );
    // Description
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->description, curToken.c_str() );
    // Use message 2K
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->useMsg2k, curToken.c_str() );
    // Use message 2K alt
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->useMsg2kAlt, curToken.c_str() );
    // Evasion type
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->evasionType, std::atoi( curToken.c_str() ) );
    // Type
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->type, (RPG::SkillType) csvTokenToSkillType[curToken.c_str()] );
    // MP consumed
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->mpCost, std::atoi( curToken.c_str() ) );
    // Is MP percent
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->isMpPercent, std::atoi( curToken.c_str() ) );
    // MP consumed percent
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->mpCostPercent, std::atoi( curToken.c_str() ) );
    // Target
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->target, csvTokenToSkillTarget[curToken.c_str()] );
    // Switch ID
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->switchId, std::atoi( curToken.c_str() ) );
    // Battle animation ID
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->battleAnimId, std::atoi( curToken.c_str() ) );
    // Use in field
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->useInField, std::atoi( curToken.c_str() ) );
    // Use in battle
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->useInBattle, std::atoi( curToken.c_str() ) );
    // Attack influence
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->atkInfluence, std::atoi( curToken.c_str() ) );
    // Intelligence influence
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->intInfluence, std::atoi( curToken.c_str() ) );
    // Variance
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->variance, std::atoi( curToken.c_str() ) );
    // Effect rating
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->effectRating, std::atoi( curToken.c_str() ) );
    // Success rate
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->successRate, std::atoi( curToken.c_str() ) );
    // Affect HP
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->affectHp, std::atoi( curToken.c_str() ) );
    // Affect MP
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->affectMp, std::atoi( curToken.c_str() ) );
    // Affect strength (attack)
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->affectStr, std::atoi( curToken.c_str() ) );
    // Affect defense
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->affectDef, std::atoi( curToken.c_str() ) );
    // Affect intelligence
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->affectInt, std::atoi( curToken.c_str() ) );
    // Affect agility
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->affectAgl, std::atoi( curToken.c_str() ) );
    // Absorb
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->absorb, std::atoi( curToken.c_str() ) );
    // Ignore defense
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->ignoreDef, std::atoi( curToken.c_str() ) );
    // Attribute resist
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->attributeResist, std::atoi( curToken.c_str() ) );
    // Condition flag
    READ_CSV_ROW_SIMPLE( RPG::skills[*primaryKeyListIterator]->conditionFlag, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Skill Conditions database values
/*!
    importSkillsConditions imports Skill Conditions database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_skills_conditions comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importSkillsConditions( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importSkillsConditions( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataSkillConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_SKILL_CONDITION;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Condition susceptibility rank
    // RPG::Skill::conditions is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_DUAL_KEY( RPG::skills[*primaryKeyIdListIterator]->conditions[*secondaryKeyIdListIterator - 1], std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Skill Attributes database values
/*!
    importSkillsAttributes imports Skill Attributes database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_skills_attributes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importSkillsAttributes( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importSkillsAttributes( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataSkillAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_SKILL_ATTRIBUTE;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Attribute susceptibility rank
    // RPG::Skill::attributes is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_DUAL_KEY( RPG::skills[*primaryKeyIdListIterator]->attributes[*secondaryKeyIdListIterator - 1], std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Skill Animations database values
/*!
    importSkillsAnimations imports Skill Animations database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_skills_animations comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importSkillsAnimations( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importSkillsAnimations( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataSkillAnimations;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_SKILL_ANIMATION;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Battler pose
    READ_CSV_ROW_DUAL_KEY( RPG::skills[*primaryKeyIdListIterator]->skillAnims[*secondaryKeyIdListIterator]->battlerPose, (RPG::AnimationBattlerPose) csvTokenToAnimationBattlerPose[curToken.c_str()] );
    // Movement
    READ_CSV_ROW_DUAL_KEY( RPG::skills[*primaryKeyIdListIterator]->skillAnims[*secondaryKeyIdListIterator]->movement, csvTokenToAnimationMoveBeforeAttack[curToken.c_str()] );
    // Add after image
    READ_CSV_ROW_DUAL_KEY( RPG::skills[*primaryKeyIdListIterator]->skillAnims[*secondaryKeyIdListIterator]->addAfterImage, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Item database values
/*!
    importItems imports Item database values (not including array values like conditions,
    attributes, etc.) from a .csv file, using a given file name. This function is used as a
    subroutine of importDatabase, and can also be called with the @dyndatabaseoverride_import_items
    comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importItems( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importItems( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataArray = overriddenDataItem;       // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ITEM;    // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_SIMPLE;
    // Get record values
    // Name
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->name, curToken.c_str() );
    // Description
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->description, curToken.c_str() );
    // Type
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->type, (RPG::ItemType) csvTokenToItemType[curToken.c_str()] );
    // Price
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->price, std::atoi( curToken.c_str() ) );
    // Number of uses
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->numUses, std::atoi( curToken.c_str() ) );
    // Attack
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->str, std::atoi( curToken.c_str() ) );
    // Defense
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->def, std::atoi( curToken.c_str() ) );
    // Intelligence
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->intl, std::atoi( curToken.c_str() ) );
    // Agility
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->agl, std::atoi( curToken.c_str() ) );
    // Two-handed
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->twoHanded, std::atoi( curToken.c_str() ) );
    // MP consumed
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->mpCost, std::atoi( curToken.c_str() ) );
    // Hit probability
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->hitProb, std::atoi( curToken.c_str() ) );
    // Critical bonus
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->critBonus, std::atoi( curToken.c_str() ) );
    // Battle animation ID
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->battleAnimId, std::atoi( curToken.c_str() ) );
    // Always attack first
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->alwaysAttackFirst, std::atoi( curToken.c_str() ) );
    // Attack twice
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->attackTwice, std::atoi( curToken.c_str() ) );
    // Attack all
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->attackAll, std::atoi( curToken.c_str() ) );
    // Ignore evasion
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->ignoreEvasion, std::atoi( curToken.c_str() ) );
    // Prevent critical hits
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->preventCrit, std::atoi( curToken.c_str() ) );
    // Increased evasion
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->increaseEvasion, std::atoi( curToken.c_str() ) );
    // Half MP consumption
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->halfMpCost, std::atoi( curToken.c_str() ) );
    // No terrain damage
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->noTerrainDmg, std::atoi( curToken.c_str() ) );
    // Cursed
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->isCursed, std::atoi( curToken.c_str() ) );
    // Target
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->target, csvTokenToItemTarget[curToken.c_str()] );
    // HP restore percent
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->hpRestorePercent, std::atoi( curToken.c_str() ) );
    // HP restore units
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->hpRestoreUnits, std::atoi( curToken.c_str() ) );
    // MP restore percent
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->mpRestorePercent, std::atoi( curToken.c_str() ) );
    // MP restore units
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->mpRestoreUnits, std::atoi( curToken.c_str() ) );
    // Only usable on the field map
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->onlyInField, std::atoi( curToken.c_str() ) );
    // Only usable on dead characters
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->onlyOnDead, std::atoi( curToken.c_str() ) );
    // Add max HP
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->addMaxHp, std::atoi( curToken.c_str() ) );
    // Add max MP
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->addMaxMp, std::atoi( curToken.c_str() ) );
    // Add attack
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->addStr, std::atoi( curToken.c_str() ) );
    // Add defense
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->addDef, std::atoi( curToken.c_str() ) );
    // Add intelligence
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->addInt, std::atoi( curToken.c_str() ) );
    // Add agility
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->addAgl, std::atoi( curToken.c_str() ) );
    // Message shown when used
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->messageShown, (RPG::ItemScrollMsgShown) csvTokenToItemScrollMsgShown[curToken.c_str()] );
    // Invoke skill when used
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->invokeSkill, std::atoi( curToken.c_str() ) );
    // Skill ID
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->skillId, std::atoi( curToken.c_str() ) );
    // Switch ID
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->switchId, std::atoi( curToken.c_str() ) );
    // Activation allowed in field
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->useInField, std::atoi( curToken.c_str() ) );
    // Activation allowed in battle
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->useInBattle, std::atoi( curToken.c_str() ) );
    // Condition effectiveness
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->conditionPercentage, std::atoi( curToken.c_str() ) );
    // Condition resist or inflict
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->conditionFlag, (RPG::ItemConditionFlag) csvTokenToItemConditionFlag[curToken.c_str()] );
    // Weapon trajectory
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->weaponTrajectory, (RPG::WeaponRangedTrajectory) csvTokenToWeaponRangedTrajectory[curToken.c_str()] );
    // Weapon targetting
    READ_CSV_ROW_SIMPLE( RPG::items[*primaryKeyListIterator]->weaponTargetting, (RPG::WeaponRangedTargetting) csvTokenToWeaponRangedTargetting[curToken.c_str()] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Item Usable By Actor database values
/*!
    importItemsUsableByActors imports Item Usable By Actor database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_items_usable_by_actors comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importItemsUsableByActors( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importItemsUsableByActors( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemUsableByActors;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_USABLE_BY_ACTOR;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Actor can use
    // RPG::Item::usableByHero is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->usableByHero[*secondaryKeyIdListIterator - 1], std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Item Usable By Classes database values
/*!
    importItemsUsableByClasses imports Item Usable By Classes database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_items_usable_by_classes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importItemsUsableByClasses( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importItemsUsableByClasses( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemUsableByClasses;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_USABLE_BY_CLASS;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Class can use
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->usableByClass[*secondaryKeyIdListIterator], std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Item Conditions database values
/*!
    importItemsConditions imports Item Conditions database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_items_conditions comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importItemsConditions( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importItemsConditions( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_CONDITION;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Condition susceptibility rank
    // RPG::Item::conditions is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->conditions[*secondaryKeyIdListIterator - 1], std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Item Attributes database values
/*!
    importItemsAttributes imports Item Attributes database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_items_attributes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importItemsAttributes( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importItemsAttributes( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_ATTRIBUTE;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Attribute susceptibility rank
    // RPG::Item::attributes is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->attributes[*secondaryKeyIdListIterator - 1], std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Item Weapon Animations database values
/*!
    importItemsWeaponAnimations imports Item Weapon Animations database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_items_weapon_animations comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importItemsWeaponAnimations( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importItemsAnimations( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataItemWeaponAnimations;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ITEM_WEAPON_ANIMATION;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Battler pose
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->battlerPose, (RPG::AnimationBattlerPose) csvTokenToAnimationBattlerPose[curToken.c_str()] );
    // Animation type
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->isBattleAnim, std::atoi( curToken.c_str() ) );
    // Weapon animation ID
    // RPG::AnimationWeapon::battleAnimId is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->battleAnimId, std::atoi( curToken.c_str() ) - 1 );
    // Movement before attack
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->movement, csvTokenToAnimationMoveBeforeAttack[curToken.c_str()] );
    // Add after image
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->addAfterImage, std::atoi( curToken.c_str() ) );
    // Number of attacks
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->numAttacks, csvTokenToWeaponNumberOfAttacks[curToken.c_str()] );
    // Ranged weapon
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->isRangedWeapon, std::atoi( curToken.c_str() ) );
    // Ranged weapon animation ID
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->rangedAnimId, std::atoi( curToken.c_str() ) );
    // Ranged weapon animation speed
    READ_CSV_ROW_DUAL_KEY( RPG::items[*primaryKeyIdListIterator]->weaponAnims[*secondaryKeyIdListIterator]->rangedAnimSpeed, csvTokenToWeaponRangedAnimSpeed[curToken.c_str()] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Monster database values
/*!
    importEnemies imports Monster database values (not including array values like Attributes,
    Conditions, etc.) from a .csv file, using a given file name. This function is used as a
    subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_enemies comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importEnemies( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importEnemies( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataArray = overriddenDataEnemy;       // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ENEMY;    // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_SIMPLE;
    // Get record values
    // Name
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->name, curToken.c_str() );
    // File name
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->filename, curToken.c_str() );
    // Color alteration
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->colorAlteration, std::atoi( curToken.c_str() ) );
    // Is transparent
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->isTransparent, std::atoi( curToken.c_str() ) );
    // Is flying
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->isFlying, std::atoi( curToken.c_str() ) );
    // Max HP
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->maxHp, std::atoi( curToken.c_str() ) );
    // Max MP
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->maxMp, std::atoi( curToken.c_str() ) );
    // Attack
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->attack, std::atoi( curToken.c_str() ) );
    // Defense
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->defense, std::atoi( curToken.c_str() ) );
    // Intelligence
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->intelligence, std::atoi( curToken.c_str() ) );
    // Agility
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->agility, std::atoi( curToken.c_str() ) );
    // Experience
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->exp, std::atoi( curToken.c_str() ) );
    // Money
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->gold, std::atoi( curToken.c_str() ) );
    // Item ID
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->itemId, std::atoi( curToken.c_str() ) );
    // Item probability
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->itemProbability, std::atoi( curToken.c_str() ) );
    // Allow critical hit
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->critFlag, std::atoi( curToken.c_str() ) );
    // Critical hit probability
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->critPercentage, std::atoi( curToken.c_str() ) );
    // Often miss
    READ_CSV_ROW_SIMPLE( RPG::dbMonsters[*primaryKeyListIterator]->oftenMiss, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Monster Conditions database values
/*!
    importEnemiesConditions imports Monster Conditions database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_enemies_conditions comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importEnemiesConditions( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importEnemiesConditions( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataEnemyConditions;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ENEMY_CONDITION;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Condition susceptibility rank
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->conditions[*secondaryKeyIdListIterator], csvTokenToDamageMultiplier[curToken] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Monster Attributes database values
/*!
    importEnemiesAttributes imports Monster Attributes database values from a .csv file, using a given
    file name. This function is used as a subroutine of importDatabase, and can also be called with
    the @dyndatabaseoverride_import_enemies_attributes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importEnemiesAttributes( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importEnemiesAttributes( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataEnemyAttributes;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ENEMY_ATTRIBUTE;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Attribute susceptibility rank
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->attributes[*secondaryKeyIdListIterator], csvTokenToDamageMultiplier[curToken] );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Monster Behavior database values
/*!
    importMonsterBehaviors imports Monster Behavior database values from a .csv file, using a given file
    name. This function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_enemies_behaviors comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importEnemiesBehaviors( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importEnemiesBehaviors( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    std::set< std::pair<int, int> > * overriddenDataStructure = &overriddenDataEnemyBehaviors;   // Pointer to the overridden data structure to use
    std::string * csvHeaderArray = CSV_HEADER_ENEMY_BEHAVIOR;   // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_DUAL_KEY;
    // Get record values
    // Precondition
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->precondition, csvTokenToMonsterPrecondition[curToken] );
    // Precondition parameter 1
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->preconditionParam[0], std::atoi( curToken.c_str() ) );
    // Precondition parameter 2
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->preconditionParam[1], std::atoi( curToken.c_str() ) );
    // Precondition switch ID
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->preconditionSwitchId, std::atoi( curToken.c_str() ) );
    // Priority
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->priority, std::atoi( curToken.c_str() ) );
    // Switch on
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->switchOn, std::atoi( curToken.c_str() ) );
    // Switch on ID
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->switchOnId, std::atoi( curToken.c_str() ) );
    // Switch off
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->switchOff, std::atoi( curToken.c_str() ) );
    // Switch off ID
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->switchOffId, std::atoi( curToken.c_str() ) );
    // Action kind
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->kind, csvTokenToActionKind[curToken] );
    // Basic action type
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->basicActionId, csvTokenToBasicAction[curToken] );
    // Skill ID
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->skillId, std::atoi( curToken.c_str() ) );
    // Monster ID (for transformation)
    READ_CSV_ROW_DUAL_KEY( RPG::dbMonsters[*primaryKeyIdListIterator]->behavior[*secondaryKeyIdListIterator]->monsterId, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Attribute database values
/*!
    importAttributes imports Attribute database values from a .csv file, using a given file name.
    This function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_attributes comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importAttributes( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importAttributes( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataArray = overriddenDataAttributes;       // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_ATTRIBUTE;    // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_SIMPLE;
    // Get record values
    // Name
    READ_CSV_ROW_SIMPLE( RPG::attributes[*primaryKeyListIterator]->name, curToken.c_str() );
    // Type
    READ_CSV_ROW_SIMPLE( RPG::attributes[*primaryKeyListIterator]->type, csvTokenToAlterationType[curToken.c_str()] );
    // Damage multiplier A
    READ_CSV_ROW_SIMPLE( RPG::attributes[*primaryKeyListIterator]->dmgA, std::atoi( curToken.c_str() ) );
    // Damage multiplier B
    READ_CSV_ROW_SIMPLE( RPG::attributes[*primaryKeyListIterator]->dmgB, std::atoi( curToken.c_str() ) );
    // Damage multiplier C
    READ_CSV_ROW_SIMPLE( RPG::attributes[*primaryKeyListIterator]->dmgC, std::atoi( curToken.c_str() ) );
    // Damage multiplier D
    READ_CSV_ROW_SIMPLE( RPG::attributes[*primaryKeyListIterator]->dmgD, std::atoi( curToken.c_str() ) );
    // Damage multiplier E
    READ_CSV_ROW_SIMPLE( RPG::attributes[*primaryKeyListIterator]->dmgE, std::atoi( curToken.c_str() ) );
    // Close the input file
    inputStream.close();
}

//! Import the data in a .csv file as Condition database values
/*!
    importConditions imports Condition database values from a .csv file, using a given file name.
    This function is used as a subroutine of importDatabase, and can also be called with the
    @dyndatabaseoverride_import_conditions comment command.

    \param r_FileName: (std::string) Name of the .csv file to get data from
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
*/
void importConditions( std::string r_FileName, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importConditions( " << r_FileName << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    std::ifstream inputStream( r_FileName.c_str() );    // File stream used to get input data
    bool * overriddenDataArray = overriddenDataConditions;       // Pointer to the relevant overridden data flag array
    std::string * csvHeaderArray = CSV_HEADER_CONDITION;    // Pointer to the array of header strings to use

    // Check whether the file opened successfully
    if( !inputStream.is_open() ) {
        #ifdef DYNDATABASEOVERRIDE_DEBUG
        *ofDebug << "File " << r_FileName << " not found\n";
        #endif // DYNDATABASEOVERRIDE_DEBUG
        return;
    }
    SETUP_IMPORT_FUNCTION_SIMPLE;
    // Get record values
    // Name
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->name, curToken.c_str() );
    // Persists after battle
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->persistsAfterBattle, std::atoi( curToken.c_str() ) );
    // Color
    // RPG::Condition::color is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->color, std::atoi( curToken.c_str() ) - 1 );
    // Priority
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->priority, std::atoi( curToken.c_str() ) );
    // Action restriction
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->actionRestriction, csvTokenToActionRestriction[curToken.c_str()] );
    // Hit percentage
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->hitPercentage, std::atoi( curToken.c_str() ) );
    // Evade all attacks
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->evadeAttacks, std::atoi( curToken.c_str() ) );
    // Reflect skills
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->reflectSkills, std::atoi( curToken.c_str() ) );
    // Equipment cannot be changed
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->lockEquipment, std::atoi( curToken.c_str() ) );
    // Susceptibility A
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->susA, std::atoi( curToken.c_str() ) );
    // Susceptibility B
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->susB, std::atoi( curToken.c_str() ) );
    // Susceptibility C
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->susC, std::atoi( curToken.c_str() ) );
    // Susceptibility D
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->susD, std::atoi( curToken.c_str() ) );
    // Susceptibility E
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->susE, std::atoi( curToken.c_str() ) );
    // Recovery turns
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->recoveryTurns, std::atoi( curToken.c_str() ) );
    // Recovery turns percent
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->recoveryTurnsPercent, std::atoi( curToken.c_str() ) );
    // Recovery physical percent
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->recoveryPhysicalPercent, std::atoi( curToken.c_str() ) );
    // Base stat alteration
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->statAlteration, csvTokenToBaseStatAlteration[curToken.c_str()] );
    // Alters attack
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->alterStr, std::atoi( curToken.c_str() ) );
    // Alters defense
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->alterDef, std::atoi( curToken.c_str() ) );
    // Alters intelligence
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->alterInt, std::atoi( curToken.c_str() ) );
    // Alters agility
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->alterAgi, std::atoi( curToken.c_str() ) );
    // Prevent skill by attack
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->preventSkillByAtk, std::atoi( curToken.c_str() ) );
    // Prevent skill attack influence
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->atkInfluence, std::atoi( curToken.c_str() ) );
    // Prevent skill by intelligence
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->preventSkillByInt, std::atoi( curToken.c_str() ) );
    // Prevent skill intelligence influence
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->intInfluence, std::atoi( curToken.c_str() ) );
    // HP alteration type
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->hpAlterationType, csvTokenToAlterationType[curToken.c_str()] );
    // HP damage percent
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->hpDamagePercentage, std::atoi( curToken.c_str() ) );
    // HP damage units
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->hpDamageUnits, std::atoi( curToken.c_str() ) );
    // HP damage on step
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->hpDamageEachStep, std::atoi( curToken.c_str() ) );
    // HP damage step number
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->hpDamageStepNum, std::atoi( curToken.c_str() ) );
    // MP alteration type
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->mpAlterationType, csvTokenToAlterationType[curToken.c_str()] );
    // MP damage percent
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->mpDamagePercentage, std::atoi( curToken.c_str() ) );
    // MP damage units
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->mpDamageUnits, std::atoi( curToken.c_str() ) );
    // MP damage on step
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->mpDamageEachStep, std::atoi( curToken.c_str() ) );
    // MP damage step number
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->mpDamageStepNum, std::atoi( curToken.c_str() ) );
    // Animation ID
    // RPG::Condition::poseId is 0-based even though the thing it refers to conceptually has 1-based indexing
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->poseId, std::atoi( curToken.c_str() ) - 1 );
    // Message 2K ally state
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->msg2kAllyState, curToken.c_str() );
    // Message 2K enemy state
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->msg2kEnemyState, curToken.c_str() );
    // Message 2K already in
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->msg2kAlreadyIn, curToken.c_str() );
    // Message 2K regular
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->msg2kRegular, curToken.c_str() );
    // Message 2K recovery
    READ_CSV_ROW_SIMPLE( RPG::conditions[*primaryKeyListIterator]->msg2kRecovery, curToken.c_str() );
    // Close the input file
    inputStream.close();
}

//! Import the database values from .csv files
/*!
    importDatabase imports database values from a collection of .csv files, using a given file name
    base for the file names. This function is used to import values for a new game or loaded
    savestate, and can also be called with the @dyndatabaseoverride_import_database comment command.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_boolSetOverriddenFlags: (bool) Whether to set the overridden data flags for overridden data to true
    \sa importStartingDatabase
    \sa importDefaultDatabase
    \sa onNewGame
    \sa onLoadGame
*/
void importDatabase( std::string r_FileNameBase, bool r_boolSetOverriddenFlags )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "importDatabase( " << r_FileNameBase << ", " << r_boolSetOverriddenFlags << " ) called\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    // Call all relevant database import functions
    std::string fileName = r_FileNameBase + ACTOR_DATABASE_FILE_NAME_END;
    importActors( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_HP_DATABASE_FILE_NAME_END;
    importActorsHp( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_MP_DATABASE_FILE_NAME_END;
    importActorsMp( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_ATTACK_DATABASE_FILE_NAME_END;
    importActorsAttack( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_DEFENSE_DATABASE_FILE_NAME_END;
    importActorsDefense( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_INTELLIGENCE_DATABASE_FILE_NAME_END;
    importActorsIntelligence( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_AGILITY_DATABASE_FILE_NAME_END;
    importActorsAgility( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_SKILL_LEARNING_DATABASE_FILE_NAME_END;
    importActorsSkillLearning( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_CONDITIONS_DATABASE_FILE_NAME_END;
    importActorsConditions( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_ATTRIBUTES_DATABASE_FILE_NAME_END;
    importActorsAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ACTOR_ATTRIBUTES_DATABASE_FILE_NAME_END;
    importActorsAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_DATABASE_FILE_NAME_END;
    importClasses( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_HP_DATABASE_FILE_NAME_END;
    importClassesHp( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_MP_DATABASE_FILE_NAME_END;
    importClassesMp( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_ATTACK_DATABASE_FILE_NAME_END;
    importClassesAttack( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_DEFENSE_DATABASE_FILE_NAME_END;
    importClassesDefense( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_INTELLIGENCE_DATABASE_FILE_NAME_END;
    importClassesIntelligence( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_AGILITY_DATABASE_FILE_NAME_END;
    importClassesAgility( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_SKILL_LEARNING_DATABASE_FILE_NAME_END;
    importClassesSkillLearning( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_CONDITIONS_DATABASE_FILE_NAME_END;
    importClassesConditions( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_ATTRIBUTES_DATABASE_FILE_NAME_END;
    importClassesAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CLASS_ATTRIBUTES_DATABASE_FILE_NAME_END;
    importClassesAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + SKILL_DATABASE_FILE_NAME_END;
    importSkills( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + SKILL_CONDITIONS_DATABASE_FILE_NAME_END;
    importSkillsConditions( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + SKILL_ATTRIBUTES_DATABASE_FILE_NAME_END;
    importSkillsAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + SKILL_ANIMATIONS_DATABASE_FILE_NAME_END;
    importSkillsAnimations( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ITEM_DATABASE_FILE_NAME_END;
    importItems( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ITEM_USABLE_BY_ACTORS_DATABASE_FILE_NAME_END;
    importItemsUsableByActors( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ITEM_USABLE_BY_CLASSES_DATABASE_FILE_NAME_END;
    importItemsUsableByClasses( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ITEM_CONDITIONS_DATABASE_FILE_NAME_END;
    importItemsConditions( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ITEM_ATTRIBUTES_DATABASE_FILE_NAME_END;
    importItemsAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ITEM_WEAPON_ANIMATIONS_DATABASE_FILE_NAME_END;
    importItemsWeaponAnimations( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ENEMY_DATABASE_FILE_NAME_END;
    importEnemies( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ENEMY_CONDITIONS_DATABASE_FILE_NAME_END;
    importEnemiesConditions( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ENEMY_ATTRIBUTES_DATABASE_FILE_NAME_END;
    importEnemiesAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ENEMY_BEHAVIORS_DATABASE_FILE_NAME_END;
    importEnemiesBehaviors( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + ATTRIBUTE_DATABASE_FILE_NAME_END;
    importAttributes( fileName, r_boolSetOverriddenFlags );
    fileName = r_FileNameBase + CONDITION_DATABASE_FILE_NAME_END;
    importConditions( fileName, r_boolSetOverriddenFlags );
}

//! Import the starting database values from .csv files
/*!
    importStartingDatabase imports the starting database values from a collection of .csv files.
    This function is used to set the database to the desired starting values shortly after the
    program starts. The database values changed by this function are NOT counted as being
    "overridden" for purposes of recording which values need to be exported in association with a
    game save state, because they are applied to all new games and restored games alike (although
    they may be overridden later by game save state restoration). The file names used by this
    function follow a convention wherein they have a start supplied in the DynRPG.ini configuration
    followed by a standardized ending based on the type of data they hold.
*/
void importStartingDatabase()
{
    importDatabase( CSV_FILE_PATH + startingDatabaseFileNameBase, false );
}

//! Import the default database values from .csv files
/*!
    importDefaultDatabase imports the default database values from a collection of .csv files.
    This function is used to set the database to the default values before starting a new game or
    loading a savestate. The database values changed by this function are NOT counted as being
    "overridden" for purposes of recording which values need to be exported in association with a
    game save state, because they are applied to all new games and restored games alike (although
    they may be overridden later by game save state restoration). The file names used by this
    function follow a convention wherein they have a start supplied by
    DEFAULT_DATABASE_FILE_NAME_BASE followed by a standardized ending based on the type of data they
    hold.

    /sa onNewGame
    /sa onLoadGame
*/
void importDefaultDatabase()
{
    importDatabase( CSV_FILE_PATH + DEFAULT_DATABASE_FILE_NAME_BASE, false );
}

//! Records the default database values at game start
/*!
    recordDefaultDatabase exports the initial database to a collection of .csv files when the game
    is first started, so that they can be used to restore the database to a baseline. This function
    is expected to be called from onInitFinished.

    \param r_FileNameBase: (std::string) Base to use for file names
    \param r_ExportAll: (bool) Whether to export all database data rather than just the parts that have been modified by the plugin
*/
void recordDefaultDatabase()
{
    // Call exportDatabase with the default database file name base and true for export all
    exportDatabase( CSV_FILE_PATH + DEFAULT_DATABASE_FILE_NAME_BASE, true );
}

//! Sets all values in a bool array to false
/*!
    boolArraySetFalse sets all the values in a given bool array to false.

    \param r_boolArray: (bool *) The bool array to set all values to false
    \param r_intArraySize: (int) The size of the array
    \sa clearOverriddenDataFlags
*/
void boolArraySetFalse( bool * r_boolArray, int r_intArraySize )
{
    for( int i = 0; i < r_intArraySize; i++ ) {
        r_boolArray[i] = false;
    }
}

//! Clears the overridden data flags
/*!
    clearOverriddenDataFlags clears the overridden data flag data structures. This is necessary
    anytime the game is considered to be entering a different savestate, such as a new game or load
    game.

    \sa onNewGame
    \sa onLoadGame
*/
void clearOverriddenDataFlags()
{
    // Set all overridden data flag data structures to all false or empty
    boolArraySetFalse( overriddenDataActor, CSV_ROWS_ACTOR - 2 );
    boolArraySetFalse( overriddenDataActorHp, CSV_ROWS_ACTOR_STAT - 2 );
    boolArraySetFalse( overriddenDataActorMp, CSV_ROWS_ACTOR_STAT - 2 );
    boolArraySetFalse( overriddenDataActorAttack, CSV_ROWS_ACTOR_STAT - 2 );
    boolArraySetFalse( overriddenDataActorDefense, CSV_ROWS_ACTOR_STAT - 2 );
    boolArraySetFalse( overriddenDataActorIntelligence, CSV_ROWS_ACTOR_STAT - 2 );
    boolArraySetFalse( overriddenDataActorAgility, CSV_ROWS_ACTOR_STAT - 2 );
    overriddenDataActorSkillLearning.clear();
    overriddenDataActorConditions.clear();
    overriddenDataActorAttributes.clear();
    boolArraySetFalse( overriddenDataClass, CSV_ROWS_CLASS - 2 );
    boolArraySetFalse( overriddenDataClassHp, CSV_ROWS_CLASS_STAT - 2 );
    boolArraySetFalse( overriddenDataClassMp, CSV_ROWS_CLASS_STAT - 2 );
    boolArraySetFalse( overriddenDataClassAttack, CSV_ROWS_CLASS_STAT - 2 );
    boolArraySetFalse( overriddenDataClassDefense, CSV_ROWS_CLASS_STAT - 2 );
    boolArraySetFalse( overriddenDataClassIntelligence, CSV_ROWS_CLASS_STAT - 2 );
    boolArraySetFalse( overriddenDataClassAgility, CSV_ROWS_CLASS_STAT - 2 );
    overriddenDataClassSkillLearning.clear();
    overriddenDataClassConditions.clear();
    overriddenDataClassAttributes.clear();
    boolArraySetFalse( overriddenDataSkill, CSV_ROWS_SKILL - 2 );
    overriddenDataSkillConditions.clear();
    overriddenDataSkillAttributes.clear();
    overriddenDataSkillAnimations.clear();
    boolArraySetFalse( overriddenDataItem, CSV_ROWS_ITEM - 2 );
    overriddenDataItemUsableByActors.clear();
    overriddenDataItemUsableByClasses.clear();
    overriddenDataItemConditions.clear();
    overriddenDataItemAttributes.clear();
    overriddenDataItemWeaponAnimations.clear();
    boolArraySetFalse( overriddenDataEnemy, CSV_ROWS_ENEMY - 2 );
    overriddenDataEnemyConditions.clear();
    overriddenDataEnemyAttributes.clear();
    overriddenDataEnemyBehaviors.clear();
    boolArraySetFalse( overriddenDataAttributes, CSV_ROWS_ATTRIBUTE - 2 );
    boolArraySetFalse( overriddenDataConditions, CSV_ROWS_CONDITION - 2 );
}

//! Translate a string into a set of ints
/*!
    stringToIntSet takes a string, which is assumed to be data from a CSV file representing a
    set of integer values, and returns a std::set<int> containing those integer values.

    \param r_DataString : (std::string) String of input data
    \return (std::set<int>) The set of integer values represented by the input string
*/
std::set<int> stringToIntSet( std::string r_DataString )
{
    static std::string dataString;      // Working string for the data
    static std::set<int> dataSet;       // Set of integer values to return
    static unsigned int tokenPos;       // End position of the current token in the data string

    // Clear the working data set
    dataSet.clear();
    // Copy the input string to a working string
    dataString = r_DataString;
    // Strip the "" from the ends of the string if they're there
    if( '"' == dataString[0] )
    {   // " detected
        dataString = dataString.substr( 1, dataString.length() - 2 );
    }
    // Loop through all tokens in the data string
    for( tokenPos = dataString.find_first_of( ';' ); std::string::npos != tokenPos; tokenPos = dataString.find_first_of( ';' ) )
    {
        // Add the number of the current token to the data set
        dataSet.insert( std::atoi( ( dataString.substr( 0, tokenPos ) ).c_str() ) );
        // Remove the current token from the data string
        dataString.erase( 0, tokenPos + 1 );
    }
    // Add the final token (which does not have an associated ;) to the data set
    dataSet.insert( std::atoi( dataString.c_str() ) );
    // Return the data set
    return dataSet;
}

//! Called after the RPG objects were initialized
/*!
    onInitFinished is called when the objects of the RPG class are finished initializing. In this
    plugin this method is used to perform plugin initialization.

    \param pluginName : ( char * ) Name of the plugin
*/
void onInitFinished()
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    ofDebug = new std::ofstream();
    ofDebug->open( "DynPlugins\\DynDatabaseOverrideDebug.txt" );
    #endif // DYNDATABASEOVERRIDE_DEBUG

    // Initialize CSV token maps
    damageMultiplierToCsvToken[RPG::DMG_A] = "A";
    damageMultiplierToCsvToken[RPG::DMG_B] = "B";
    damageMultiplierToCsvToken[RPG::DMG_C] = "C";
    damageMultiplierToCsvToken[RPG::DMG_D] = "D";
    damageMultiplierToCsvToken[RPG::DMG_E] = "E";
    for( std::map<RPG::DamageMultiplier_T, std::string>::iterator iter = damageMultiplierToCsvToken.begin(); iter != damageMultiplierToCsvToken.end(); ++iter ) {
        csvTokenToDamageMultiplier[iter->second] = iter->first;
    }
    actionKindToCsvToken[RPG::AK_BASIC] = "Basic";
    actionKindToCsvToken[RPG::AK_SKILL] = "Skill";
    actionKindToCsvToken[RPG::AK_TRANSFORM] = "Transform";
    actionKindToCsvToken[RPG::AK_ITEM] = "Item";
    actionKindToCsvToken[RPG::AK_NONE] = "None";
    actionKindToCsvToken[RPG::AK_RANDOM] = "Random";
    for( std::map<RPG::ActionKind_T, std::string>::iterator iter = actionKindToCsvToken.begin(); iter != actionKindToCsvToken.end(); ++iter ) {
        csvTokenToActionKind[iter->second] = iter->first;
    }
    basicActionToCsvToken[RPG::BA_ATTACK] = "Attack";
    basicActionToCsvToken[RPG::BA_DOUBLE_ATTACK] = "Double Attack";
    basicActionToCsvToken[RPG::BA_DEFEND] = "Defend";
    basicActionToCsvToken[RPG::BA_OBSERVE] = "Observe Battle";
    basicActionToCsvToken[RPG::BA_CHARGE] = "Charge Up";
    basicActionToCsvToken[RPG::BA_SELF_DESTRUCT] = "Self-Destruct";
    basicActionToCsvToken[RPG::BA_ESCAPE] = "Escape";
    basicActionToCsvToken[RPG::BA_NONE] = "No Action";
    basicActionToCsvToken[RPG::BA_CHANGE_ROW] = "Change Row";
    for( std::map<RPG::BasicAction_T, std::string>::iterator iter = basicActionToCsvToken.begin(); iter != basicActionToCsvToken.end(); ++iter ) {
        csvTokenToBasicAction[iter->second] = iter->first;
    }
    skillTypeToCsvToken[RPG::SKILL_NORMAL] = "Normal";
    skillTypeToCsvToken[RPG::SKILL_TELEPORT] = "Teleport";
    skillTypeToCsvToken[RPG::SKILL_ESCAPE] = "Escape";
    skillTypeToCsvToken[RPG::SKILL_SWITCH] = "Switch";
    for( int i = RPG::SKILL_SWITCH + 1; i < NUM_SKILL_TYPE; i++ ) {
        char buffer [2];
        std::string csvToken = "Custom Skill Type ";
        csvToken.append( itoa( i, buffer, 10 ) );
        skillTypeToCsvToken[i] = csvToken;
    }
    for( std::map<int, std::string>::iterator iter = skillTypeToCsvToken.begin(); iter != skillTypeToCsvToken.end(); ++iter ) {
        csvTokenToSkillType[iter->second] = iter->first;
    }
    skillTargetToCsvToken[RPG::SKILL_TARGET_ENEMY] = "Single enemy";
    skillTargetToCsvToken[RPG::SKILL_TARGET_ALL_ENEMIES] = "All enemies";
    skillTargetToCsvToken[RPG::SKILL_TARGET_SELF] = "Self";
    skillTargetToCsvToken[RPG::SKILL_TARGET_ALLY] = "Single ally";
    skillTargetToCsvToken[RPG::SKILL_TARGET_ALL_ALLIES] = "All allies";
    for( std::map<RPG::SkillTarget, std::string>::iterator iter = skillTargetToCsvToken.begin(); iter != skillTargetToCsvToken.end(); ++iter ) {
        csvTokenToSkillTarget[iter->second] = iter->first;
    }
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_STAND] = "Idle Stance";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_PUNCH_RIGHT] = "Right Hand";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_PUNCH_LEFT] = "Left Hand";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_SKILL] = "Skill Use";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_DEAD] = "Dead";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_DAMAGE] = "Damage";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_DAZED] = "Bad Status";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_BLOCK] = "Defending";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_RUN_LEFT] = "Walking Left";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_RUN_RIGHT] = "Walking Right";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_VICTORY] = "Victory";
    animationBattlerPoseToCsvToken[RPG::ANIM_POSE_ITEM] = "Item";
    for( int i = RPG::ANIM_POSE_ITEM + 1; i < NUM_ANIM_BATTLER_POSE; i++ ) {
        char buffer [2];
        std::string csvToken = "Custom Pose ";
        csvToken.append( itoa( i, buffer, 10 ) );
        animationBattlerPoseToCsvToken[i] = csvToken;
    }
    for( std::map<int, std::string>::iterator iter = animationBattlerPoseToCsvToken.begin(); iter != animationBattlerPoseToCsvToken.end(); ++iter ) {
        csvTokenToAnimationBattlerPose[iter->second] = iter->first;
    }
    animationMoveBeforeAttackToCsvToken[RPG::ANIM_MOVE_BEFORE_ACTION_NONE] = "None";
    animationMoveBeforeAttackToCsvToken[RPG::ANIM_MOVE_BEFORE_ACTION_STEP_FORWARD] = "Step Forward";
    animationMoveBeforeAttackToCsvToken[RPG::ANIM_MOVE_BEFORE_ACTION_JUMP_FORWARD] = "Jump Forward";
    animationMoveBeforeAttackToCsvToken[RPG::ANIM_MOVE_BEFORE_ACTION_MOVE_TO_TARGET] = "Move to Target";
    for( std::map<RPG::AnimationMoveBeforeAttack, std::string>::iterator iter = animationMoveBeforeAttackToCsvToken.begin(); iter != animationMoveBeforeAttackToCsvToken.end(); ++iter ) {
        csvTokenToAnimationMoveBeforeAttack[iter->second] = iter->first;
    }
    weaponNumberOfAttacksToCsvToken[RPG::WEAPON_NUM_ATTACKS_ONCE] = "Once";
    weaponNumberOfAttacksToCsvToken[RPG::WEAPON_NUM_ATTACKS_TWICE] = "Twice";
    weaponNumberOfAttacksToCsvToken[RPG::WEAPON_NUM_ATTACKS_THREE_TIMES] = "Three times";
    for( std::map<RPG::WeaponNumberOfAttacks, std::string>::iterator iter = weaponNumberOfAttacksToCsvToken.begin(); iter != weaponNumberOfAttacksToCsvToken.end(); ++iter ) {
        csvTokenToWeaponNumberOfAttacks[iter->second] = iter->first;
    }
    weaponRangedAnimSpeedToCsvToken[RPG::WEAPON_RANGED_ANIM_FAST] = "Fast";
    weaponRangedAnimSpeedToCsvToken[RPG::WEAPON_RANGED_ANIM_MEDIUM] = "Medium";
    weaponRangedAnimSpeedToCsvToken[RPG::WEAPON_RANGED_ANIM_SLOW] = "Slow";
    for( std::map<RPG::WeaponRangedAnimSpeed, std::string>::iterator iter = weaponRangedAnimSpeedToCsvToken.begin(); iter != weaponRangedAnimSpeedToCsvToken.end(); ++iter ) {
        csvTokenToWeaponRangedAnimSpeed[iter->second] = iter->first;
    }
    weaponRangedTrajectoryToCsvToken[RPG::WEAPON_RANGED_TRAJ_FLY_TO_TARGET] = "Fly Straight to Target";
    weaponRangedTrajectoryToCsvToken[RPG::WEAPON_RANGED_TRAJ_RETURN_TO_HERO] = "Return to User After Striking";
    for( std::map<RPG::WeaponRangedTrajectory, std::string>::iterator iter = weaponRangedTrajectoryToCsvToken.begin(); iter != weaponRangedTrajectoryToCsvToken.end(); ++iter ) {
        csvTokenToWeaponRangedTrajectory[iter->second] = iter->first;
    }
    weaponRangedTargettingToCsvToken[RPG::WEAPON_RANGED_TARGET_SINGLE] = "Single Target";
    weaponRangedTargettingToCsvToken[RPG::WEAPON_RANGED_TARGET_SCREEN_CENTER] = "Fly Down the Center of the Screen";
    weaponRangedTargettingToCsvToken[RPG::WEAPON_RANGED_TARGET_ALL_SIMUL] = "Strike Each Enemy Simultaneously";
    weaponRangedTargettingToCsvToken[RPG::WEAPON_RANGED_TARGET_ALL_INTURN] = "Strike Each Enemy in Turn";
    for( std::map<RPG::WeaponRangedTargetting, std::string>::iterator iter = weaponRangedTargettingToCsvToken.begin(); iter != weaponRangedTargettingToCsvToken.end(); ++iter ) {
        csvTokenToWeaponRangedTargetting[iter->second] = iter->first;
    }
    itemTypeToCsvToken[RPG::ITEM_COMMON] = "Common Item";
    itemTypeToCsvToken[RPG::ITEM_WEAPON] = "Weapon";
    itemTypeToCsvToken[RPG::ITEM_SHIELD] = "Shield";
    itemTypeToCsvToken[RPG::ITEM_ARMOR] = "Armor";
    itemTypeToCsvToken[RPG::ITEM_HELMET] = "Helmet";
    itemTypeToCsvToken[RPG::ITEM_OTHER] = "Accessory";
    itemTypeToCsvToken[RPG::ITEM_MEDICINE] = "Medicine";
    itemTypeToCsvToken[RPG::ITEM_BOOK] = "Skill Book";
    itemTypeToCsvToken[RPG::ITEM_SEED] = "Seed";
    itemTypeToCsvToken[RPG::ITEM_SCROLL] = "Skill Scroll";
    itemTypeToCsvToken[RPG::ITEM_SWITCH] = "Switch";
    for( std::map<RPG::ItemType, std::string>::iterator iter = itemTypeToCsvToken.begin(); iter != itemTypeToCsvToken.end(); ++iter ) {
        csvTokenToItemType[iter->second] = iter->first;
    }
    itemTargetToCsvToken[RPG::ITEM_TARGET_SINGLE] = "Single Ally";
    itemTargetToCsvToken[RPG::ITEM_TARGET_ALL] = "All Allies";
    for( std::map<RPG::ItemTarget, std::string>::iterator iter = itemTargetToCsvToken.begin(); iter != itemTargetToCsvToken.end(); ++iter ) {
        csvTokenToItemTarget[iter->second] = iter->first;
    }
    itemScrollMsgShownToCsvToken[RPG::ITEM_SCROLL_MSG_ITEM] = "Item Name";
    itemScrollMsgShownToCsvToken[RPG::ITEM_SCROLL_MSG_SKILL] = "Name of Associated Skill";
    for( std::map<RPG::ItemScrollMsgShown, std::string>::iterator iter = itemScrollMsgShownToCsvToken.begin(); iter != itemScrollMsgShownToCsvToken.end(); ++iter ) {
        csvTokenToItemScrollMsgShown[iter->second] = iter->first;
    }
    itemConditionFlagToCsvToken[RPG::ITEM_CONDITION_INFLICT] = "Inflict";
    itemConditionFlagToCsvToken[RPG::ITEM_CONDITION_REMOVE] = "Resist";
    for( std::map<RPG::ItemConditionFlag, std::string>::iterator iter = itemConditionFlagToCsvToken.begin(); iter != itemConditionFlagToCsvToken.end(); ++iter ) {
        csvTokenToItemConditionFlag[iter->second] = iter->first;
    }
    monsterPreconditionToCsvToken[RPG::MON_PRE_NONE] = "None";
    monsterPreconditionToCsvToken[RPG::MON_PRE_SWITCH] = "Switch";
    monsterPreconditionToCsvToken[RPG::MON_PRE_TURNS] = "Turns Elapsed";
    monsterPreconditionToCsvToken[RPG::MON_PRE_MONSTER_PRESENT] = "Enemies Present";
    monsterPreconditionToCsvToken[RPG::MON_PRE_MONSTER_HP] = "Monster HP";
    monsterPreconditionToCsvToken[RPG::MON_PRE_MONSTER_MP] = "Monster MP";
    monsterPreconditionToCsvToken[RPG::MON_PRE_AVG_LEVEL] = "Party Level Average";
    monsterPreconditionToCsvToken[RPG::MON_PRE_AVG_EXHAUSTION] = "Party Exhaustion";
    for( std::map<RPG::MonsterPrecondition_T, std::string>::iterator iter = monsterPreconditionToCsvToken.begin(); iter != monsterPreconditionToCsvToken.end(); ++iter ) {
        csvTokenToMonsterPrecondition[iter->second] = iter->first;
    }
    attributeTypeToCsvToken[RPG::ATTRIBUTE_WEAPON] = "Weapon";
    attributeTypeToCsvToken[RPG::ATTRIBUTE_MAGIC] = "Magic";
    for( std::map<RPG::AttributeType_T, std::string>::iterator iter = attributeTypeToCsvToken.begin(); iter != attributeTypeToCsvToken.end(); ++iter ) {
        csvTokenToAttributeType[iter->second] = iter->first;
    }
    actionRestrictionToCsvToken[RPG::AR_NONE] = "No Restriction";
    actionRestrictionToCsvToken[RPG::AR_NO_ACTION_ALLOWED] = "No Action Allowed";
    actionRestrictionToCsvToken[RPG::AR_ATTACK_ENEMIES_RANDOMLY] = "Attack Enemies Randomly";
    actionRestrictionToCsvToken[RPG::AR_ATTACK_ALLIES_RANDOMLY] = "Attack Allies Randomly";
    for( std::map<RPG::ActionRestriction_T, std::string>::iterator iter = actionRestrictionToCsvToken.begin(); iter != actionRestrictionToCsvToken.end(); ++iter ) {
        csvTokenToActionRestriction[iter->second] = iter->first;
    }
    baseStatAlterationToCsvToken[RPG::BSA_HALF] = "Half";
    baseStatAlterationToCsvToken[RPG::BSA_DOUBLE] = "Double";
    baseStatAlterationToCsvToken[RPG::BSA_NONE] = "None";
    for( std::map<RPG::BaseStatAlteration, std::string>::iterator iter = baseStatAlterationToCsvToken.begin(); iter != baseStatAlterationToCsvToken.end(); ++iter ) {
        csvTokenToBaseStatAlteration[iter->second] = iter->first;
    }
    alterationTypeToCsvToken[RPG::AT_ALTER_DAMAGE] = "Damage";
    alterationTypeToCsvToken[RPG::AT_ALTER_RECOVER] = "Recovery";
    alterationTypeToCsvToken[RPG::AT_ALTER_NONE] = "None";
    for( std::map<RPG::AlterationType, std::string>::iterator iter = alterationTypeToCsvToken.begin(); iter != alterationTypeToCsvToken.end(); ++iter ) {
        csvTokenToAlterationType[iter->second] = iter->first;
    }
    // Import the starting database
    importStartingDatabase();
    // Record the default database
    recordDefaultDatabase();
}

//! Imports the relevant data values from CSV files when the game is loaded
/*! onNewGame is called before the a new game is started. In this plugin this method
    is used to reset the overridden data flags and database to the default state.
*/
void onNewGame()
{
    // Clear all overridden data flags
    clearOverriddenDataFlags();
    // Import the default database
    importDefaultDatabase();
    // Set any active data to match the database
    syncActiveToDatabaseForNewGame();
}

//! Exports any overridden data values to CSV files when the game is saved
/*! onSaveGame is called before the game creates or updates a save file. In this plugin this method
    is used to export any data values that have been overridden (NOT including overrides from
    starting file list) so that they can be imported again when the game state is loaded.

    \param id: (int) Savestate ID.
    \param savePluginData: (void(*)(char *data, int length)) Call this function to save custom plugin data.
*/
void onSaveGame( int id, void __cdecl (*savePluginData)(char *data, int length) )
{
    // Call exportDatabase with the save state file name base and false for export all
    char buffer [3];
    std::string fileName = CSV_FILE_PATH + saveStateFileNameBase;
    fileName.append( itoa( id, buffer, 10 ) );
    exportDatabase( fileName, false );
}

//! Imports the relevant data values from CSV files when the game is loaded
/*! onLoadGame is called before the game loads a save file. In this plugin this method
    is used to import any data values that were recorded for that savestate as overridden (NOT
    including overrides from starting file list).

    \param id: (int) Savestate ID.
    \param data: (char *) Pointer to plugin data.
    \param length: (int) Length of plugin data.
*/
void onLoadGame( int id, char * data, int length )
{
    // Clear all overridden data flags
    clearOverriddenDataFlags();
    // Import the default database
    importDefaultDatabase();
    // Call importDatabase with the save state file name base and true for set overridden flags
    char buffer [3];
    std::string fileName = CSV_FILE_PATH + saveStateFileNameBase;
    fileName.append( itoa( id, buffer, 10 ) );
    importDatabase( fileName, true );
}

//! Respond to potential comment commands
/*!
    onComment is called when the game runs across a comment line in its event scripting. Comment
    commands are how the user communicates with the plugin, so this function checks for comment
    commands and responds accordingly.

    \param text (const char*) The comment's content as simple text
    \param parsedData (const RPG::ParsedCommentData*) The already parsed data
    \param nextScriptLine (RPG::EventScriptLine*) The next event script line after the comment
    \param scriptData (RPG::EventScriptData*) Pointer to the RPG::EventScriptData object of the current event script
    \param eventId (int) The ID of the current event (negative for common events, zero for battle events)
    \param pageId (int) The ID of the current event page (zero for common and battle events - sorry, no battle event page ID yet)
    \param lineId (int) The zero-based line number
    \param nextLineId (int*) Pointer to the next executed line number (-1 for default)
    \return (bool) false will prevent other plugins from receiving this notification, use true otherwise
*/
bool onComment( const char* text,
                const RPG::ParsedCommentData* 	parsedData,
                RPG::EventScriptLine* 	nextScriptLine,
                RPG::EventScriptData* 	scriptData,
                int 	eventId,
                int 	pageId,
                int 	lineId,
                int* 	nextLineId )
{
    #ifdef DYNDATABASEOVERRIDE_DEBUG
    *ofDebug << "onComment called with text: " << text << "\n";
    #endif // DYNDATABASEOVERRIDE_DEBUG

    static std::string cmd;
    static std::string baseFileName;

    // Check for and respond to plugin comment commands
    cmd = parsedData->command;
    if( 0 == cmd.compare( "dyndatabaseoverride_export_database" ) )
    {   // Export the entire current database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        exportDatabase( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_database" ) )
    {   // Import all data from a chosen group of .csv files to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importDatabase( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors" ) )
    {   // Import Actor data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActors( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_hp" ) )
    {   // Import Actor HP data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsHp( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_mp" ) )
    {   // Import Actor MP data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsMp( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_attack" ) )
    {   // Import Actor Attack data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsAttack( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_defense" ) )
    {   // Import Actor Defense data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsDefense( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_intelligence" ) )
    {   // Import Actor Intelligence data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsIntelligence( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_agility" ) )
    {   // Import Actor Agility data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsAgility( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_skill_learning" ) )
    {   // Import Actor Skill Learning data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsSkillLearning( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_conditions" ) )
    {   // Import Actor Conditions data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsConditions( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_actors_attributes" ) )
    {   // Import Actor Attributes data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importActorsAttributes( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes" ) )
    {   // Import Class data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClasses( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_hp" ) )
    {   // Import Class HP data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesHp( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_mp" ) )
    {   // Import Class MP data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesMp( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_attack" ) )
    {   // Import Class Attack data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesAttack( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_defense" ) )
    {   // Import Class Defense data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesDefense( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_intelligence" ) )
    {   // Import Class Intelligence data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesIntelligence( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_agility" ) )
    {   // Import Class Agility data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesAgility( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_skill_learning" ) )
    {   // Import Class Skill Learning data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesSkillLearning( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_conditions" ) )
    {   // Import Class Conditions data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesConditions( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_classes_attributes" ) )
    {   // Import Class Attributes data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importClassesAttributes( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_skills" ) )
    {   // Import Skill data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importSkills( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_skills_conditions" ) )
    {   // Import Skill Conditions data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importSkillsConditions( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_skills_attributes" ) )
    {   // Import Skill Attributes data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importSkillsAttributes( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_skills_animations" ) )
    {   // Import Skill Animations data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importSkillsAnimations( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_items" ) )
    {   // Import Item data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importItems( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_items_usable_by_actors" ) )
    {   // Import Item Usable By Actors data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importItemsUsableByActors( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_items_usable_by_classes" ) )
    {   // Import Item Usable By Classes data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importItemsUsableByClasses( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_items_conditions" ) )
    {   // Import Item Conditions data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importItemsConditions( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_items_attributes" ) )
    {   // Import Item Attributes data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importItemsAttributes( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_items_weapon_animations" ) )
    {   // Import Item Weapon Animations data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importItemsWeaponAnimations( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_enemies" ) )
    {   // Import Enemy data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importEnemies( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_enemies_conditions" ) )
    {   // Import Enemy Conditions data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importEnemiesConditions( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_enemies_attributes" ) )
    {   // Import Enemy Attributes data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importEnemiesAttributes( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_enemies_behaviors" ) )
    {   // Import Enemy Behavior data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importEnemiesBehaviors( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_attributes" ) )
    {   // Import Attribute data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importAttributes( baseFileName, true );
        return false;
    }
    if( 0 == cmd.compare( "dyndatabaseoverride_import_conditions" ) )
    {   // Import Condition data from a chosen .csv file to the database
        baseFileName = CSV_FILE_PATH;
        baseFileName += parsedData->parameters[0].text;
        importConditions( baseFileName, true );
        return false;
    }
    // No comment commands for this plugin detected; pass notification on for other plugins
    return true;
}

#ifdef DYNDATABASEOVERRIDE_DEBUG
//! Clean up after use
/*!
    onExit is called when the game closes. In this plugin this is used to perform any needed
    cleanup.
*/
void onExit()
{
    ofDebug->close();
}
#endif // DYNDATABASEOVERRIDE_DEBUG
